//
// file :         acqthread.cpp
//
// description :  gige camera acquisition thread
//
// project :      TANGO Device Server
//
// $Author: claudio $ Giulio Gaio
//
// $Revision: 1.63 $ 1.0
//
// $Log: acqthread.cpp,v $
// Revision 1.63  2020-05-11 09:05:59  claudio
// #11562 : compile on 1004 x86_64 with tango-9.3.3
//
// Revision 1.62  2019-01-25 16:14:52  giulio
// Fixed bug of 180 degrees rotation + HwROI
//
// Revision 1.61  2019-01-25 08:08:54  giulio
// Support to 180 degrees HWROI support
//
// Revision 1.60  2019-01-24 15:42:29  giulio
// Added aAV2300 and acA1440-73gm
//
// Revision 1.59  2018-10-05 10:03:56  giulio
// Fixed some bugs
//
// Revision 1.58  2018-10-03 13:55:16  giulio
// Allowing set scale/offset when ccd is off, bugfixing
//
// Revision 1.57  2018-06-19 07:06:53  giulio
// Added acA2040-35-gm and acA2500-14gm support, Makefile for pylon5
//
// Revision 1.56  2017-08-25 08:53:13  giulio
// Fixed bug in exposure api
//
// Revision 1.55  2017-04-19 13:00:04  giulio
// If fitting/ sum threshold failed then state is set to ALARM
//
// Revision 1.54  2017-04-19 07:00:53  giulio
// Fixed feedback mode
//
// Revision 1.53  2017-03-20 12:22:22  giulio
// Added aCA2000-50gm, added support for tango based trajectory feedback
//
// Revision 1.52  2016-07-28 11:49:09  giulio
// Added acA1300-75gm
//
// Revision 1.51  2016-04-04 07:39:08  giulio
// Added acA1300-60gm
//
// Revision 1.50  2016-03-16 14:01:47  giulio
// Adde acA645-100gm and acA1920-50gm
//
// Revision 1.49  2016-02-03 09:29:58  giulio
// Added support for scA1600-28gm and acA1300-30gm
//
// Revision 1.48  2015-11-24 10:35:04  giulio
// Integrated acA1300-gm
//
// Revision 1.47  2015-06-10 09:43:56  giulio
// Temporary solution to fix acA1600 digital gain configuration
//
// Revision 1.46  2015-02-10 10:13:04  giulio
// Added acA1600 camera
//
// Revision 1.45  2015-01-19 10:43:31  giulio
// Added av1600 camera, solved problems with ROI
//
// Revision 1.44  2014-12-30 08:33:57  giulio
// Added Model attribute
//
// Revision 1.43  2014-08-28 07:49:45  giulio
// Added imagesum attribute and scA1400 ccd
//
// Revision 1.42  2013-12-04 09:53:20  giulio
// Moved to Pylon3, added some eth tuning properties (band reservation, interpacket delay)
//
// Revision 1.41  2013-08-21 07:59:48  giulio
// Default is with rnm enabled
//
// Revision 1.40  2013-08-07 08:00:26  giulio
// Added mutexes in get_data commands; moved to Tango8; changed shutdown timeouts
//
// Revision 1.39  2013-07-09 08:34:16  giulio
// Added packet size configuration
//
// Revision 1.38  2013-07-08 07:45:14  giulio
// Clean heap memory by using malloc_trim(0)
//
// Revision 1.37  2013-07-01 09:32:36  giulio
// Fixed bug in hw roi (very small roi bug)
//
// Revision 1.36  2013-06-10 14:50:55  giulio
// Supported 75gm 16bit data format
//
// Revision 1.35  2013-05-17 13:14:08  giulio
// Added disable trigger timeout, stream grabber configurable
//
// Revision 1.34  2013-05-16 12:41:23  giulio
// Disable hw roi when image rotated, fix hor hw roi overflow
//
// Revision 1.33  2013-04-11 14:29:53  giulio
// Added digital image stabilization, saved roi_threshold
//
// Revision 1.32  2013-02-12 09:01:37  giulio
// Minimized bug in connect-disconnect with 2.3.3
//
// Revision 1.31  2013-02-08 15:12:34  giulio
// Added autoreconnect disable + avA1000 ccd
//
// Revision 1.30  2013-01-24 08:47:29  giulio
// Added binning
//
// Revision 1.29  2013-01-23 14:01:18  giulio
// Fixed bug on ext. trigger absence (server locked)
//
// Revision 1.28  2012-11-29 15:35:11  giulio
// Fixed 16bit hwroi position rescale
//
// Revision 1.27  2012-11-29 15:30:49  giulio
// Decreased basler n_buffer to 1
//
// Revision 1.26  2012-11-02 15:00:59  giulio
// Initialize background image
//
// Revision 1.25  2012-10-09 16:13:01  giulio
// Fixed bug in the pause mode (lock detected)
//
// Revision 1.24  2012-09-28 12:16:09  giulio
// Shifted ace minimum gains to 0
//
// Revision 1.23  2012-09-28 11:32:30  giulio
// Review of gigevcam gain fitting for Ace ccd
//
// Revision 1.22  2012-09-28 09:09:24  giulio
// Added support for acA780-gm camera
//
// Revision 1.21  2012-09-24 13:56:29  giulio
// Add  hw roi limits check
//
// Revision 1.20  2012-09-18 12:15:39  giulio
// Fixed error counter at first cycle
//
// Revision 1.19  2012-09-17 11:44:53  giulio
// Major updates (hw roi, auto gain/exposure, rnm..)
//
// Revision 1.18  2012-08-09 10:23:46  giulio
// Removed long sleep(500000) in acqthread
//
// Revision 1.17  2012-07-19 09:34:13  giulio
// Delete some delete attribute instructions automatically doubled by pogo7
//
// Revision 1.16  2012-07-19 07:58:03  giulio
// Sleep something after setting FAULT state in acqthread
//
// Revision 1.15  2012-07-11 10:26:54  giulio
// Commented debug stream in ccd api
//
// Revision 1.14  2012-07-02 14:48:36  giulio
// Increased erro msg info in ccd api
//
// Revision 1.13  2012-06-15 15:40:57  giulio
// Fixed bug in start device (wait link flag), splitted exceptions in thread destruction
//
// Revision 1.12  2012-05-30 12:42:54  giulio
// Support for 32/64 bit platforms, fixed init procedure
//
// Revision 1.11  2011-12-28 08:00:53  giulio
// Extended exposure timeout to its maximu value. Store all ccd parameters through SaveScale command
//
// Revision 1.10  2011/10/24 07:31:48  giulio
// Added new ccd support (sony chipset)
//
// Revision 1.9  2011/06/14 12:19:08  giulio
// Fixed GetImag16 bug
//
// Revision 1.8  2011/05/13 09:44:00  giulio
// Fixed some bugs, moved to tango7, supported Pylon 2.3.3
//
// Revision 1.7  2011/01/25 10:24:03  giulio
// Added debug thread attribute
//
// Revision 1.6  2011/01/21 09:01:34  giulio
// Fixed save property issue
//
// Revision 1.5  2010/12/15 09:40:13  giulio
// Added support for scA640 and piA1000
//
// Revision 1.4  2010/08/30 14:54:55  giulio
// Fixed get_image16 bug. Circular buffers return 0 when data is not avaliable
//
// Revision 1.3  2010/04/30 09:00:53  giulio
// Added Gaussian fitting methods
//
// Revision 1.2  2010/04/15 13:25:54  giulio
// New revision
//
// Revision 1.1.1.1  2009/04/20 08:23:30  giulio
// Import iniziale di gigevcam.
//
//
// copyleft :   Sincrotrone Trieste S.C.p.A.
//              Strada Statale 14 - km 163,5 in AREA Science Park
//              34012 Basovizza, Trieste ITALY
//

#include "acqthread.h"
#include <malloc.h>



void ScanName(char *input, char *mod, char* ser, char *ipaddr)
{
	int32_t cnt = 0;
	char *p, *c, tmp[GIGEVCAM_DESC];
	
	memset(tmp,0,GIGEVCAM_DESC);
	memset(ipaddr,0,GIGEVCAM_DESC);
	/* string example: scA780-54gm#0030530C3FBF#10.0.0.10:0 */
	
	p = strtok(input, "#");
	while (p != NULL) {
		if (cnt == 0) {
			sprintf(mod,"%s",p);cnt++;
		}
		else if (cnt == 1) {
			sprintf(ser,"%s",p);cnt++;
		}
		else if (cnt == 2) {
			sprintf(tmp,"%s",p);cnt++;
			/* terminate string at first occurrence of ':' */
			c = strchr(tmp, ':');
			if (c != NULL)
			  *c = 0;
			strncpy(ipaddr,tmp,strlen(tmp));
		}
		p = strtok(NULL, "#");
 
	}

}

#ifdef GIGEVCAM_RNM
int64_t acqthread::ReverseBytes64(int64_t value)
{
  return (value & 0x00000000000000FFULL) << 56 | (value & 0x000000000000FF00ULL) << 40 |
		(value & 0x0000000000FF0000ULL) << 24 | (value & 0x00000000FF000000ULL) << 8 |
		(value & 0x000000FF00000000ULL) >> 8 | (value & 0x0000FF0000000000ULL) >> 24 |
		(value & 0x00FF000000000000ULL) >> 40 | (value & 0xFF00000000000000ULL) >> 56;
}
#endif

/*
 * update_configuration
 * Update on the fly the ImageProc parameters
 */
void acqthread::update_configuration(void)
{

	/* if this is the first run */
	if (myacq->cam_param[myacq->cam_idx].configured == false) {
		/* overload manual roi param in case max image size doesn't match */
		if ((myacq->cam_param[myacq->cam_idx].manual_roi_param[0] > width) ||
				(myacq->cam_param[myacq->cam_idx].manual_roi_param[2] > width) ||
				(myacq->cam_param[myacq->cam_idx].manual_roi_param[1] > height) ||
				(myacq->cam_param[myacq->cam_idx].manual_roi_param[3] > height) ||
				(myacq->cam_param[myacq->cam_idx].manual_roi_param[0] < 0) ||
				(myacq->cam_param[myacq->cam_idx].manual_roi_param[1] < 0) ||
				(myacq->cam_param[myacq->cam_idx].manual_roi_param[2] < 0) ||
				(myacq->cam_param[myacq->cam_idx].manual_roi_param[3] < 0)) {
			myacq->cam_param[myacq->cam_idx].manual_roi_param[0] = 0;
			myacq->cam_param[myacq->cam_idx].manual_roi_param[1] = 0;
			myacq->cam_param[myacq->cam_idx].manual_roi_param[2] = width;
			myacq->cam_param[myacq->cam_idx].manual_roi_param[3] = height;
		}

		/* get sensor width and height */
		myacq->cam_param[myacq->cam_idx].max_width = max_width;
		myacq->cam_param[myacq->cam_idx].max_height = max_height;

		/* main hw parameters */
		CameraSetExposure((int32_t)myacq->cam_param[myacq->cam_idx].exposure);
		CameraSetGain((int32_t) myacq->cam_param[myacq->cam_idx].gain);
		if (myacq->cam_param[myacq->cam_idx].trigger == false)
			CameraSetFrameRate(myacq->cam_param[myacq->cam_idx].frame_rate);

	}

	if (myacq->cam_param[myacq->cam_idx].img_depth == 8) {
		/* set horizontal/vertical reference position offset */
		if (myacq->take_reference == true) {
			myacq->take_reference = false;
			for (int32_t i = 0; i < myacq->imageBufferSize; i++) {
				myacq->img8[i]->ref_hor_pos_offset = myacq->img8[myacq->img_cur]->hor_pos + 2 * myacq->img8[myacq->img_cur]->hor_pos_offset;
				myacq->img8[i]->ref_ver_pos_offset = myacq->img8[myacq->img_cur]->ver_pos;					
			}
		}
		/* set image size */

		myacq->img8[myacq->img_nxt]->set_size(width, height);

		/* set low pass filter order */
		myacq->img8[myacq->img_nxt]->set_lp_order(myacq->cam_param[myacq->cam_idx].lp_order);
		/* set decimation factor */
		myacq->img8[myacq->img_nxt]->set_hor_grid(hor_grid_size);
		myacq->img8[myacq->img_nxt]->set_ver_grid(ver_grid_size);

		/* set mirror */
		myacq->img8[myacq->img_nxt]->set_mirror(myacq->cam_param[myacq->cam_idx].mirror);
		/* enable image processing */
		myacq->img8[myacq->img_nxt]->set_enable_process(myacq->cam_param[myacq->cam_idx].process_enable);
		/* set auto roi */
		myacq->img8[myacq->img_nxt]->set_enable_auto_roi(myacq->cam_param[myacq->cam_idx].auto_roi);
		/* set roi threshold */
		myacq->img8[myacq->img_nxt]->set_roi_threshold(myacq->cam_param[myacq->cam_idx].roi_threshold);
		myacq->img8[myacq->img_nxt]->set_search_background_param(myacq->cam_param[myacq->cam_idx].search_background_param);
		/* plot mode and level */
		myacq->img8[myacq->img_nxt]->set_plot_mode(myacq->cam_param[myacq->cam_idx].plot_mode);
		myacq->img8[myacq->img_nxt]->set_plot_level(myacq->cam_param[myacq->cam_idx].plot_level);
		myacq->img8[myacq->img_nxt]->set_plot_offset_axis(myacq->cam_param[myacq->cam_idx].plot_offset_axis);		
		/* enable fast calculation */
		myacq->img8[myacq->img_nxt]->set_enable_fast_calc(myacq->cam_param[myacq->cam_idx].fast_calc);
		/* enable median filtering */
		myacq->img8[myacq->img_nxt]->set_enable_median_filter(myacq->cam_param[myacq->cam_idx].enable_median_filter);
		/* set manual roi */
		myacq->img8[myacq->img_nxt]->set_enable_manual_roi(myacq->cam_param[myacq->cam_idx].sw_roi);
		myacq->img8[myacq->img_nxt]->set_manual_roi(myacq->cam_param[myacq->cam_idx].manual_roi_param);
		/* set process mode */
		myacq->img8[myacq->img_nxt]->set_algorithm(myacq->cam_param[myacq->cam_idx].process_mode);
		/* update fitting timeout */
		if (myacq->frequency > 0)
			myacq->img8[myacq->img_nxt]->set_fit_tout(1000/myacq->frequency/2); /* leave half of the time period for fitting */
		/* digital image stabilization parameters */
		myacq->img8[myacq->img_nxt]->set_ref_raw_pos(myacq->cam_param[myacq->cam_idx].ref_raw_hor_pos, myacq->cam_param[myacq->cam_idx].ref_raw_ver_pos);
		myacq->img8[myacq->img_nxt]->set_enable_dis(myacq->cam_param[myacq->cam_idx].dis_enable);


		/* if rotation not multiple of 90, enable rotation interpolation */
		if ((int32_t)myacq->cam_param[myacq->cam_idx].rotation % 90) {
			myacq->img8[myacq->img_nxt]->set_enable_rotation_filter(true);
		}
		else
			myacq->img8[myacq->img_nxt]->set_enable_rotation_filter(false);

	}
	else if (myacq->cam_param[myacq->cam_idx].img_depth == 16) {

		/* set horizontal/vertical reference position offset */
		if (myacq->take_reference == true) {
			myacq->take_reference = false;
			for (int32_t i = 0; i < myacq->imageBufferSize; i++) {
				myacq->img16[i]->ref_hor_pos_offset = myacq->img16[myacq->img_cur]->hor_pos  + 2 * myacq->img16[myacq->img_cur]->hor_pos_offset;
				myacq->img16[i]->ref_ver_pos_offset = myacq->img16[myacq->img_cur]->ver_pos;	
			}
		}

		/* set image size */
		myacq->img16[myacq->img_nxt]->set_size((int32_t)width, (int32_t) height);
		/* set low pass filter order */
		myacq->img16[myacq->img_nxt]->set_lp_order(myacq->cam_param[myacq->cam_idx].lp_order);
		/* set decimation factor */
		myacq->img16[myacq->img_nxt]->set_hor_grid(hor_grid_size);
		myacq->img16[myacq->img_nxt]->set_ver_grid(ver_grid_size);

		/* set mirror */
		myacq->img16[myacq->img_nxt]->set_mirror(myacq->cam_param[myacq->cam_idx].mirror);
		/* enable image processing */
		myacq->img16[myacq->img_nxt]->set_enable_process(myacq->cam_param[myacq->cam_idx].process_enable);
		/* set auto roi */
		myacq->img16[myacq->img_nxt]->set_enable_auto_roi(myacq->cam_param[myacq->cam_idx].auto_roi);
		/* set roi threshold */
		myacq->img16[myacq->img_nxt]->set_roi_threshold(myacq->cam_param[myacq->cam_idx].roi_threshold);
		myacq->img16[myacq->img_nxt]->set_search_background_param(myacq->cam_param[myacq->cam_idx].search_background_param);
		/* plot mode and level */
		myacq->img16[myacq->img_nxt]->set_plot_mode(myacq->cam_param[myacq->cam_idx].plot_mode);
		myacq->img16[myacq->img_nxt]->set_plot_level(myacq->cam_param[myacq->cam_idx].plot_level);
		myacq->img16[myacq->img_nxt]->set_plot_offset_axis(myacq->cam_param[myacq->cam_idx].plot_offset_axis);
		/* enable fast calculation */
		myacq->img16[myacq->img_nxt]->set_enable_fast_calc(myacq->cam_param[myacq->cam_idx].fast_calc);
		/* enable median filtering */
		myacq->img16[myacq->img_nxt]->set_enable_median_filter(myacq->cam_param[myacq->cam_idx].enable_median_filter);
		/* set manual roi */
		myacq->img16[myacq->img_nxt]->set_enable_manual_roi(myacq->cam_param[myacq->cam_idx].sw_roi);
		myacq->img16[myacq->img_nxt]->set_manual_roi(myacq->cam_param[myacq->cam_idx].manual_roi_param);
		/* set process mode */
		myacq->img16[myacq->img_nxt]->set_algorithm(myacq->cam_param[myacq->cam_idx].process_mode);
		/* update fitting timeout */
		if (myacq->frequency > 0)
			myacq->img16[myacq->img_nxt]->set_fit_tout(1000/myacq->frequency/2); /* leave half of the time period for fitting */

		/* digital image stabilization parameters */
		myacq->img16[myacq->img_nxt]->set_ref_raw_pos(myacq->cam_param[myacq->cam_idx].ref_raw_hor_pos, myacq->cam_param[myacq->cam_idx].ref_raw_ver_pos);
		myacq->img16[myacq->img_nxt]->set_enable_dis(myacq->cam_param[myacq->cam_idx].dis_enable);

		/* if rotation not multiple of 90, enable rotation interpolation */
		if ((int32_t)myacq->cam_param[myacq->cam_idx].rotation % 90)
			myacq->img16[myacq->img_nxt]->set_enable_rotation_filter(true);
		else
			myacq->img16[myacq->img_nxt]->set_enable_rotation_filter(false);
	}

	/* check permutation external calibration table */		
	if (*myacq->attr_Permutation_read != myacq->cam_param[myacq->cam_idx].permutation) {
		myacq->perform_permutation(*myacq->attr_Permutation_read);
		myacq->cam_param[myacq->cam_idx].permutation = *myacq->attr_Permutation_read;
	}

	myacq->cam_param[myacq->cam_idx].configured = true;

}

/*
 * acqthread::save_data
 */
//bool acqthread::save_data(const GrabResult *Result)
bool acqthread::save_data(void *imagePtr, uint32_t frame_id)
{
	struct timeval tv;
	bool avg_thres_ok = true;

	gettimeofday(&tv,NULL);

	/* 8 bit */
	if (myacq->cam_param[myacq->cam_idx].img_depth == 8) {
		
		if (myacq->trigger_background) {
			myacq->cam_param[myacq->cam_idx].bckimg8->push_permutation((unsigned char*) imagePtr,
				myacq->cam_param[myacq->cam_idx].img_permutation,
				(uint32_t) frame_id, (uint32_t)  myacq->bunch_number);
			myacq->trigger_background = false;
		}
		if (myacq->cam_param[myacq->cam_idx].background_subtraction) {
			myacq->img8[myacq->img_nxt]->push_permutation_with_background((unsigned char*) imagePtr, myacq->cam_param[myacq->cam_idx].img_permutation,
				(uint32_t)frame_id, (uint32_t) myacq->bunch_number,(const unsigned char*)myacq->cam_param[myacq->cam_idx].bckimg8->image);
		}
		else {
			myacq->img8[myacq->img_nxt]->push_permutation((unsigned char*) imagePtr, myacq->cam_param[myacq->cam_idx].img_permutation, 
				(uint32_t) frame_id, (uint32_t)myacq->bunch_number);
		}
		

	}
	/* 16 bit */
	else if (myacq->cam_param[myacq->cam_idx].img_depth == 16) {
		
		if (myacq->trigger_background) {
			myacq->cam_param[myacq->cam_idx].bckimg16->push_permutation((unsigned short*) imagePtr, 
				myacq->cam_param[myacq->cam_idx].img_permutation,
				(uint32_t) frame_id,(uint32_t) myacq->bunch_number);
			myacq->trigger_background = false;
		}
		if (myacq->cam_param[myacq->cam_idx].background_subtraction) {
			myacq->img16[myacq->img_nxt]->push_permutation_with_background((unsigned short*) imagePtr, myacq->cam_param[myacq->cam_idx].img_permutation,
				(uint32_t) frame_id, (uint32_t) myacq->bunch_number,(const unsigned short *)myacq->cam_param[myacq->cam_idx].bckimg16->image);
		}
		else {
			myacq->img16[myacq->img_nxt]->push_permutation((unsigned short*) imagePtr, myacq->cam_param[myacq->cam_idx].img_permutation, 
				(uint32_t) frame_id, (uint32_t) myacq->bunch_number);
		}
		
	}

	myacq->img_cur = myacq->img_nxt;
	myacq->img_nxt++;

	/* save data into extended circular buffers */
	if (myacq->cam_param[myacq->cam_idx].img_depth == 8) {
		

		if (*myacq->attr_ImageSumSamples_read) {
			myacq->in_mutex->lock();
			myacq->imagesum_counter++;
			int img_sub_idx;
			if (myacq->imagesum_counter > *myacq->attr_ImageSumSamples_read) {
				img_sub_idx = (myacq->img_cur + myacq->imageBufferSize - *myacq->attr_ImageSumSamples_read) % myacq->imageBufferSize;
				for (int i = 0; i < myacq->img8[myacq->img_cur]->get_width()*myacq->img8[myacq->img_cur]->get_height(); i++) {
					myacq->attr_ImageSum_read[i] += (Tango::DevLong) myacq->img8[myacq->img_cur]->image[i]-(Tango::DevLong) myacq->img8[img_sub_idx]->image[i];
				}
			}
			else {
				for (int i = 0; i < myacq->img8[myacq->img_cur]->get_width()*myacq->img8[myacq->img_cur]->get_height(); i++) {
					myacq->attr_ImageSum_read[i] += (Tango::DevLong) myacq->img8[myacq->img_cur]->image[i];
				}
			}
			myacq->in_mutex->unlock();
		}

		if (myacq->circ_buffer) {
			push_data((void *) &myacq->img8[myacq->img_cur]->phase, tv, myacq->bunch_number, myacq->buf_phase);
			push_data((void *) &myacq->img8[myacq->img_cur]->area, tv, myacq->bunch_number, myacq->buf_area);
			push_data((void *) &myacq->img8[myacq->img_cur]->hor_pos, tv, myacq->bunch_number, myacq->buf_hor_pos);
			push_data((void *) &myacq->img8[myacq->img_cur]->hor_sigma, tv, myacq->bunch_number, myacq->buf_hor_sigma);
			push_data((void *) &myacq->img8[myacq->img_cur]->saturation, tv, myacq->bunch_number, myacq->buf_saturation);
			push_data((void *) &myacq->img8[myacq->img_cur]->ver_pos, tv, myacq->bunch_number, myacq->buf_ver_pos);
			push_data((void *) &myacq->img8[myacq->img_cur]->ver_sigma, tv, myacq->bunch_number, myacq->buf_ver_sigma);
			push_data((void *) &myacq->img8[myacq->img_cur]->intensity, tv, myacq->bunch_number, myacq->buf_intensity);
			push_data((void *) &myacq->img8[myacq->img_cur]->fit_error, tv, myacq->bunch_number, myacq->buf_fit_error);
			push_data((void *) &myacq->bunch_number, tv, myacq->bunch_number, myacq->buf_acquisition_counter);
			push_data((void *) &myacq->img8[myacq->img_cur]->max_value, tv, myacq->bunch_number, myacq->buf_max_val);
			push_data((void *) myacq->attr_Gain_read, tv, myacq->bunch_number, myacq->buf_gain);
			push_data((void *) myacq->attr_Exposure_read, tv, myacq->bunch_number, myacq->buf_exposure);
			double mean_intensity = 0;
			if ((myacq->img8[myacq->img_cur]->area != 0) && (myacq->img8[myacq->img_cur]->hor_calibration != 0) &&
				(myacq->img8[myacq->img_cur]->ver_calibration != 0)) {
				mean_intensity = myacq->img8[myacq->img_cur]->intensity / 
													myacq->img8[myacq->img_cur]->area /
													myacq->img8[myacq->img_cur]->hor_calibration /
													myacq->img8[myacq->img_cur]->ver_calibration;
			}
			double hor_pos = myacq->img8[myacq->img_cur]->hor_pos, ver_pos = myacq->img8[myacq->img_cur]->ver_pos;
			/* force rnm value to 0 in case of fitting error or signal under threshold */
			*myacq->attr_MeanIntensity_read = mean_intensity;
			if ((mean_intensity < *myacq->attr_MeanIntensityThreshold_read) || myacq->img8[myacq->img_cur]->fit_error) {
				hor_pos = NAN; ver_pos = NAN;
				if (*myacq->attr_FeedbackMode_read) {
					if (*myacq->attr_MeanIntensity_read < (double)*myacq->attr_MeanIntensityThreshold_read) {
						avg_thres_ok = false;
						myacq->set_status("signal under threshold");
					}
					else if (myacq->img8[myacq->img_cur]->fit_error) {
						avg_thres_ok = false;
						myacq->set_status("profile fitting failed");
					}	
				}
			}

#ifdef GIGEVCAM_RNM
			if ((myacq->rnmIndex >= 0) && (myacq->rnmEnabled)){
				char check = RNM_DATA_OK;
				int64_t hor_pos_big_endian = ReverseBytes64(*(int64_t *)&hor_pos);
				int64_t ver_pos_big_endian = ReverseBytes64(*(int64_t *)&ver_pos);
				int64_t mean_intensity_big_endian = ReverseBytes64(*(int64_t *)&mean_intensity);
				rnm_ccd_horpos_set_value(myacq->rnmIndex, myacq->rnmIndex+1, (void *) &hor_pos_big_endian, &check);
				rnm_ccd_verpos_set_value(myacq->rnmIndex, myacq->rnmIndex+1, (void *) &ver_pos_big_endian, &check);
				rnm_ccd_intensity_set_value(myacq->rnmIndex, myacq->rnmIndex+1, (void *) &mean_intensity_big_endian, &check);
			}
#endif

		}
	}
	else if (myacq->cam_param[myacq->cam_idx].img_depth == 16) {

		if (*myacq->attr_ImageSumSamples_read) {
			myacq->in_mutex->lock();
			myacq->imagesum_counter++;
			int img_sub_idx;
			if (myacq->imagesum_counter >= *myacq->attr_ImageSumSamples_read) {
				img_sub_idx = (myacq->img_cur + myacq->imageBufferSize - *myacq->attr_ImageSumSamples_read) % myacq->imageBufferSize;
				for (int i = 0; i < myacq->img16[myacq->img_cur]->get_width()*myacq->img16[myacq->img_cur]->get_height(); i++) {
					myacq->attr_ImageSum_read[i] += (Tango::DevLong) myacq->img16[myacq->img_cur]->image[i]-(Tango::DevLong) myacq->img16[img_sub_idx]->image[i];
				}
			}
			else {
				for (int i = 0; i < myacq->img16[myacq->img_cur]->get_width()*myacq->img16[myacq->img_cur]->get_height(); i++) {
					myacq->attr_ImageSum_read[i] += (Tango::DevLong) myacq->img16[myacq->img_cur]->image[i];
				}
			}
			myacq->in_mutex->unlock();
		}

		if (myacq->circ_buffer) {
			push_data((void *) &myacq->img16[myacq->img_cur]->phase, tv, myacq->bunch_number, myacq->buf_phase);
			push_data((void *) &myacq->img16[myacq->img_cur]->area, tv, myacq->bunch_number, myacq->buf_area);
			push_data((void *) &myacq->img16[myacq->img_cur]->hor_pos, tv, myacq->bunch_number, myacq->buf_hor_pos);
			push_data((void *) &myacq->img16[myacq->img_cur]->hor_sigma, tv, myacq->bunch_number, myacq->buf_hor_sigma);
			push_data((void *) &myacq->img16[myacq->img_cur]->saturation, tv, myacq->bunch_number, myacq->buf_saturation);
			push_data((void *) &myacq->img16[myacq->img_cur]->ver_pos, tv, myacq->bunch_number, myacq->buf_ver_pos);
			push_data((void *) &myacq->img16[myacq->img_cur]->ver_sigma, tv, myacq->bunch_number, myacq->buf_ver_sigma);
			push_data((void *) &myacq->img16[myacq->img_cur]->intensity, tv, myacq->bunch_number, myacq->buf_intensity);
			push_data((void *) &myacq->img16[myacq->img_cur]->fit_error, tv, myacq->bunch_number, myacq->buf_fit_error);
			push_data((void *) &myacq->bunch_number, tv, myacq->bunch_number, myacq->buf_acquisition_counter);
			push_data((void *) &myacq->img16[myacq->img_cur]->max_value, tv, myacq->bunch_number, myacq->buf_max_val);
			push_data((void *) myacq->attr_Gain_read, tv, myacq->bunch_number, myacq->buf_gain);
			push_data((void *) myacq->attr_Exposure_read, tv, myacq->bunch_number, myacq->buf_exposure);
		}

		double mean_intensity = 0;
		if ((myacq->img16[myacq->img_cur]->area != 0) && (myacq->img16[myacq->img_cur]->hor_calibration != 0) &&
			(myacq->img16[myacq->img_cur]->ver_calibration != 0)) {
			mean_intensity = myacq->img16[myacq->img_cur]->intensity / 
												myacq->img16[myacq->img_cur]->area /
												myacq->img16[myacq->img_cur]->hor_calibration /
												myacq->img16[myacq->img_cur]->ver_calibration;
		}
		double hor_pos = myacq->img16[myacq->img_cur]->hor_pos,
						ver_pos = myacq->img16[myacq->img_cur]->ver_pos;
		/* force rnm value to 0 in case of fitting error or signal under threshold */
		*myacq->attr_MeanIntensity_read = mean_intensity; 
		if ((mean_intensity < *myacq->attr_MeanIntensityThreshold_read) || myacq->img16[myacq->img_cur]->fit_error) {
			hor_pos = NAN; ver_pos = NAN;
			if (*myacq->attr_FeedbackMode_read) {
				if (*myacq->attr_MeanIntensity_read < (double)*myacq->attr_MeanIntensityThreshold_read) {
					avg_thres_ok = false;
					myacq->set_status("signal under threshold");
				}
				else if (myacq->img8[myacq->img_cur]->fit_error) {
					avg_thres_ok = false;
					myacq->set_status("profile fitting failed");
				}	
			}
		}

#ifdef GIGEVCAM_RNM
		if ((myacq->rnmIndex >= 0) && (myacq->rnmEnabled)){
			char check = RNM_DATA_OK;
			int64_t hor_pos_big_endian = ReverseBytes64(*(int64_t *)&hor_pos);
			int64_t ver_pos_big_endian = ReverseBytes64(*(int64_t *)&ver_pos);
			int64_t mean_intensity_big_endian = ReverseBytes64(*(int64_t *)&mean_intensity);
			rnm_ccd_horpos_set_value(myacq->rnmIndex, myacq->rnmIndex+1, (void *) &hor_pos_big_endian, &check);
			rnm_ccd_verpos_set_value(myacq->rnmIndex, myacq->rnmIndex+1, (void *) &ver_pos_big_endian, &check);
			rnm_ccd_intensity_set_value(myacq->rnmIndex, myacq->rnmIndex+1, (void *) &mean_intensity_big_endian, &check);
		}
#endif
	}

	if (myacq->img_nxt >= myacq->imageBufferSize)
		myacq->img_nxt = 0;

	if (avg_thres_ok) {
		*myacq->attr_FeedbackError_read = false;
	}
	else {
		*myacq->attr_FeedbackError_read = true;
	}

	return avg_thres_ok;

}

/*
 * acqthread::
 */
bool acqthread::init_data(void)
{
	myacq->in_mutex->lock();

	{ string msg_log("init_data: start"); myacq->info_log(msg_log); }

	try {
		int64_t img_width = width;
		int64_t img_height = height;

		/* vector used for coefficient permutation for image rotation */
		//myacq->cam_param[myacq->cam_idx].img_permutation = new long(img_height*img_width);

		myacq->cam_param[myacq->cam_idx].img_permutation = (int32_t *) malloc(sizeof(int32_t)*img_height*img_width);
		{ string msg_log("init_data: malloc img_permutation done"); myacq->info_log(msg_log); }
		for (int32_t i = 0; i < (img_height*img_width); i++) {
			myacq->cam_param[myacq->cam_idx].img_permutation[i] = i;
		}

		/* initialize feedback counters and flags */
		myacq->cam_param[myacq->cam_idx].exposure_auto_cnt = 0;
		myacq->cam_param[myacq->cam_idx].gain_auto_cnt = 0;
		myacq->cam_param[myacq->cam_idx].exposure_auto_once = false;
		myacq->cam_param[myacq->cam_idx].gain_auto_once = false;
		myacq->cam_param[myacq->cam_idx].permutation = false;

		/* optimize grid sizes */
		if ((myacq->cam_param[myacq->cam_idx].gridsize != 0) && (img_width != 0) &&
			(img_width >= myacq->cam_param[myacq->cam_idx].gridsize)) {
			hor_grid_size = (int)((double)img_width/(int)(img_width/myacq->cam_param[myacq->cam_idx].gridsize));
		}
		else {
			hor_grid_size = myacq->cam_param[myacq->cam_idx].gridsize;
		}
		/* set vertical grid */
		if ((myacq->cam_param[myacq->cam_idx].gridsize != 0) && (img_height != 0) &&
			(img_height >= myacq->cam_param[myacq->cam_idx].gridsize)) {
			ver_grid_size = (int)((double)img_height/(int)(img_width/myacq->cam_param[myacq->cam_idx].gridsize));
		}
		else {
			ver_grid_size = myacq->cam_param[myacq->cam_idx].gridsize;
		}

		{ string msg_log("init_data: optimizing grid sizes done"); myacq->info_log(msg_log); }

		memset(myacq->attr_ImageSum_read, 0, img_height * img_width * sizeof(Tango::DevLong));
		myacq->imagesum_counter = 0;

		if (myacq->cam_param[myacq->cam_idx].img_depth == 8) {
			/* initialize the image structure */

			for (int32_t i = 0; i < myacq->imageBufferSize; i++) {

				ImageProc <unsigned char> *tmp = new ImageProc <unsigned char>(img_height,img_width,(int32_t)myacq->fitNumCores);
				if (!tmp)
					return false;

				tmp->set_lp_order((int32_t) myacq->cam_param[myacq->cam_idx].lp_order);
				tmp->set_hor_grid((int32_t) hor_grid_size);
				tmp->set_ver_grid((int32_t) ver_grid_size);
				tmp->binning = myacq->cam_param[myacq->cam_idx].binning;
				tmp->set_enable_auto_roi(myacq->autoRoi);
				tmp->set_enable_process(myacq->cam_param[myacq->cam_idx].process_enable);
				tmp->set_enable_median_filter(myacq->cam_param[myacq->cam_idx].enable_median_filter);
				tmp->set_mirror(myacq->cam_param[myacq->cam_idx].mirror);

				tmp-> set_enable_fast_calc(myacq->fastCalculation);
				tmp->set_plot_mode(myacq->plotMode[0]);
				tmp->set_roi_threshold(myacq->autoRoiParam[2]);
				myacq->img8.push_back(tmp);

				/* set calibration and mechanical offsets */
				tmp->hor_calibration = myacq->cam_param[myacq->cam_idx].hor_calibration;
				tmp->ver_calibration = myacq->cam_param[myacq->cam_idx].ver_calibration;
				tmp->hor_pos_offset = myacq->cam_param[myacq->cam_idx].hor_pos_offset * myacq->cam_param[myacq->cam_idx].hor_sign;
				//tmp->ver_pos_offset = myacq->cam_param[myacq->cam_idx].ver_pos_offset * myacq->cam_param[myacq->cam_idx].ver_sign;
				tmp->ver_pos_offset = myacq->cam_param[myacq->cam_idx].ver_pos_offset;

				tmp->hor_sign = myacq->cam_param[myacq->cam_idx].hor_sign;
				tmp->ver_sign = myacq->cam_param[myacq->cam_idx].ver_sign;

				tmp->ref_hor_pos_offset = myacq->cam_param[myacq->cam_idx].ref_hor_pos_offset + myacq->cam_param[myacq->cam_idx].delta_ref_hor_pos_offset;
				tmp->ref_ver_pos_offset = myacq->cam_param[myacq->cam_idx].ref_ver_pos_offset + myacq->cam_param[myacq->cam_idx].delta_ref_ver_pos_offset;

				/* hw roi change image offset */
				if (myacq->cam_param[myacq->cam_idx].hw_roi == true) {
					if (myacq->cam_param[myacq->cam_idx].mirror == false) {
						tmp->img_hor_pos_offset = tmp->img_hor_pos_offset + (max_width/2-
							tmp->img_hor_pos_offset - myacq->cam_param[myacq->cam_idx].hw_roi_param[0]);
					}
					else {
						tmp->img_hor_pos_offset = tmp->img_hor_pos_offset + (max_width/2-
							tmp->img_hor_pos_offset - myacq->cam_param[myacq->cam_idx].hw_roi_param[0]);
					}
					tmp->img_ver_pos_offset = tmp->img_ver_pos_offset + (max_height/2-
						tmp->img_ver_pos_offset-myacq->cam_param[myacq->cam_idx].hw_roi_param[1]);
				}
				/* binning change image offset */
				if (myacq->cam_param[myacq->cam_idx].binning > 1) {
					if (myacq->cam_param[myacq->cam_idx].mirror == false) {
						
						tmp->img_hor_pos_offset = tmp->img_hor_pos_offset + (max_width/myacq->cam_param[myacq->cam_idx].binning-
							tmp->img_hor_pos_offset - max_width/2/myacq->cam_param[myacq->cam_idx].binning);
					}
					else {
						tmp->img_hor_pos_offset = tmp->img_hor_pos_offset + (max_width/myacq->cam_param[myacq->cam_idx].binning-
							tmp->img_hor_pos_offset - max_width/2/myacq->cam_param[myacq->cam_idx].binning);
					}
					tmp->img_ver_pos_offset = tmp->img_ver_pos_offset + (max_height/myacq->cam_param[myacq->cam_idx].binning-
						tmp->img_ver_pos_offset-max_height/2/myacq->cam_param[myacq->cam_idx].binning);
				}


			}

			/* initialize permutation vector */
			if (myacq->cam_param[myacq->cam_idx].mirror)
				myacq->img8[0]->permutation_mirrorh(myacq->cam_param[myacq->cam_idx].img_permutation);
			if (myacq->cam_param[myacq->cam_idx].rotation != 0)
				myacq->img8[0]->permutation_rotation(myacq->cam_param[myacq->cam_idx].rotation, myacq->cam_param[myacq->cam_idx].img_permutation);

			/* background subtraction support variable */
			myacq->cam_param[myacq->cam_idx].bckimg8 = 
				new ImageProc <unsigned char> (img_height,img_width,(int32_t)myacq->fitNumCores);

			myacq->cam_param[myacq->cam_idx].bckimg8->reset();
	
		} /* end image 8bit */
		else if (myacq->cam_param[myacq->cam_idx].img_depth == 16) {
			/* initialize the image structure */
			for (int32_t i = 0; i < myacq->imageBufferSize; i++) {

				ImageProc <unsigned short> *tmp = new ImageProc <unsigned short>(img_height,img_width,(int32_t)myacq->fitNumCores);
				if (!tmp)
					return false;

				tmp->set_lp_order((int32_t)myacq->autoRoiParam[1]);
				tmp->set_hor_grid(hor_grid_size);
				tmp->set_ver_grid(ver_grid_size);
				tmp->binning = myacq->cam_param[myacq->cam_idx].binning;
				tmp->set_enable_auto_roi(myacq->autoRoi);
				tmp->set_enable_process(myacq->cam_param[myacq->cam_idx].process_enable);
				tmp->set_enable_median_filter(myacq->cam_param[myacq->cam_idx].enable_median_filter);
				tmp->set_mirror(myacq->cam_param[myacq->cam_idx].mirror);

				tmp-> set_enable_fast_calc(myacq->fastCalculation);
				tmp->set_plot_mode(myacq->plotMode[0]);
				tmp->set_roi_threshold(myacq->autoRoiParam[2]);
				myacq->img16.push_back(tmp);
				/* set calibration and mechanical offsets */
				tmp->hor_calibration = myacq->cam_param[myacq->cam_idx].hor_calibration;
				tmp->ver_calibration = myacq->cam_param[myacq->cam_idx].ver_calibration;
				tmp->hor_pos_offset = myacq->cam_param[myacq->cam_idx].hor_pos_offset * myacq->cam_param[myacq->cam_idx].hor_sign;
				tmp->ver_pos_offset = myacq->cam_param[myacq->cam_idx].ver_pos_offset * myacq->cam_param[myacq->cam_idx].ver_sign;
				tmp->hor_sign = myacq->cam_param[myacq->cam_idx].hor_sign;
				tmp->ver_sign = myacq->cam_param[myacq->cam_idx].ver_sign;				

				tmp->ref_hor_pos_offset = myacq->cam_param[myacq->cam_idx].ref_hor_pos_offset + myacq->cam_param[myacq->cam_idx].delta_ref_hor_pos_offset;
				tmp->ref_ver_pos_offset = myacq->cam_param[myacq->cam_idx].ref_ver_pos_offset + myacq->cam_param[myacq->cam_idx].delta_ref_ver_pos_offset;


				/* hw roi changes image offset */
				if (myacq->cam_param[myacq->cam_idx].hw_roi == true) {
					if (myacq->cam_param[myacq->cam_idx].mirror == false) {
						tmp->img_hor_pos_offset = tmp->img_hor_pos_offset + (max_width/2-
						tmp->img_hor_pos_offset - myacq->cam_param[myacq->cam_idx].hw_roi_param[0]);
					}
					else {
						tmp->img_hor_pos_offset = tmp->img_hor_pos_offset + (max_width/2-
						tmp->img_hor_pos_offset - myacq->cam_param[myacq->cam_idx].hw_roi_param[0]);
					}
					tmp->img_ver_pos_offset = tmp->img_ver_pos_offset + (max_height/2-
						tmp->img_ver_pos_offset-myacq->cam_param[myacq->cam_idx].hw_roi_param[1]);
				}
				/* binning changes image offset */
				if (myacq->cam_param[myacq->cam_idx].binning > 1) {
					if (myacq->cam_param[myacq->cam_idx].mirror == false) {
						tmp->img_hor_pos_offset = tmp->img_hor_pos_offset + (max_width/myacq->cam_param[myacq->cam_idx].binning-
							tmp->img_hor_pos_offset - max_width/2/myacq->cam_param[myacq->cam_idx].binning);
					}
					else {
						tmp->img_hor_pos_offset = tmp->img_hor_pos_offset + (max_width/myacq->cam_param[myacq->cam_idx].binning-
							tmp->img_hor_pos_offset - max_width/2/myacq->cam_param[myacq->cam_idx].binning);
					}
					tmp->img_ver_pos_offset = tmp->img_ver_pos_offset + (max_height/myacq->cam_param[myacq->cam_idx].binning-
						tmp->img_ver_pos_offset-max_height/2/myacq->cam_param[myacq->cam_idx].binning);
				}

			}

			/* initialize permutation vector */
			if (myacq->cam_param[myacq->cam_idx].mirror)
				myacq->img16[0]->permutation_mirrorh(myacq->cam_param[myacq->cam_idx].img_permutation);
			if (myacq->cam_param[myacq->cam_idx].rotation != 0)
				myacq->img16[0]->permutation_rotation(myacq->cam_param[myacq->cam_idx].rotation, myacq->cam_param[myacq->cam_idx].img_permutation);

			/* support variable for background subtraction */
			myacq->cam_param[myacq->cam_idx].bckimg16 = 
				new ImageProc <unsigned short> (img_height,img_width,(int32_t)myacq->fitNumCores);

			myacq->cam_param[myacq->cam_idx].bckimg16->reset();

		}

		{ string msg_log("init_data: image buffer allocated"); myacq->info_log(msg_log); }

	}
	catch (...) {
		cout << "acqthread::init_data(): exception" << endl;
	}

	myacq->in_mutex->unlock();

	return true;

}


void acqthread::delete_data(void)
{
	bool img8_flag = false,
		img16_flag = false;

	myacq->in_mutex->lock();

	try {
		if ( !myacq->img8.empty() ) {
			img8_flag = true;
			delete myacq->cam_param[myacq->cam_idx].bckimg8;
		}

		if( !myacq->img16.empty() ) {
			img16_flag = true;
			delete myacq->cam_param[myacq->cam_idx].bckimg16;
		}

		if (myacq->cam_param[myacq->cam_idx].img_permutation) {
			free((void *)myacq->cam_param[myacq->cam_idx].img_permutation);
			myacq->cam_param[myacq->cam_idx].img_permutation = 0;
		}

		/* keep permutation index and background image when disconnected */
		while( !myacq->img8.empty() ) {
			delete myacq->img8.back();
			myacq->img8.pop_back();
		}
		if (img8_flag)
			myacq->img8.clear();

		while( !myacq->img16.empty() ) {
			delete myacq->img16.back();
			myacq->img16.pop_back();
		}
		if (img16_flag)
			myacq->img16.clear();

	}
	catch (...) {
		cout << "acqthread::delete_data(): exception" << endl;
	}
	
	myacq->in_mutex->unlock();

}

bool acqthread::CameraGetFrameRate(double *frame_rate)
{

	*frame_rate = 0; 
	return true;

}

// set exposure value	
bool acqthread::CameraSetFrameRate(double frame_rate)
{	

	return true;

}


// get exposure value
bool acqthread::CameraGetExposure(int32_t *exposure)
{
	try {
		dev_camera->read_attribute("Exposure") >> *exposure;
	}	
	catch(...) {
		*exposure = 0;
	}		
	return true;

}


// set exposure value	
bool acqthread::CameraSetExposure(int32_t exposure)
{	

	return true;

}


// get gain
bool acqthread::CameraGetGain(double *gain)
{
	try {
		dev_camera->read_attribute("Gain") >> *gain;
		*gain = digital_gain;
	}	
	catch(...) {
		*gain = 0;
	}		
	return true;
}

// set gain	
bool acqthread::CameraSetGain(double gain)
{
	digital_gain = gain;
	return true;

}

// get trigger mode
bool acqthread::CameraGetImageDepth(short *imagedepth)
{

	*imagedepth = 0;
	
	return true;

}

// get trigger mode
bool acqthread::CameraGetTrigger(bool *trigger)
{
	try {
		dev_camera->read_attribute("Trigger") >> *trigger;
	}
	catch(...) {
		*trigger = false;
	}
	return true;
}

// get binning
bool acqthread::CameraGetBinning(int32_t *binning)
{
	try {
		dev_camera->read_attribute("Binning") >> *binning;
	}
	catch(...) {
		*binning = 1;
	}			
	return true;
}


// get hw roi
bool acqthread::CameraGetHWRoi(bool *hwroi)
{
	try {
		dev_camera->read_attribute("HWRoi") >> *hwroi;
	}
	catch(...) {
		*hwroi = false;
	}	
	return true;
}


// get trigger mode
bool acqthread::CameraSetTrigger(bool trigger)
{

	return true; 
}


// manage roi
void acqthread::CameraSetRoi(void)
{


}

// set acquisition mode
void acqthread::CameraSetAcqMode(void)
{
	

}

// set acquisition mode
void acqthread::CameraInit(void)
{



}

// set acquisition mode
void acqthread::StreamGrabberInit(void)
{
	/* allocate streamgrabber */

}

// delete myCamera
void acqthread::CameraDelete(void)
{	
	//if (&myCamera != NULL) {
	if (1) {
		// Close camera
		if (open_flag == true) {
			//myCamera->Close();
			{ string msg_log("switch off camera"); myacq->info_log(msg_log); }
		}
		// deleted mycamera object
		//delete myCamera;
		{ 
                 std::string msg_log("camera deleted");
                 myacq->info_log(msg_log);
                 }
	}
}

// delete myCamera
void acqthread::StreamGrabberDelete(void)
{

}

// automatic gain control feedback
bool acqthread::auto_gain_feedback()
{
	int32_t max_value = 0;
	double err = 0;
	bool run_loop = false, deadband_flag = false;

	if ((myacq->cam_param[myacq->cam_idx].gain_auto == true) || 
		   myacq->cam_param[myacq->cam_idx].gain_auto_once)  {
		/* first get the maximum pixel value */
		if (myacq->cam_param[myacq->cam_idx].img_depth == 8)
			max_value = myacq->img8[myacq->img_cur]->max_value;
		else if (myacq->cam_param[myacq->cam_idx].img_depth == 16)
			max_value = myacq->img16[myacq->img_cur]->max_value;
		/* then calculate the error */
		err = myacq->cam_param[myacq->cam_idx].auto_target_value - max_value;
		/* if error over threshold, enable deadband flag */
		if (abs(err) > myacq->cam_param[myacq->cam_idx].auto_feedback_deadband) {
			deadband_flag = true;
		}
		/* if setpoint reached, disable out of deadband flag */
		if (abs(err) <= 1)
			deadband_flag = false;
		/* if out of deadband, run feedback */
		if (deadband_flag)
			run_loop = true;
		else
			run_loop = false;

		/* run once flag gain control (run once for max 30 shots) */
		if (myacq->cam_param[myacq->cam_idx].gain_auto_once) {
			if (myacq->cam_param[myacq->cam_idx].gain_once_cnt++ > 30) {
				myacq->cam_param[myacq->cam_idx].gain_auto_once = false;
				myacq->cam_param[myacq->cam_idx].gain_once_cnt = 0;
			}
			else {
				run_loop = true;
				/* trick to allow run once also when under threshold */
				if (max_value < myacq->cam_param[myacq->cam_idx].auto_feedback_target_thres)
					max_value = myacq->cam_param[myacq->cam_idx].auto_target_value/2+1;
			}
		}
		/* if signal under threshold, pause feedback */
		if (myacq->cam_param[myacq->cam_idx].auto_feedback_target_thres > max_value)
			run_loop = false;
		/* initialize minimum/maximum gain raw */
		if (myacq->cam_param[myacq->cam_idx].gain_auto_cnt == 0) {
			myacq->cam_param[myacq->cam_idx].gain_auto_coeff = 0;
		}
		/* downsampling counter */
		if ((myacq->cam_param[myacq->cam_idx].gain_auto_cnt % (myacq->cam_param[myacq->cam_idx].auto_feedback_dws+1)) != 0) 
			run_loop = false;
		/* run loop */
		if (run_loop) {

			/* feedback loop core */
			if ((max_value >= myacq->cam_param[myacq->cam_idx].auto_feedback_target_thres) && (myacq->cam_param[myacq->cam_idx].auto_target_value > 0)) {
				if (err > 0) {
					digital_gain = digital_gain + err * myacq->cam_param[myacq->cam_idx].auto_feedback_gain;
				}
				else if (err < 0) {
					digital_gain = digital_gain + err * myacq->cam_param[myacq->cam_idx].auto_feedback_gain;
				}
				if (digital_gain > myacq->cam_param[myacq->cam_idx].gain_auto_max)
					digital_gain = myacq->cam_param[myacq->cam_idx].gain_auto_max;
				else if (digital_gain <= myacq->cam_param[myacq->cam_idx].gain_auto_min)
					digital_gain = myacq->cam_param[myacq->cam_idx].gain_auto_min;					
			}

		} /* end if run_loop */
		/* increment loop counter */
		myacq->cam_param[myacq->cam_idx].gain_auto_cnt++;
	}

	return true;


}

// automatic exposure control feedback
bool acqthread::auto_exposure_feedback()
{

	return true;

}

// stop streaming
void acqthread::CameraStop()
{
	myacq->run_flag = false;
}


void acqthread::run(void *) 
{

	int ReturnVal = SUCCESS;
	unsigned int count, camera_index = 0;
	bool avg_thres_ok = true,
		init_data_done_flag = false,
		deviceproxy_done_flag = false,
		alloc_frame_buffers = false,
		init_buffers_done_flag = false;

	int32_t cnt_err = 0, cnt_err_old = 0, cnt_frame = 0, old_frame = 0, curr_frame = 0, diff_frame,
		bn_err = 0, bn_err_old = 0;
	struct timespec time_old, time_new, tim_bn_err, tim_cnt_err, now, 
		start_wait_new_img, start_acq_new_img, start_read_attributes;
	int32_t old_bunch_number = 0, img_width = 0, img_height = 0;

	bool err_flag = false;

	double gain_freq_loop = 1;

	Tango::DeviceAttribute myAttribute;
	vector<unsigned char> imageArray8;
	vector<unsigned short> imageArray16;	
	vector<uint32_t> imageArray32;

	myacq->link_flag = true;
	myacq->set_state(Tango::INIT);
	gettimeofday(&myacq->thread_state_time,NULL);

	digital_gain = 1;

	try {
		dev_camera = 0;
		dev_camera = new Tango::DeviceProxy(myacq->cam_param[myacq->cam_idx].ipaddress);
		Tango::DevLong img_width, img_weight;
		Tango::DevShort img_depth;
		dev_camera->read_attribute("Width") >> img_width;
		dev_camera->read_attribute("Height") >> img_height;
		max_width = img_width; width = max_width;
		max_height = img_height; height = max_height;
		dev_camera->read_attribute("ImageDepth") >> img_depth;
		myacq->cam_param[myacq->cam_idx].img_depth = img_depth;	
		*myacq->attr_ImageDepth_read = myacq->cam_param[myacq->cam_idx].img_depth;
		// connect also to the storage device (for spectrometer)
		if (strlen(myacq->imageStorageDevice.c_str()) > 0) 
			dev_camera_storage = new Tango::DeviceProxy(myacq->imageStorageDevice);
		else
			dev_camera_storage = dev_camera;
		deviceproxy_done_flag = true;	
	}	
	catch(Tango::DevFailed &e) {
		err_flag = true;
	}	
	catch(...) {
		err_flag = true;
	}

	
	
	
		
	if (err_flag == false)	
		init_data_done_flag = init_data();
	
	if (err_flag == false)
		init_buffers_done_flag = myacq->init_buffers(myacq); 

	clock_gettime(CLOCK_REALTIME, &start_read_attributes);

	bool acq_flag = myacq->run_flag;
	while ((acq_flag) && (err_flag == false)) {

		acq_flag = myacq->run_flag;
		
		if (myacq->pause_flag == true) {
			myacq->set_status("acquisition paused");
			usleep(10000);
			gettimeofday(&myacq->thread_state_time,NULL);
		}
		else {
			// wait new image
			if (err_flag == false) {
				clock_gettime(CLOCK_REALTIME, &start_wait_new_img);
				try {
					dev_camera->read_attribute("CameraCounter") >> curr_frame;
					while (old_frame == curr_frame) {		
						dev_camera->read_attribute("CameraCounter") >> curr_frame;
						clock_gettime(CLOCK_REALTIME, &now);
						if (myacq->timespec2ms(myacq->difftime(start_wait_new_img,now)) > GIGEVCAM_IMG_TIME_OUT) {
							err_flag = true;
							break;
						}
						if (*myacq->attr_FrameRate_read > 0)
							usleep(500000/(*myacq->attr_FrameRate_read));		
						else
							usleep(100000);	
					}
				}	
				catch(Tango::DevFailed &e) {
					err_flag = true;
				}	
				catch(...) {
					err_flag = true;
				}	
			}				
		
			// acquire new image
			if (err_flag == false) {
				clock_gettime(CLOCK_REALTIME, &start_acq_new_img);			
				try {
					if (myacq->cam_param[myacq->cam_idx].img_depth == 8) {
						*myacq->attr_ImageDepth_read = 8;
						if (*myacq->attr_NumAverageFrames_read <= 1) {
							myAttribute = dev_camera->read_attribute("Image8");
							myAttribute >> imageArray8;
							if ((myAttribute.get_dim_y() != width) || (myAttribute.get_dim_x() != height)) {
								width = myAttribute.get_dim_x(); max_width = width;
								height = myAttribute.get_dim_y(); max_height = height;
								update_configuration();
							}
							int32_t img_size = width*height;
							for (int j = 0; j < img_size; j++) {
								imageArray8[j] = imageArray8[j] * digital_gain;
							}										
							save_data(imageArray8.data(), curr_frame);
						}
						else {
							Tango::DeviceData Din,Dout;
							vector <int32_t> data_in;
							data_in.push_back(0); // read through the bunch number
							data_in.push_back((int32_t)*myacq->attr_NumAverageFrames_read);
							Din << data_in;
							Dout = dev_camera_storage->command_inout("GetImage8", Din);
							Dout >> imageArray8;
							int32_t img_size = width*height;
							
							if (imageArray8.size() == (img_size*(*myacq->attr_NumAverageFrames_read))) {
								imageArray32.resize(img_size);
								std::fill(imageArray32.begin(), imageArray32.end(), 0);
								for (int i = 0; i < *myacq->attr_NumAverageFrames_read; i++) {
									for (int j = 0; j < img_size; j++) {
										imageArray32[j] += imageArray8[j + i * img_size];
									}								
								}
								imageArray8.resize(img_size);
								for (int j = 0; j < img_size; j++) {
									imageArray8[j] = imageArray32[j] * digital_gain;
								}
							}
							else {
								imageArray8.resize(img_size);
								std::fill(imageArray8.begin(), imageArray8.end(), 0);
							}
							save_data(imageArray8.data(), curr_frame);							
						}
					}
					else if (myacq->cam_param[myacq->cam_idx].img_depth == 16) {
						*myacq->attr_ImageDepth_read = 16;
						if (*myacq->attr_NumAverageFrames_read <= 1) {
							myAttribute = dev_camera->read_attribute("Image16");
							myAttribute >> imageArray16;
							if ((myAttribute.get_dim_y() != width) || (myAttribute.get_dim_x() != height)) {
								width = myAttribute.get_dim_x(); max_width = width;
								height = myAttribute.get_dim_y(); max_height = height;
								update_configuration();
							}
							int32_t img_size = width*height;
							for (int j = 0; j < img_size; j++) {
								imageArray16[j] = imageArray16[j] * digital_gain;
							}												
							save_data(imageArray16.data(), curr_frame);
						}
						else {
							Tango::DeviceData Din,Dout;
							vector <int32_t> data_in;
							data_in.push_back(0); // read through the bunch number
							data_in.push_back((int32_t)*myacq->attr_NumAverageFrames_read);
							Din << data_in;
							Dout = dev_camera_storage->command_inout("GetImage16", Din);
							Dout >> imageArray16;
							int32_t img_size = width*height;
							
							if (imageArray16.size() == (img_size*(*myacq->attr_NumAverageFrames_read))) {
								imageArray32.resize(img_size);
								std::fill(imageArray32.begin(), imageArray32.end(), 0);
								for (int i = 0; i < *myacq->attr_NumAverageFrames_read; i++) {
									for (int j = 0; j < img_size; j++) {
										imageArray32[j] += imageArray16[j + i * img_size];
									}								
								}
								imageArray16.resize(img_size);
								for (int j = 0; j < img_size; j++) {
									imageArray16[j] = imageArray32[j] * digital_gain;
								}
																											
							}
							else {
								imageArray16.resize(img_size);
								std::fill(imageArray16.begin(), imageArray16.end(), 0);
							}
							save_data(imageArray16.data(), curr_frame);							
						}						
						
								
					}
					else {
						err_flag = true;
					}
				}
				catch(Tango::DevFailed &e) {
					err_flag = true;
				}	
				catch(...) {
					err_flag = true;
				}	
			}
							

			if (err_flag == false) {
		
				// get bunch number from rnm
#ifdef GIGEVCAM_RNM
				if (myacq->rnmEnabled) {
					char check_flag;
#ifdef CONFIG_64
					rnm_bunch_number_get_value(0, 1, &myacq->bunch_number, &check_flag);
#else
					rnm_bunch_number_get_value(0, 1, (long *) &myacq->bunch_number, &check_flag);
#endif
					/* vcheck bunch number lost */
					if ((old_bunch_number > 0) && ((myacq->bunch_number - old_bunch_number) != 1)) {
						bn_err++;
						myacq->error_counters[GIGEVCAM_ERR_BN_LOST]++;
					}
					old_bunch_number = myacq->bunch_number;

					
				}
				else {
					myacq->bunch_number = curr_frame;
				}
#else
				myacq->bunch_number = curr_frame;
#endif
				// end get bunch number from rnm

				// check lost packets, tune fitting period timeout
				diff_frame = abs((old_frame % 0xfffe)-(curr_frame % 0xfffe));
				old_frame = curr_frame;
				
				if (diff_frame > 1) {
					cnt_err++;
					myacq->error_counters[GIGEVCAM_ERR_CNT_LOST]++;
				} 
				clock_gettime(CLOCK_REALTIME, &now);

				cnt_frame++;
				clock_gettime(CLOCK_REALTIME, &time_new);
				double curr_frequency = 1/(myacq->timespec2ms(myacq->difftime(time_old,time_new)));
				if ((curr_frequency > 0) && (curr_frequency < 700)) {
					myacq->frequency = curr_frequency*0.1+myacq->frequency*0.9; /* low pass filter */
					
					if (myacq->frequency < *myacq->attr_FrameRate_read)
						gain_freq_loop = gain_freq_loop - 0.001;
					else
						gain_freq_loop = gain_freq_loop + 0.001;
							
				}
				time_old.tv_sec = time_new.tv_sec; time_old.tv_nsec = time_new.tv_nsec;

				update_configuration();
			
				if (myacq->timespec2ms(myacq->difftime(start_read_attributes,now)) > 1) {
					try {
						CameraGetTrigger(myacq->attr_Trigger_read);
						CameraGetGain(myacq->attr_Gain_read);
						CameraGetExposure(myacq->attr_Exposure_read);
						CameraGetBinning(myacq->attr_Binning_read);
						CameraGetHWRoi(myacq->attr_HwRoi_read);
						clock_gettime(CLOCK_REALTIME, &start_read_attributes);	
					}
					catch(Tango::DevFailed &e) {
						err_flag = true;
						break;
					}	
					catch(...) {
						err_flag = true;
						break;
					}																
				}	

				auto_gain_feedback();
				auto_exposure_feedback();

				/* notify frame lost error */
				stringstream err_msg_cnt, err_msg_bn;
				if (abs(cnt_err_old-cnt_err) > 0) 
					clock_gettime(CLOCK_REALTIME, &tim_cnt_err);
				if (myacq->timespec2ms(myacq->difftime(tim_cnt_err,now)) < GIGEVCAM_ERR_TIME_BARRIER)  
					err_msg_cnt << "missing frames error  ";

				/* notify bunch number lost error */
				if (abs(bn_err_old-bn_err) > 0) 
					clock_gettime(CLOCK_REALTIME, &tim_bn_err);
				if (myacq->timespec2ms(myacq->difftime(tim_bn_err,now)) < GIGEVCAM_ERR_TIME_BARRIER)  
					err_msg_bn << "bunch number error";

				/* store old error counters */
				cnt_err_old = cnt_err;bn_err_old = bn_err;
				clock_gettime(CLOCK_REALTIME, &now);

				/* make error message persistent */
				if (((myacq->timespec2ms(myacq->difftime(tim_cnt_err,now)) < GIGEVCAM_ERR_TIME_BARRIER) ||
						(myacq->timespec2ms(myacq->difftime(tim_bn_err,now)) < GIGEVCAM_ERR_TIME_BARRIER))
				    && (myacq->get_state() == (Tango::ON))) {
					stringstream err_msg;
					err_msg << err_msg_cnt.str() << err_msg_bn.str();
					myacq->set_status(err_msg.str());
				}
				
				myacq->set_state(Tango::ON);
				if (avg_thres_ok == true) { 
					if ((myacq->timespec2ms(myacq->difftime(tim_bn_err,now)) > GIGEVCAM_ERR_TIME_BARRIER) && 
						(myacq->timespec2ms(myacq->difftime(tim_cnt_err,now)) > GIGEVCAM_ERR_TIME_BARRIER)) {
						myacq->set_status("acquisition running");
					}
					myacq->set_state(Tango::ON);
				}
				else {
					myacq->set_state(Tango::ALARM);
				}				
				
				
			}
			else if (myacq->get_state() != (Tango::FAULT)) {
				myacq->set_state(Tango::OFF);
				myacq->set_status("acquisition failed, stream disabled");	
				string msg_log("camera connected,"); 
				char dbg_thread[20]; sprintf(dbg_thread,"trig_err(%d/%d)",(int)myacq->trig_err_consec_cnt,GIGEVCAM_TRIG_ERR_MAX_CNT);
				msg_log.append(dbg_thread); myacq->info_log(msg_log);

				*myacq->attr_DebugThread_read = 42;
				usleep(200000);
				if ((myacq->trig_err_consec_cnt++ > GIGEVCAM_TRIG_ERR_MAX_CNT) && (myacq->autoReconnect == false)) {
					myacq->run_flag = false;
					myacq->trig_err_consec_cnt = 0;
					myacq->set_state(Tango::OFF);
					myacq->set_status("acquisition stopped");
					string msg_log("Maximum number waiting trigger reached, acquisition stopped"); myacq->info_log(msg_log);
				}

			} /* if EVT_CameraGetFrame(...) == SUCCESS) */
			
			while (1) {
				clock_gettime(CLOCK_REALTIME, &now);
				if ((myacq->timespec2ms(myacq->difftime(start_wait_new_img,now)) > (1 / *myacq->attr_FrameRate_read * gain_freq_loop)) || (*myacq->attr_FrameRate_read <= 0))
					break;
				usleep(10000); 
			}

		} /* if pause flag */

		gettimeofday(&myacq->thread_state_time,NULL);
		//if (myacq->run_flag == false)
		//	break;

	}

	gettimeofday(&myacq->thread_state_time,NULL);


	if (init_data_done_flag)
		delete_data();

	if (init_buffers_done_flag)
		myacq->delete_buffers(myacq);

	if ((myacq->run_flag == false) && (err_flag == false)) {
		myacq->set_state(Tango::OFF); myacq->set_status("acquisition stopped");
	}

	if ((deviceproxy_done_flag == false) || (err_flag == true)) {
		myacq->set_state(Tango::FAULT); myacq->set_status("camera not found or not supported");
	}

	if ((deviceproxy_done_flag == true) && (err_flag == true)) {
		myacq->set_state(Tango::FAULT); myacq->set_status("communication error");
	}

	if (dev_camera)
		delete dev_camera;

	myacq->link_flag = false;
	
	return;
	
} 
















