//
// file :         acqthread.h.cpp
//
// description :  thread class
//
// project :      TANGO Device Server
//
// $Author: claudio $ Giulio Gaio
//
// $Revision: 1.10 $ 1.0
//
// $Log: acqthread.h,v $
// Revision 1.10  2020-05-11 09:05:59  claudio
// #11562 : compile on 1004 x86_64 with tango-9.3.3
//
// Revision 1.9  2017-03-20 12:22:22  giulio
// Added aCA2000-50gm, added support for tango based trajectory feedback
//
// Revision 1.8  2013-08-07 08:00:26  giulio
// Added mutexes in get_data commands; moved to Tango8; changed shutdown timeouts
//
// Revision 1.7  2013-07-01 09:32:36  giulio
// Fixed bug in hw roi (very small roi bug)
//
// Revision 1.6  2012-09-17 11:44:53  giulio
// Major updates (hw roi, auto gain/exposure, rnm..)
//
// Revision 1.5  2012-05-30 12:42:54  giulio
// Support for 32/64 bit platforms, fixed init procedure
//
// Revision 1.4  2011-06-14 12:19:08  giulio
// Fixed GetImag16 bug
//
// Revision 1.3  2011/05/13 09:44:00  giulio
// Fixed some bugs, moved to tango7, supported Pylon 2.3.3
//
// Revision 1.2  2010/04/15 13:25:54  giulio
// New revision
//
// Revision 1.2  2009/04/22 09:46:14  giulio
// Added rt_buffer for long term acquisition
//
// Revision 1.1.1.1  2009/04/20 08:23:30  giulio
// Import iniziale di gigevcam.
//
//
// copyleft :   Sincrotrone Trieste S.C.p.A.
//              Strada Statale 14 - km 163,5 in AREA Science Park
//              34012 Basovizza, Trieste ITALY
// 

#ifndef ACQTHREAD_H
#define ACQTHREAD_H

#include <omnithread.h>
#include "GigeVCam.h"

#define GIGEVCAM_ACQTHREAD_NUM_RX         5
#define GIGEVCAM_ACQTHREAD_TRIGGER_IDX    0
#define GIGEVCAM_ACQTHREAD_GAIN_IDX       1
#define GIGEVCAM_ACQTHREAD_EXPOSURE_IDX   2
#define GIGEVCAM_ACQTHREAD_BLACKLEVEL_IDX 3
#define GIGEVCAM_ACQTHREAD_IMAGEDEPTH_IDX 4

#define GIGEVCAM_AUTO_MODE_OFF   0
#define GIGEVCAM_AUTO_MODE_ONCE  1  /* run once (30 shots) */
#define GIGEVCAM_AUTO_MODE_ON    2  /* continuos */

#define GIGEVCAM_MAX_CAMERAS 100
#define MAX_CAMERAS 100
#define MAX_FRAMES 3

#define GIGEVCAM_FRAMES_BUFFERS 30

#define SUCCESS 0

class acqthread : public omni_thread
{
	
	public:
		/* acqthread::acqthread(): initialize init loop */
		acqthread(GigeVCam_ns::GigeVCam *s) {myacq = s;};
		/* acqthread::~acqthread(): object destructor */
		~acqthread() {};
		void CameraStop();
		bool CameraGetTriggerFrequency(float *);
		bool CameraGetTrigger(bool *);
		bool CameraSetTrigger(bool);
		bool CameraGetImageDepth(short *);
		bool CameraSetGain(double);
		bool CameraGetGain(double *);
		bool CameraSetExposure(int32_t);
		bool CameraGetExposure(int32_t *);
		bool CameraSetFrameRate(double);
		bool CameraGetFrameRate(double *);		
		bool CameraSetTriggerDelay(int32_t);
		bool CameraGetTriggerDelay(int32_t *);
		bool CameraGetHWRoi(bool *);
		bool CameraGetBinning(int32_t *);
		void StreamGrabberDelete(void);
		void CameraDelete(void);
		void delete_data(void);
		long Width(void) {return width;};
		long Height(void){return height;};

/*typedef struct EmergentFrame
{
	enum PIXEL_FORMAT pixel_type;
	unsigned int size_x;      
	unsigned int size_y;      
	int stride;
	unsigned int offset_x;      
	unsigned int offset_y;      
	unsigned int padding_x;      
	unsigned int padding_y;      
	unsigned int trailer_size_y;
	unsigned short frame_id;
	UCHAR_PTR imagePtr;
	unsigned int bufferSize; 
	unsigned __int64 timestamp;
	unsigned __int64  nsecs;
	int convertColor;
	int convertBitDepth;
	unsigned char privateData[256]; // place holder of c++ class: Emergent::CEmergentFrame
} EmergentFrame;*/

/*
struct GigEVisionDeviceInfo
{
	// camera info, sent from camera
	unsigned short specVersionMajor;
	unsigned short specVersionMinor;
	unsigned int deviceMode;
	char macAddress[18];
	unsigned int ipConfigOptions;
	unsigned int ipConfigCurrent;
	char currentIp[IPV4_BYTES_SIZE];
	char currentSubnetMask[IPV4_BYTES_SIZE];
	char defaultGateway[IPV4_BYTES_SIZE];
	char manufacturerName[32];
	char modelName[32];
	char deviceVersion[32];
	char manufacturerSpecifiedInfo[48];
	char serialNumber[16];
	char userDefinedName[16];
	
	// info of the nic that camera is connected to
	struct NetworkInterfaceController nic;

	//bool operator==(const GigEVisionDeviceInfo& rhs) const;
};*/
		
	private:

		Tango::DeviceProxy *dev_camera, *dev_camera_storage;
		GigeVCam_ns::GigeVCam *myacq;
		struct timeval camera_get_trigger,        /* 0 */
			             camera_get_gain,      /* 1 */
			             camera_get_exposure,       /* 2 */
			             camera_get_framerate,    /* 3 */
			             camera_get_imagedepth,    /* 4 */
			             camera_get_binning;       /* 5 */
		int32_t camera_get_counter;
		uint32_t width, height, max_width, max_height, hor_grid_size, ver_grid_size, offset_x, offset_y;
		bool init_camera_flag, open_flag, streamgrabber_flag;
		bool CameraSetup(uint32_t);
		void CameraSetRoi(void);
		bool CameraGrab(void);
		void CameraSetAcqMode(void);
		void CameraInit(void);
		void StreamGrabberInit(void);
		void ProcessImage(unsigned char *);
		//bool save_data(const GrabResult *Result);
		bool save_data(void *imagePtr, uint32_t frame_id);
		void update_configuration(void);
#ifdef GIGEVCAM_RNM
		int64_t ReverseBytes64(int64_t);
#endif
		bool init_data(void);
		bool auto_gain_feedback();
		bool auto_exposure_feedback();
		double digital_gain;

	protected:
		/* acqthread::run():	start init loop */
		void run(void *);
		
};	/* end class acqthread() */


#endif

/* EOF */
