/*----- PROTECTED REGION ID(GigeVCam.cpp) ENABLED START -----*/
//=============================================================================
//
// file :        GigeVCam.cpp
//
// description : C++ source for the GigeVCam and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               GigeVCam are implemented in this file.
//
// project :     Gigabit Camera Server.
//
// $Author: giulio $
//
// $Revision: 1.58 $
// $Date: 2019-03-09 11:32:40 $
//
// SVN only:
// $HeadURL:  $
//
// CVS only:
// $Source: /home/cvsadm/cvsroot/fermi/servers/gigevcam/src/GigeVCam.cpp,v $
// $Log: GigeVCam.cpp,v $
// Revision 1.58  2019-03-09 11:32:40  giulio
// Rnmshare initialized in main.cpp to avoid crashes at rnmshare_init
//
// Revision 1.57  2019-01-25 08:08:54  giulio
// Support to 180 degrees HWROI support
//
// Revision 1.56  2019-01-24 15:42:29  giulio
// Added aAV2300 and acA1440-73gm
//
// Revision 1.55  2018-10-05 15:18:58  giulio
// Bugfixing when saving offsets
//
// Revision 1.54  2018-10-05 10:03:56  giulio
// Fixed some bugs
//
// Revision 1.53  2018-10-03 13:55:16  giulio
// Allowing set scale/offset when ccd is off, bugfixing
//
// Revision 1.52  2018-06-19 07:06:53  giulio
// Added acA2040-35-gm and acA2500-14gm support, Makefile for pylon5
//
// Revision 1.51  2017-04-19 07:00:53  giulio
// Fixed feedback mode
//
// Revision 1.50  2017-03-20 12:22:22  giulio
// Added aCA2000-50gm, added support for tango based trajectory feedback
//
// Revision 1.49  2016-07-28 11:49:09  giulio
// Added acA1300-75gm
//
// Revision 1.48  2016-04-04 07:39:07  giulio
// Added acA1300-60gm
//
// Revision 1.47  2016-03-16 14:01:47  giulio
// Adde acA645-100gm and acA1920-50gm
//
// Revision 1.46  2015-05-26 12:45:41  giulio
// Fixed get_area bug, fixed calib x/y when binning=2
//
// Revision 1.45  2015-02-16 13:26:53  giulio
// Temporary commit
//
// Revision 1.44  2015-02-10 10:13:04  giulio
// Added acA1600 camera
//
// Revision 1.43  2015-01-19 10:43:30  giulio
// Added av1600 camera, solved problems with ROI
//
// Revision 1.42  2014-12-30 08:33:57  giulio
// Added Model attribute
//
// Revision 1.41  2014-08-28 07:49:45  giulio
// Added imagesum attribute and scA1400 ccd
//
// Revision 1.40  2013-12-04 09:53:20  giulio
// Moved to Pylon3, added some eth tuning properties (band reservation, interpacket delay)
//
// Revision 1.39  2013-08-07 08:00:26  giulio
// Added mutexes in get_data commands; moved to Tango8; changed shutdown timeouts
//
// Revision 1.38  2013-07-09 08:34:16  giulio
// Added packet size configuration
//
// Revision 1.37  2013-07-01 09:32:36  giulio
// Fixed bug in hw roi (very small roi bug)
//
// Revision 1.36  2013-06-17 11:52:56  giulio
// Bug in get_imag8 fixed
//
// Revision 1.35  2013-06-17 07:03:23  giulio
// Fixed bug in get profiles
//
// Revision 1.34  2013-05-29 12:49:17  giulio
// Implemented hwroiparam
//
// Revision 1.33  2013-05-20 14:36:40  giulio
// Fixed bug in DIS algorithm when roi closed to the borders
//
// Revision 1.32  2013-05-17 13:14:08  giulio
// Added disable trigger timeout, stream grabber configurable
//
// Revision 1.31  2013-05-16 12:41:23  giulio
// Disable hw roi when image rotated, fix hor hw roi overflow
//
// Revision 1.30  2013-04-16 14:45:37  giulio
// Removed rpath, added rt-buffer 6,7 mode to profile data
//
// Revision 1.29  2013-04-11 14:29:53  giulio
// Added digital image stabilization, saved roi_threshold
//
// Revision 1.28  2013-02-12 09:01:37  giulio
// Minimized bug in connect-disconnect with 2.3.3
//
// Revision 1.27  2013-02-08 15:12:34  giulio
// Added autoreconnect disable + avA1000 ccd
//
// Revision 1.26  2013-01-24 08:47:29  giulio
// Added binning
//
// Revision 1.25  2013-01-23 14:01:18  giulio
// Fixed bug on ext. trigger absence (server locked)
//
// Revision 1.24  2012-09-28 09:09:24  giulio
// Added support for acA780-gm camera
//
// Revision 1.23  2012-09-17 11:44:53  giulio
// Major updates (hw roi, auto gain/exposure, rnm..)
//
// Revision 1.22  2012-07-19 09:34:13  giulio
// Delete some delete attribute instructions automatically doubled by pogo7
//
// Revision 1.21  2012-07-02 16:06:11  giulio
// Added saveparam command
//
// Revision 1.20  2012-06-15 15:40:57  giulio
// Fixed bug in start device (wait link flag), splitted exceptions in thread destruction
//
// Revision 1.19  2012-06-14 08:20:12  giulio
// Fixed bug in modify configuration (do not store ccd hw parameters)
//
// Revision 1.18  2012-06-11 11:51:50  giulio
// run-flag forced to false in stop command
//
// Revision 1.17  2012-05-30 12:42:54  giulio
// Support for 32/64 bit platforms, fixed init procedure
//
// Revision 1.16  2012-01-30 12:47:01  giulio
// Added attributes for image transfer optimization
//
// Revision 1.15  2012-01-18 11:16:28  giulio
// Fixed a minor bug in get buffer keys in bunch data mode
//
// Revision 1.14  2011/12/28 08:00:53  giulio
// Extended exposure timeout to its maximu value. Store all ccd parameters through SaveScale command
//
// Revision 1.13  2011/06/14 12:19:08  giulio
// Fixed GetImag16 bug
//
// Revision 1.12  2011/05/13 09:44:00  giulio
// Fixed some bugs, moved to tango7, supported Pylon 2.3.3
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <tango.h>
#include <GigeVCam.h>
#include <GigeVCamClass.h>
#include "acqthread.h"

#include <check_get_rt_buffer.h>

/*----- PROTECTED REGION END -----*/	//	GigeVCam.cpp

/**
 *  GigeVCam class description:
 *    This server connects to a camera using an external thread. Inside this thread, images are acquired and stored
 *    in an array of objects that also do the image processing.
 *    Currently this server support Basler Gigabit cameras.
 *    To make the server work correctly, it is necessary to setup the ethernet interface with
 *    an mtu > 8000 and disable autonegotiation.
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name           |  Method name
//================================================================
//  State                  |  Inherited (no method)
//  Status                 |  Inherited (no method)
//  Start                  |  start
//  Stop                   |  stop
//  Reset                  |  reset
//  ListCamera             |  list_camera
//  AcquireBackground      |  acquire_background
//  GetImage8              |  get_image8
//  GetImage16             |  get_image16
//  GetHorPos              |  get_hor_pos
//  GetVerPos              |  get_ver_pos
//  GetHorSigma            |  get_hor_sigma
//  GetVerSigma            |  get_ver_sigma
//  GetArea                |  get_area
//  GetAcquisitionCounter  |  get_acquisition_counter
//  GetSaturation          |  get_saturation
//  SetReference           |  set_reference
//  GetPhase               |  get_phase
//  SaveScale              |  save_scale
//  RestoreScale           |  restore_scale
//  SetMeasurePoint        |  set_measure_point
//  ClearMeasurePoints     |  clear_measure_points
//  GetIntensity           |  get_intensity
//  GetHorProfile          |  get_hor_profile
//  GetVerProfile          |  get_ver_profile
//  GetHorFitProfile       |  get_hor_fit_profile
//  GetVerFitProfile       |  get_ver_fit_profile
//  SaveParam              |  save_param
//  ClearReference         |  clear_reference
//  ExposureAutoOnce       |  exposure_auto_once
//  GainAutoOnce           |  gain_auto_once
//  GetFitError            |  get_fit_error
//  GetBunchNumber         |  get_bunch_number
//  GetMaxVal              |  get_max_val
//  GetGain                |  get_gain
//  GetExposure            |  get_exposure
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  HorPos                   |  Tango::DevDouble	Scalar
//  VerPos                   |  Tango::DevDouble	Scalar
//  MaxVal                   |  Tango::DevLong	Scalar
//  CountMaxVal              |  Tango::DevLong	Scalar
//  HorSigma                 |  Tango::DevDouble	Scalar
//  VerSigma                 |  Tango::DevDouble	Scalar
//  Area                     |  Tango::DevDouble	Scalar
//  HVCovar                  |  Tango::DevDouble	Scalar
//  CameraCounter            |  Tango::DevLong	Scalar
//  AcquisitionCounter       |  Tango::DevLong	Scalar
//  Height                   |  Tango::DevLong	Scalar
//  Width                    |  Tango::DevLong	Scalar
//  ProcessingTime           |  Tango::DevLong	Scalar
//  Frequency                |  Tango::DevDouble	Scalar
//  Trigger                  |  Tango::DevBoolean	Scalar
//  Exposure                 |  Tango::DevLong	Scalar
//  Gain                     |  Tango::DevDouble	Scalar
//  RoiThreshold             |  Tango::DevDouble	Scalar
//  AutoRoi                  |  Tango::DevBoolean	Scalar
//  SwRoi                    |  Tango::DevBoolean	Scalar
//  HwRoi                    |  Tango::DevBoolean	Scalar
//  PlotMode                 |  Tango::DevShort	Scalar
//  IconSize                 |  Tango::DevLong	Scalar
//  LowPassOrder             |  Tango::DevShort	Scalar
//  Rotation                 |  Tango::DevDouble	Scalar
//  Mirror                   |  Tango::DevBoolean	Scalar
//  ImageDepth               |  Tango::DevShort	Scalar
//  BackgroundSubtraction    |  Tango::DevBoolean	Scalar
//  EnableProcessing         |  Tango::DevBoolean	Scalar
//  FastProcessing           |  Tango::DevBoolean	Scalar
//  Saturation               |  Tango::DevDouble	Scalar
//  PlotLevel                |  Tango::DevShort	Scalar
//  Pause                    |  Tango::DevBoolean	Scalar
//  Label                    |  Tango::DevString	Scalar
//  HorPosRelative           |  Tango::DevDouble	Scalar
//  VerPosRelative           |  Tango::DevDouble	Scalar
//  Phase                    |  Tango::DevDouble	Scalar
//  ScaleHor                 |  Tango::DevDouble	Scalar
//  ScaleVer                 |  Tango::DevDouble	Scalar
//  OffsetHor                |  Tango::DevDouble	Scalar
//  OffsetVer                |  Tango::DevDouble	Scalar
//  PlotOffsetAxis           |  Tango::DevBoolean	Scalar
//  Intensity                |  Tango::DevDouble	Scalar
//  SignHor                  |  Tango::DevDouble	Scalar
//  SignVer                  |  Tango::DevDouble	Scalar
//  ProcessMode              |  Tango::DevLong	Scalar
//  DebugThread              |  Tango::DevLong	Scalar
//  FrameRate                |  Tango::DevDouble	Scalar
//  ImageBufferSize          |  Tango::DevLong	Scalar
//  ImageAcquisitionCounter  |  Tango::DevLong	Scalar
//  ExposureAutoEnable       |  Tango::DevBoolean	Scalar
//  ExposureAutoMin          |  Tango::DevLong	Scalar
//  ExposureAutoMax          |  Tango::DevLong	Scalar
//  GainAutoEnable           |  Tango::DevBoolean	Scalar
//  GainAutoMin              |  Tango::DevDouble	Scalar
//  GainAutoMax              |  Tango::DevDouble	Scalar
//  AutoTargetValue          |  Tango::DevLong	Scalar
//  AutoFeedbackGain         |  Tango::DevDouble	Scalar
//  AutoFeedbackDws          |  Tango::DevLong	Scalar
//  AutoFeedbackTargetThres  |  Tango::DevLong	Scalar
//  AutoFeedbackDeadband     |  Tango::DevLong	Scalar
//  FitError                 |  Tango::DevLong	Scalar
//  BunchNumber              |  Tango::DevLong	Scalar
//  MeanIntensityThreshold   |  Tango::DevLong	Scalar
//  Binning                  |  Tango::DevLong	Scalar
//  DISEnable                |  Tango::DevBoolean	Scalar
//  DisableTriggerTimeout    |  Tango::DevBoolean	Scalar
//  ImageSumSamples          |  Tango::DevLong	Scalar
//  Model                    |  Tango::DevString	Scalar
//  FeedbackMode             |  Tango::DevBoolean	Scalar
//  MeanIntensity            |  Tango::DevDouble	Scalar
//  FeedbackError            |  Tango::DevBoolean	Scalar
//  EnableMedianFilter       |  Tango::DevBoolean	Scalar
//  Permutation              |  Tango::DevBoolean	Scalar
//  MasterCamera             |  Tango::DevString	Scalar
//  NumAverageFrames         |  Tango::DevLong	Scalar
//  SearchBackgroundParam    |  Tango::DevDouble	Spectrum  ( max = 2)
//  HorProfile               |  Tango::DevDouble	Spectrum  ( max = 4200)
//  VerProfile               |  Tango::DevDouble	Spectrum  ( max = 4200)
//  ErrorCounters            |  Tango::DevLong	Spectrum  ( max = 10)
//  TimeStamp                |  Tango::DevLong	Spectrum  ( max = 2)
//  AutoRoiParam             |  Tango::DevLong	Spectrum  ( max = 4)
//  RoiParam                 |  Tango::DevLong	Spectrum  ( max = 4)
//  HorFitProfile            |  Tango::DevDouble	Spectrum  ( max = 4200)
//  VerFitProfile            |  Tango::DevDouble	Spectrum  ( max = 4200)
//  HwRoiParam               |  Tango::DevLong	Spectrum  ( max = 4)
//  Image8                   |  Tango::DevUChar	Image  ( max = 4200 x 4200)
//  Image8Icon               |  Tango::DevLong	Image  ( max = 800 x 800)
//  Image16                  |  Tango::DevUShort	Image  ( max = 4200 x 4200)
//  Image16Icon              |  Tango::DevLong	Image  ( max = 800 x 800)
//  Image8Counter            |  Tango::DevUChar	Image  ( max = 4200 x 4200)
//  Image16Counter           |  Tango::DevUShort	Image  ( max = 4200 x 4200)
//  ImageSum                 |  Tango::DevLong	Image  ( max = 4200 x 4200)
//================================================================

namespace GigeVCam_ns
{
/*----- PROTECTED REGION ID(GigeVCam::namespace_starting) ENABLED START -----*/

	//	static initializations

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : GigeVCam::GigeVCam()
 *	Description : Constructors for a Tango device
 *                implementing the classGigeVCam
 */
//--------------------------------------------------------
GigeVCam::GigeVCam(Tango::DeviceClass *cl, string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(GigeVCam::constructor_1) ENABLED START -----*/

	init_device();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::constructor_1
}
//--------------------------------------------------------
GigeVCam::GigeVCam(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(GigeVCam::constructor_2) ENABLED START -----*/

	init_device();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::constructor_2
}
//--------------------------------------------------------
GigeVCam::GigeVCam(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(GigeVCam::constructor_3) ENABLED START -----*/

	init_device();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : GigeVCam::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void GigeVCam::delete_device()
{
	DEBUG_STREAM << "GigeVCam::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::delete_device) ENABLED START -----*/

	//	Delete device allocated objects
	//	Delete device's allocated object

	/* delete camera parameters struct */

	{ string msg_log("delete_device start"); info_log(msg_log); }
	struct timeval start, now;

	run_flag = false;
	pause_flag = false;
	*attr_DisableTriggerTimeout_read = false;

	gettimeofday(&start,NULL);

	while (link_flag) {
		usleep(500000); 
		gettimeofday(&now,NULL);
		if ((now.tv_sec-start.tv_sec) > (0.6*GIGEVCAM_LOCK_TOUT)) {
			string msg_log("Failed to release acquisition thread, shutting down"); 
			info_log(msg_log);
			exit(1);
		}
	}

	delete [] cam_param;
	delete in_mutex;

#ifdef GIGEVCAM_RNM
	// This RNM api crashes the server
	//if (rnmEnabled) {
	//	rnmshare_deinit();
	//}
#endif

	delete attr_HorPos_read;
	delete attr_VerPos_read;
	delete attr_Intensity_read;
	delete attr_HorSigma_read;
	delete attr_VerSigma_read;
	delete attr_Area_read;

	delete attr_Phase_read;
	delete attr_Gain_read;
	delete [] attr_SearchBackgroundParam_read;

	delete attr_HorPosRelative_read;
	delete attr_VerPosRelative_read;
	delete attr_ScaleHor_read;
	delete attr_ScaleVer_read;
	delete attr_OffsetHor_read;
	delete attr_OffsetVer_read;
	delete attr_HVCovar_read;
	delete attr_Frequency_read;
	delete attr_RoiThreshold_read;
	delete attr_Rotation_read;
	delete attr_SignHor_read;
	delete attr_SignVer_read;

	delete attr_DebugThread_read;
	delete attr_MaxVal_read;
	delete attr_Saturation_read;
	delete attr_CountMaxVal_read;
	delete attr_Exposure_read;
	delete attr_ProcessMode_read;
	delete [] attr_TimeStamp_read;
	delete [] attr_ErrorCounters_read;
	delete attr_CameraCounter_read;
	delete attr_AcquisitionCounter_read;
	delete attr_ProcessingTime_read;
	delete attr_Height_read;
	delete attr_Width_read;
	delete attr_IconSize_read;

	delete attr_PlotMode_read;
	delete attr_LowPassOrder_read;
	delete attr_ImageDepth_read;
	delete attr_PlotLevel_read;

	delete attr_Trigger_read;
	delete attr_AutoRoi_read;
	delete attr_SwRoi_read;
	delete attr_HwRoi_read;
	delete attr_Permutation_read;	
	delete attr_Mirror_read;
	delete attr_BackgroundSubtraction_read;
	delete attr_EnableProcessing_read;
	delete attr_FastProcessing_read;
	delete attr_Pause_read;
	delete attr_PlotOffsetAxis_read;
 	delete attr_FitError_read;
	delete attr_BunchNumber_read;

	if (*attr_Label_read) 
		delete *attr_Label_read;
	if (attr_Label_read) {
		delete[] attr_Label_read;
		attr_Label_read = 0;
	}

	if (*attr_MasterCamera_read) 
		delete *attr_MasterCamera_read;
  
	if (attr_MasterCamera_read) {
		delete[] attr_MasterCamera_read;
		attr_MasterCamera_read = 0;
	}
	

	{
	string msg_log("delete_device end");
	info_log(msg_log);
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::delete_device
	delete[] attr_FrameRate_read;
	delete[] attr_ImageBufferSize_read;
	delete[] attr_ImageAcquisitionCounter_read;
	delete[] attr_ExposureAutoEnable_read;
	delete[] attr_ExposureAutoMin_read;
	delete[] attr_ExposureAutoMax_read;
	delete[] attr_GainAutoEnable_read;
	delete[] attr_GainAutoMin_read;
	delete[] attr_GainAutoMax_read;
	delete[] attr_AutoTargetValue_read;
	delete[] attr_AutoFeedbackGain_read;
	delete[] attr_AutoFeedbackDws_read;
	delete[] attr_AutoFeedbackTargetThres_read;
	delete[] attr_AutoFeedbackDeadband_read;
	delete[] attr_MeanIntensityThreshold_read;
	delete[] attr_Binning_read;
	delete[] attr_DISEnable_read;
	delete[] attr_DisableTriggerTimeout_read;
	delete[] attr_ImageSumSamples_read;
	delete[] attr_Model_read;
	delete[] attr_FeedbackMode_read;
	delete[] attr_MeanIntensity_read;
	delete[] attr_FeedbackError_read;
	delete[] attr_EnableMedianFilter_read;
	delete[] attr_NumAverageFrames_read;
	delete[] attr_ImageSum_read;
}

//--------------------------------------------------------
/**
 *	Method      : GigeVCam::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void GigeVCam::init_device()
{
	DEBUG_STREAM << "GigeVCam::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::init_device_before) ENABLED START -----*/

	//	Initialization before get_device_property() call
	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::init_device_before
	

	//	Get the device properties from database
	get_device_property();
	
	attr_FrameRate_read = new Tango::DevDouble[1];
	attr_ImageBufferSize_read = new Tango::DevLong[1];
	attr_ImageAcquisitionCounter_read = new Tango::DevLong[1];
	attr_ExposureAutoEnable_read = new Tango::DevBoolean[1];
	attr_ExposureAutoMin_read = new Tango::DevLong[1];
	attr_ExposureAutoMax_read = new Tango::DevLong[1];
	attr_GainAutoEnable_read = new Tango::DevBoolean[1];
	attr_GainAutoMin_read = new Tango::DevDouble[1];
	attr_GainAutoMax_read = new Tango::DevDouble[1];
	attr_AutoTargetValue_read = new Tango::DevLong[1];
	attr_AutoFeedbackGain_read = new Tango::DevDouble[1];
	attr_AutoFeedbackDws_read = new Tango::DevLong[1];
	attr_AutoFeedbackTargetThres_read = new Tango::DevLong[1];
	attr_AutoFeedbackDeadband_read = new Tango::DevLong[1];
	attr_MeanIntensityThreshold_read = new Tango::DevLong[1];
	attr_Binning_read = new Tango::DevLong[1];
	attr_DISEnable_read = new Tango::DevBoolean[1];
	attr_DisableTriggerTimeout_read = new Tango::DevBoolean[1];
	attr_ImageSumSamples_read = new Tango::DevLong[1];
	attr_Model_read = new Tango::DevString[1];
	attr_FeedbackMode_read = new Tango::DevBoolean[1];
	attr_MeanIntensity_read = new Tango::DevDouble[1];
	attr_FeedbackError_read = new Tango::DevBoolean[1];
	attr_EnableMedianFilter_read = new Tango::DevBoolean[1];
	attr_NumAverageFrames_read = new Tango::DevLong[1];
	attr_ImageSum_read = new Tango::DevLong[4200*4200];
	/*----- PROTECTED REGION ID(GigeVCam::init_device) ENABLED START -----*/

	//	Initialize device
	*attr_DisableTriggerTimeout_read = false;
	Tango::Database *db = new Tango::Database();
	try {
		Tango::DbData db_data;
		db_data.push_back((Tango::DbDatum("Host")));
		db_data.push_back((Tango::DbDatum("Port")));
		db->get_property("Database",db_data);
		db_data[0] >> host_rw;
		db_data[1] >> port_rw;
	}
	catch(Tango::DevFailed &e) {
	      INFO_STREAM << __FUNCTION__ << " Error reading Database property='" << e.errors[0].desc << "'";
	      delete db;
	}
	delete db;

	attr_FrameRate_read = new Tango::DevDouble[1];
	attr_ImageBufferSize_read = new Tango::DevLong[1];
	attr_ImageAcquisitionCounter_read = new Tango::DevLong[1];
	attr_HorPos_read = new Tango::DevDouble();
	attr_VerPos_read = new Tango::DevDouble();
	attr_Intensity_read = new Tango::DevDouble();
	attr_HorSigma_read = new Tango::DevDouble();
	attr_VerSigma_read = new Tango::DevDouble();
	attr_Area_read = new Tango::DevDouble();
	attr_Saturation_read = new Tango::DevDouble();
	attr_Phase_read = new Tango::DevDouble();
	attr_Gain_read = new Tango::DevDouble();
	attr_BunchNumber_read = new Tango::DevLong();

	attr_HorPosRelative_read = new Tango::DevDouble();
	attr_VerPosRelative_read = new Tango::DevDouble();
	attr_ScaleHor_read = new Tango::DevDouble();
	attr_ScaleVer_read = new Tango::DevDouble();
	attr_OffsetHor_read = new Tango::DevDouble();
	attr_OffsetVer_read = new Tango::DevDouble();
	attr_HVCovar_read = new Tango::DevDouble();
	attr_Frequency_read = new Tango::DevDouble();
	attr_RoiThreshold_read = new Tango::DevDouble();
	attr_Rotation_read = new Tango::DevDouble();
	attr_SignHor_read = new Tango::DevDouble();
	attr_SignVer_read = new Tango::DevDouble();
	attr_SearchBackgroundParam_read = new Tango::DevDouble[2];

	attr_Exposure_read = new Tango::DevLong();
	attr_DebugThread_read = new Tango::DevLong();
	attr_FitError_read = new Tango::DevLong();
	attr_MaxVal_read = new Tango::DevLong();
	attr_CountMaxVal_read = new Tango::DevLong();
	attr_ProcessMode_read = new Tango::DevLong();
	attr_TimeStamp_read = new Tango::DevLong[2];
	attr_ErrorCounters_read = new Tango::DevLong[5];
	attr_CameraCounter_read = new Tango::DevLong();
	attr_AcquisitionCounter_read = new Tango::DevLong();
	attr_ProcessingTime_read = new Tango::DevLong();
	attr_Height_read = new Tango::DevLong();
	attr_Width_read = new Tango::DevLong();
	attr_IconSize_read = new Tango::DevLong();

	attr_PlotMode_read = new Tango::DevShort();
	attr_LowPassOrder_read = new Tango::DevShort();
	attr_ImageDepth_read = new Tango::DevShort();
	attr_PlotLevel_read = new Tango::DevShort();

	attr_Trigger_read = new Tango::DevBoolean();
	attr_AutoRoi_read = new Tango::DevBoolean();
	attr_SwRoi_read = new Tango::DevBoolean();
	attr_HwRoi_read = new Tango::DevBoolean();
	attr_Permutation_read = new Tango::DevBoolean();	
	attr_Mirror_read = new Tango::DevBoolean();
	attr_BackgroundSubtraction_read = new Tango::DevBoolean();
	attr_EnableProcessing_read = new Tango::DevBoolean();
	attr_FastProcessing_read = new Tango::DevBoolean();
	attr_Pause_read = new Tango::DevBoolean();
	attr_PlotOffsetAxis_read = new Tango::DevBoolean();
	attr_HwRoiParam_read = new Tango::DevLong[4];
	attr_EnableMedianFilter_read = new Tango::DevBoolean[1];

	*attr_Model_read = new char[GIGEVCAM_DESC];
	memset(*attr_Model_read, 0, GIGEVCAM_DESC);

	*attr_ImageSumSamples_read = 0;
	*attr_FeedbackMode_read = false;
	*attr_MeanIntensity_read = 0;
	*attr_FeedbackError_read = false;
	
	attr_Label_read = new char*;
	*attr_Label_read = new char[GIGEVCAM_DESC];

	attr_MasterCamera_read = new char*;
	*attr_MasterCamera_read = new char[GIGEVCAM_DESC];

	in_mutex = new omni_mutex();

	gettimeofday(&thread_state_time,NULL);
	gettimeofday(&last_state_time,NULL);

	set_state(Tango::OFF);
	set_status("camera not selected");
	
	/* connect by default to the first camera */
	cam_idx = 0; img_cur = 0; img_nxt = 0;
	attr_ImageAcquisitionCounter_write = 0;
	attr_Exposure_write = 0;
	attr_Gain_write = 0;
	
	auto_restart_cnt = 0;

	trigger_background = false;
	take_reference = false;
	
	for (int32_t i = 0; i < GIGEVCAM_ERR_CNT_SIZE; i++)
		error_counters[i] = 0;

	get_configuration();

#ifdef GIGEVCAM_RNM_NOCOMPILE
	if (rnmEnabled) {
		cout << "Gigecam::init_device(): before opening rnmshare" << endl;
		if (rnmshare_init(0) < 0) {
			cout << "Gigecam::init_device(): failed to open rnm" << endl;
			rnmEnabled = false;
			Tango::Except::throw_exception (
				(const char *)"Failed to conntect to rnm",
				(const char *)"Rnm error",
				(const char *)"Gigecam::init_device()");
		}
		else {
			cout << "Enable rnm bunch number acquisition" << endl;
			int32_t bunch_number;
			char check; 
			rnm_bunch_number_get_value(0, 1, &bunch_number, &check);
			cout << "Init rnmshare done in main.cpp, bunchnumber: " << bunch_number << endl;
		}		
	}
#endif

	gettimeofday(&socket_state_time, NULL);

	link_flag = false; pause_flag = false; run_flag = false;


	/* initialize buffer pointers */
	buf_phase = 0;
	buf_area = 0;
	buf_hor_pos = 0;
	buf_hor_sigma = 0;
	buf_saturation = 0;
	buf_ver_pos = 0;
	buf_ver_sigma = 0;
	buf_intensity = 0;
	buf_fit_error = 0;
	buf_acquisition_counter = 0;
	buf_max_val = 0;
	buf_gain = 0;
	buf_exposure = 0;

	/* for bunch number estimation (currently not used) */
	bunch_number_lock.old_frame_num = 0;
	bunch_number_lock.abs_frame_num = 0;
	bunch_number_lock.old_bunch_num = 0;
	bunch_number_lock.buf_idx = 0;
	for (int32_t i = 0; i < GIGEVCAM_LOCK_BUF_SIZE; i++) {
		bunch_number_lock.abs_frame_num_buf[i] = 0;
		bunch_number_lock.estim_bunch_num_buf[i] = 0;
		bunch_number_lock.real_bunch_num_buf[i] = 0;
	}	
	bunch_number_lock.consec_estim_err_counter = 0;

	bunch_number = 0;
	comm_err_consec_cnt = 0;
	trig_err_consec_cnt = 0;

	/* if only one camera supported start acquisition immediatly */
	/* in this case cam_idx is always 0 */
	circ_buffer = true;


	{ string msg_log("device started"); info_log(msg_log); }
	if ((num_cameras == 1) && (autoConnect == true)) {
		set_state(Tango::FAULT);
		run_flag = true;
		acqloop = NULL;
		acqloop = new acqthread(this);
		acqloop->start();
	}
	{ string msg_log("init_device end"); info_log(msg_log); }
	shutdown_flag = false;
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::init_device
}

//--------------------------------------------------------
/**
 *	Method      : GigeVCam::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void GigeVCam::get_device_property()
{
	/*----- PROTECTED REGION ID(GigeVCam::get_device_property_before) ENABLED START -----*/

	//	Initialize property data members
	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_device_property_before


	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("Configuration"));
	dev_prop.push_back(Tango::DbDatum("AutoRoi"));
	dev_prop.push_back(Tango::DbDatum("AutoRoiParam"));
	dev_prop.push_back(Tango::DbDatum("ImageProcessing"));
	dev_prop.push_back(Tango::DbDatum("PlotMode"));
	dev_prop.push_back(Tango::DbDatum("FastCalculation"));
	dev_prop.push_back(Tango::DbDatum("ImageBufferSize"));
	dev_prop.push_back(Tango::DbDatum("DataBufferSize"));
	dev_prop.push_back(Tango::DbDatum("SearchBackgroundParam"));
	dev_prop.push_back(Tango::DbDatum("RnmEnabled"));
	dev_prop.push_back(Tango::DbDatum("AutoConnect"));
	dev_prop.push_back(Tango::DbDatum("RnmIndex"));
	dev_prop.push_back(Tango::DbDatum("ProcessMode"));
	dev_prop.push_back(Tango::DbDatum("RoiParam"));
	dev_prop.push_back(Tango::DbDatum("SwRoi"));
	dev_prop.push_back(Tango::DbDatum("HwRoi"));
	dev_prop.push_back(Tango::DbDatum("RefPosOffset"));
	dev_prop.push_back(Tango::DbDatum("GainAutoParam"));
	dev_prop.push_back(Tango::DbDatum("ExposureAutoParam"));
	dev_prop.push_back(Tango::DbDatum("HwRoiParam"));
	dev_prop.push_back(Tango::DbDatum("AutoTargetValue"));
	dev_prop.push_back(Tango::DbDatum("AutoFeedbackGain"));
	dev_prop.push_back(Tango::DbDatum("AutoFeedbackDws"));
	dev_prop.push_back(Tango::DbDatum("AutoFeedbackDeadband"));
	dev_prop.push_back(Tango::DbDatum("AutoFeedbackTargetThres"));
	dev_prop.push_back(Tango::DbDatum("Binning"));
	dev_prop.push_back(Tango::DbDatum("AutoReconnect"));
	dev_prop.push_back(Tango::DbDatum("RoiThreshold"));
	dev_prop.push_back(Tango::DbDatum("StreamGrabberBuffer"));
	dev_prop.push_back(Tango::DbDatum("PacketSize"));
	dev_prop.push_back(Tango::DbDatum("FitNumCores"));
	dev_prop.push_back(Tango::DbDatum("InterPacketDelay"));
	dev_prop.push_back(Tango::DbDatum("ReserveBandwidth"));
	dev_prop.push_back(Tango::DbDatum("PermutationTable"));
	dev_prop.push_back(Tango::DbDatum("ImageStorageDevice"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on GigeVCamClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		GigeVCamClass	*ds_class =
			(static_cast<GigeVCamClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize Configuration from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  configuration;
		else {
			//	Try to initialize Configuration from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  configuration;
		}
		//	And try to extract Configuration value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  configuration;

		//	Try to initialize AutoRoi from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoRoi;
		else {
			//	Try to initialize AutoRoi from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoRoi;
		}
		//	And try to extract AutoRoi value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoRoi;

		//	Try to initialize AutoRoiParam from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoRoiParam;
		else {
			//	Try to initialize AutoRoiParam from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoRoiParam;
		}
		//	And try to extract AutoRoiParam value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoRoiParam;

		//	Try to initialize ImageProcessing from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  imageProcessing;
		else {
			//	Try to initialize ImageProcessing from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  imageProcessing;
		}
		//	And try to extract ImageProcessing value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  imageProcessing;

		//	Try to initialize PlotMode from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  plotMode;
		else {
			//	Try to initialize PlotMode from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  plotMode;
		}
		//	And try to extract PlotMode value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  plotMode;

		//	Try to initialize FastCalculation from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  fastCalculation;
		else {
			//	Try to initialize FastCalculation from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  fastCalculation;
		}
		//	And try to extract FastCalculation value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  fastCalculation;

		//	Try to initialize ImageBufferSize from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  imageBufferSize;
		else {
			//	Try to initialize ImageBufferSize from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  imageBufferSize;
		}
		//	And try to extract ImageBufferSize value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  imageBufferSize;

		//	Try to initialize DataBufferSize from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  dataBufferSize;
		else {
			//	Try to initialize DataBufferSize from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  dataBufferSize;
		}
		//	And try to extract DataBufferSize value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  dataBufferSize;

		//	Try to initialize SearchBackgroundParam from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  searchBackgroundParam;
		else {
			//	Try to initialize SearchBackgroundParam from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  searchBackgroundParam;
		}
		//	And try to extract SearchBackgroundParam value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  searchBackgroundParam;

		//	Try to initialize RnmEnabled from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  rnmEnabled;
		else {
			//	Try to initialize RnmEnabled from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  rnmEnabled;
		}
		//	And try to extract RnmEnabled value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  rnmEnabled;

		//	Try to initialize AutoConnect from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoConnect;
		else {
			//	Try to initialize AutoConnect from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoConnect;
		}
		//	And try to extract AutoConnect value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoConnect;

		//	Try to initialize RnmIndex from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  rnmIndex;
		else {
			//	Try to initialize RnmIndex from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  rnmIndex;
		}
		//	And try to extract RnmIndex value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  rnmIndex;

		//	Try to initialize ProcessMode from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  processMode;
		else {
			//	Try to initialize ProcessMode from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  processMode;
		}
		//	And try to extract ProcessMode value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  processMode;

		//	Try to initialize RoiParam from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  roiParam;
		else {
			//	Try to initialize RoiParam from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  roiParam;
		}
		//	And try to extract RoiParam value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  roiParam;

		//	Try to initialize SwRoi from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  swRoi;
		else {
			//	Try to initialize SwRoi from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  swRoi;
		}
		//	And try to extract SwRoi value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  swRoi;

		//	Try to initialize HwRoi from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  hwRoi;
		else {
			//	Try to initialize HwRoi from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  hwRoi;
		}
		//	And try to extract HwRoi value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  hwRoi;

		//	Try to initialize RefPosOffset from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  refPosOffset;
		else {
			//	Try to initialize RefPosOffset from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  refPosOffset;
		}
		//	And try to extract RefPosOffset value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  refPosOffset;

		//	Try to initialize GainAutoParam from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  gainAutoParam;
		else {
			//	Try to initialize GainAutoParam from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  gainAutoParam;
		}
		//	And try to extract GainAutoParam value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  gainAutoParam;

		//	Try to initialize ExposureAutoParam from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  exposureAutoParam;
		else {
			//	Try to initialize ExposureAutoParam from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  exposureAutoParam;
		}
		//	And try to extract ExposureAutoParam value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  exposureAutoParam;

		//	Try to initialize HwRoiParam from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  hwRoiParam;
		else {
			//	Try to initialize HwRoiParam from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  hwRoiParam;
		}
		//	And try to extract HwRoiParam value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  hwRoiParam;

		//	Try to initialize AutoTargetValue from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoTargetValue;
		else {
			//	Try to initialize AutoTargetValue from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoTargetValue;
		}
		//	And try to extract AutoTargetValue value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoTargetValue;

		//	Try to initialize AutoFeedbackGain from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoFeedbackGain;
		else {
			//	Try to initialize AutoFeedbackGain from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoFeedbackGain;
		}
		//	And try to extract AutoFeedbackGain value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoFeedbackGain;

		//	Try to initialize AutoFeedbackDws from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoFeedbackDws;
		else {
			//	Try to initialize AutoFeedbackDws from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoFeedbackDws;
		}
		//	And try to extract AutoFeedbackDws value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoFeedbackDws;

		//	Try to initialize AutoFeedbackDeadband from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoFeedbackDeadband;
		else {
			//	Try to initialize AutoFeedbackDeadband from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoFeedbackDeadband;
		}
		//	And try to extract AutoFeedbackDeadband value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoFeedbackDeadband;

		//	Try to initialize AutoFeedbackTargetThres from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoFeedbackTargetThres;
		else {
			//	Try to initialize AutoFeedbackTargetThres from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoFeedbackTargetThres;
		}
		//	And try to extract AutoFeedbackTargetThres value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoFeedbackTargetThres;

		//	Try to initialize Binning from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  binning;
		else {
			//	Try to initialize Binning from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  binning;
		}
		//	And try to extract Binning value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  binning;

		//	Try to initialize AutoReconnect from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoReconnect;
		else {
			//	Try to initialize AutoReconnect from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoReconnect;
		}
		//	And try to extract AutoReconnect value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoReconnect;

		//	Try to initialize RoiThreshold from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  roiThreshold;
		else {
			//	Try to initialize RoiThreshold from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  roiThreshold;
		}
		//	And try to extract RoiThreshold value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  roiThreshold;

		//	Try to initialize StreamGrabberBuffer from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  streamGrabberBuffer;
		else {
			//	Try to initialize StreamGrabberBuffer from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  streamGrabberBuffer;
		}
		//	And try to extract StreamGrabberBuffer value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  streamGrabberBuffer;

		//	Try to initialize PacketSize from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  packetSize;
		else {
			//	Try to initialize PacketSize from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  packetSize;
		}
		//	And try to extract PacketSize value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  packetSize;

		//	Try to initialize FitNumCores from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  fitNumCores;
		else {
			//	Try to initialize FitNumCores from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  fitNumCores;
		}
		//	And try to extract FitNumCores value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  fitNumCores;

		//	Try to initialize InterPacketDelay from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  interPacketDelay;
		else {
			//	Try to initialize InterPacketDelay from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  interPacketDelay;
		}
		//	And try to extract InterPacketDelay value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  interPacketDelay;

		//	Try to initialize ReserveBandwidth from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  reserveBandwidth;
		else {
			//	Try to initialize ReserveBandwidth from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  reserveBandwidth;
		}
		//	And try to extract ReserveBandwidth value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  reserveBandwidth;

		//	Try to initialize PermutationTable from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  permutationTable;
		else {
			//	Try to initialize PermutationTable from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  permutationTable;
		}
		//	And try to extract PermutationTable value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  permutationTable;

		//	Try to initialize ImageStorageDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  imageStorageDevice;
		else {
			//	Try to initialize ImageStorageDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  imageStorageDevice;
		}
		//	And try to extract ImageStorageDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  imageStorageDevice;

	}

	/*----- PROTECTED REGION ID(GigeVCam::get_device_property_after) ENABLED START -----*/

	//	Check device property data members init

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_device_property_after
}

//--------------------------------------------------------
/**
 *	Method      : GigeVCam::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void GigeVCam::always_executed_hook()
{
	DEBUG_STREAM << "GigeVCam::always_executed_hook()  " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::always_executed_hook) ENABLED START -----*/

	//	code always executed before all requests
	struct timeval curr_state_time;

	gettimeofday(&curr_state_time, NULL);

	/* to avoid stopping the server after opening a gui */
	if (abs(last_state_time.tv_sec - curr_state_time.tv_sec) > (GIGEVCAM_LOCK_TOUT*0.7))
		thread_state_time.tv_sec = curr_state_time.tv_sec;

	last_state_time.tv_sec = curr_state_time.tv_sec;

	if (((curr_state_time.tv_sec - socket_state_time.tv_sec) > GIGEVCAM_RECONN_TOUT) && (run_flag == true) && (link_flag == false) && ((get_state() == Tango::FAULT))) {
		if ((comm_err_consec_cnt < GIGEVCAM_COM_ERR_MAX_CNT) || (autoReconnect == true)) {
			comm_err_consec_cnt++;
			acqloop = new acqthread(this);
			string msg_log("Reconnecting camera, "); 
			char dbg_thread[20]; sprintf(dbg_thread,"retry(%d/%d)",(int)comm_err_consec_cnt,GIGEVCAM_COM_ERR_MAX_CNT);
			msg_log.append(dbg_thread); info_log(msg_log);
			acqloop->start();
			gettimeofday(&socket_state_time,NULL);
		}
		else {
			run_flag = false;
			comm_err_consec_cnt = 0;
			set_state(Tango::OFF);
			set_status("acquisition stopped");
			string msg_log("Maximum number of automatic reconnections reached, acquisition stopped");
			info_log(msg_log);
		}
	}

	/* automatic recovery of basler lock bug */
	if (((curr_state_time.tv_sec - thread_state_time.tv_sec) > GIGEVCAM_LOCK_TOUT) && (*attr_DisableTriggerTimeout_read == false) &&
	  (pause_flag == false) && (link_flag == true)) {
		if (shutdown_flag == false) {
			set_state(Tango::FAULT);
			set_status("Server locked, shutting down");
			string msg_log("LibBasler err: server locked, shutting down,"); 
			char dbg_thread[20]; sprintf(dbg_thread,"debug(%d)\n",*attr_DebugThread_read);
			msg_log.append(dbg_thread);
			info_log(msg_log);
			shutdown_flag = true;
		}
		if 	((curr_state_time.tv_sec - thread_state_time.tv_sec) > (GIGEVCAM_LOCK_TOUT*1.2)) {
			string msg_log("LibBasler err: server locked, server exit"); info_log(msg_log);
			exit(1);
		}
	}
	else {
		shutdown_flag = false;
	}
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : GigeVCam::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void GigeVCam::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "GigeVCam::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_attr_hardware) ENABLED START -----*/

	//	Add your own code
	//	Add your own code here

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method      : GigeVCam::write_attr_hardware()
 *	Description : Hardware writing for attributes
 */
//--------------------------------------------------------
void GigeVCam::write_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "GigeVCam::write_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::write_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute HorPos related method
 *	Description: Horizontal position of the spot int the image. (Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_HorPos(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_HorPos(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_HorPos) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_HorPos_read = img8[img_cur]->hor_pos * cam_param[cam_idx].hor_sign;
	}
	else if (img16.size() > 0) {
		*attr_HorPos_read = img16[img_cur]->hor_pos * cam_param[cam_idx].hor_sign;
	}
	else
		*attr_HorPos_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_HorPos_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_HorPos
}
//--------------------------------------------------------
/**
 *	Read attribute VerPos related method
 *	Description: Horizontal position of the spot int the image.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_VerPos(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_VerPos(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_VerPos) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_VerPos_read = img8[img_cur]->ver_pos * cam_param[cam_idx].ver_sign;
	}
	else if (img16.size() > 0) {
		*attr_VerPos_read = img16[img_cur]->ver_pos * cam_param[cam_idx].ver_sign;
	}
	else
		*attr_VerPos_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_VerPos_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_VerPos
}
//--------------------------------------------------------
/**
 *	Read attribute MaxVal related method
 *	Description: Maximum pixel value.(Measure)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_MaxVal(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_MaxVal(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_MaxVal) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_MaxVal_read = img8[img_cur]->max_value;
	}
	else if (img16.size() > 0) {
		*attr_MaxVal_read = img16[img_cur]->max_value;
	}
	else
		*attr_MaxVal_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_MaxVal_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_MaxVal
}
//--------------------------------------------------------
/**
 *	Read attribute CountMaxVal related method
 *	Description: Number of pixels equal to the maximum pixel value (255/4095).(Measure)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_CountMaxVal(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_CountMaxVal(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_CountMaxVal) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_CountMaxVal_read = img8[img_cur]->count_max_value;
	}
	else if (img16.size() > 0) {
		*attr_CountMaxVal_read = img16[img_cur]->count_max_value;
	}
	else
		*attr_CountMaxVal_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_CountMaxVal_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_CountMaxVal
}
//--------------------------------------------------------
/**
 *	Read attribute HorSigma related method
 *	Description: Horizontal standard deviation of the spot in the image.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_HorSigma(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_HorSigma(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_HorSigma) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_HorSigma_read = img8[img_cur]->hor_sigma;
	}
	else if (img16.size() > 0) {
		*attr_HorSigma_read = img16[img_cur]->hor_sigma;
	}
	else
		*attr_HorSigma_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_HorSigma_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_HorSigma
}
//--------------------------------------------------------
/**
 *	Read attribute VerSigma related method
 *	Description: Vertical standard deviation of the spot in the image.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_VerSigma(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_VerSigma(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_VerSigma) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_VerSigma_read = img8[img_cur]->ver_sigma;
	}
	else if (img16.size() > 0) {
		*attr_VerSigma_read = img16[img_cur]->ver_sigma;
	}
	else
		*attr_VerSigma_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_VerSigma_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_VerSigma
}
//--------------------------------------------------------
/**
 *	Read attribute Area related method
 *	Description: Area of the spot inside the roi.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Area(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Area(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Area) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_Area_read = img8[img_cur]->area;
	}
	else if (img16.size() > 0) {
		*attr_Area_read = img16[img_cur]->area;
	}
	else
		*attr_Area_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_Area_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Area
}
//--------------------------------------------------------
/**
 *	Read attribute HVCovar related method
 *	Description: Covariance between horizontal and vertical plane.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_HVCovar(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_HVCovar(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_HVCovar) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_HVCovar_read = img8[img_cur]->covar;
	}
	else if (img16.size() > 0) {
		*attr_HVCovar_read = img16[img_cur]->covar;
	}
	else
		*attr_HVCovar_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_HVCovar_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_HVCovar
}
//--------------------------------------------------------
/**
 *	Read attribute CameraCounter related method
 *	Description: Camera counter (0-65535).(Connection Diagnostics)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_CameraCounter(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_CameraCounter(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_CameraCounter) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_CameraCounter_read = img8[img_cur]->camera_counter;
	}
	else if (img16.size() > 0) {
		*attr_CameraCounter_read = img16[img_cur]->camera_counter;
	}
	else
		*attr_CameraCounter_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_CameraCounter_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_CameraCounter
}
//--------------------------------------------------------
/**
 *	Read attribute AcquisitionCounter related method
 *	Description: Absolute acquisition image counter.(Connection Diagnostics)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_AcquisitionCounter(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_AcquisitionCounter(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_AcquisitionCounter) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_AcquisitionCounter_read = img8[img_cur]->acquisition_counter;
	}
	else if (img16.size() > 0) {
		*attr_AcquisitionCounter_read = img16[img_cur]->acquisition_counter;
	}
	else
		*attr_AcquisitionCounter_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_AcquisitionCounter_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_AcquisitionCounter
}
//--------------------------------------------------------
/**
 *	Read attribute Height related method
 *	Description: Image height
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Height(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Height(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Height) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_Height_read = img8[img_cur]->get_height();
	}
	else if (img16.size() > 0) {
		*attr_Height_read = img16[img_cur]->get_height();
	}
	else
		*attr_Height_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_Height_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Height
}
//--------------------------------------------------------
/**
 *	Read attribute Width related method
 *	Description: Image width
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Width(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Width(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Width) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_Width_read = img8[img_cur]->get_width();
	}
	else if (img16.size() > 0) {
		*attr_Width_read = img16[img_cur]->get_width();
	}
	else
		*attr_Width_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_Width_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Width
}
//--------------------------------------------------------
/**
 *	Read attribute ProcessingTime related method
 *	Description: Processing time in us.(Processing Setup)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ProcessingTime(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ProcessingTime(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ProcessingTime) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_ProcessingTime_read = img8[img_cur]->processing_time;
	}
	else if (img16.size() > 0) {
		*attr_ProcessingTime_read = img16[img_cur]->processing_time;
	}
	else
		*attr_ProcessingTime_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_ProcessingTime_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ProcessingTime
}
//--------------------------------------------------------
/**
 *	Read attribute Frequency related method
 *	Description: Frame rate in Hz. (Camera Setup)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Frequency(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Frequency(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Frequency) ENABLED START -----*/
	*attr_Frequency_read = frequency;
	attr.set_value(attr_Frequency_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Frequency
}
//--------------------------------------------------------
/**
 *	Read attribute Trigger related method
 *	Description: Trigger mode (Camera Setup)
 *               false = internal trigger (max frame rate)
 *               true = external trigger
 *               The trigger will be on the rising edge of the pulse.
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Trigger(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Trigger(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Trigger) ENABLED START -----*/


	if ((img8.size() == 0) && (img16.size() == 0))  {
		*attr_Trigger_read = false;
	}

	*attr_Trigger_read = cam_param[cam_idx].trigger;

	attr.set_value(attr_Trigger_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Trigger
}
//--------------------------------------------------------
/**
 *	Write attribute Trigger related method
 *	Description: Trigger mode (Camera Setup)
 *               false = internal trigger (max frame rate)
 *               true = external trigger
 *               The trigger will be on the rising edge of the pulse.
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_Trigger(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_Trigger(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_Trigger) ENABLED START -----*/
	bool triggermode;
	INFO_STREAM << "Gigecam::write_Trigger(): " << w_val << endl;
	
	struct timeval t1;


	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_Trigger
}
//--------------------------------------------------------
/**
 *	Read attribute Exposure related method
 *	Description: Exposure time in us.(Camera Setup)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Exposure(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Exposure(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Exposure) ENABLED START -----*/
	if ((img8.size() == 0) && (img16.size() == 0))  {
		*attr_Exposure_read = 0;
	}

	attr.set_value(attr_Exposure_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Exposure
}
//--------------------------------------------------------
/**
 *	Write attribute Exposure related method
 *	Description: Exposure time in us.(Camera Setup)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_Exposure(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_Exposure(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_Exposure) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_Exposure(): " << w_val << endl;

	if (((img8.size() > 0) || (img16.size() > 0)) && (acqloop != NULL))  {

		if (*attr_ImageSumSamples_read) {
			in_mutex->lock();
			if (img8.size() > 0) {
				memset(attr_ImageSum_read, 0, img8[img_cur]->get_width()*img8[img_cur]->get_height()*4);
			}
			else if (img16.size() > 0) {
				memset(attr_ImageSum_read, 0, img16[img_cur]->get_width()*img16[img_cur]->get_height()*4);
			}
			in_mutex->unlock();
			imagesum_counter = 0;
		}

		acqloop->CameraSetExposure((int32_t)w_val);
		attr_Exposure_write = w_val;
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_Exposure
}
//--------------------------------------------------------
/**
 *	Read attribute Gain related method
 *	Description: Gain value in a.u.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Gain(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Gain(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Gain) ENABLED START -----*/

	if ((img8.size() == 0) && (img16.size() == 0))  {
		*attr_Gain_read = 0;
	}

	attr.set_value(attr_Gain_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Gain
}
//--------------------------------------------------------
/**
 *	Write attribute Gain related method
 *	Description: Gain value in a.u.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_Gain(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_Gain(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_Gain) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_Gain(): " << w_val << endl;

	if (((img8.size() > 0) || (img16.size() > 0)) && (acqloop != NULL))  {
		if (*attr_ImageSumSamples_read) {
			in_mutex->lock();
			if (img8.size() > 0) {
				memset(attr_ImageSum_read, 0, img8[img_cur]->get_width()*img8[img_cur]->get_height()*4);
			}
			else if (img16.size() > 0) {
				memset(attr_ImageSum_read, 0, img16[img_cur]->get_width()*img16[img_cur]->get_height()*4);
			}
			imagesum_counter = 0;
			in_mutex->unlock();
		}
		acqloop->CameraSetGain(w_val);
		attr_Gain_write = w_val;
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_Gain
}
//--------------------------------------------------------
/**
 *	Read attribute RoiThreshold related method
 *	Description: Scaling factor of the automatic roi threshold.
 *               Automatic roi threshol is the background level that is automatically found by autoroi routine.(Processing Setup)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_RoiThreshold(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_RoiThreshold(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_RoiThreshold) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_RoiThreshold_read = img8[img_cur]->get_roi_threshold();
	}
	else if (img16.size() > 0) {
		*attr_RoiThreshold_read  = img16[img_cur]->get_roi_threshold();
	}
	else
		*attr_RoiThreshold_read  = 0;
	in_mutex->unlock();

	attr.set_value(attr_RoiThreshold_read );

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_RoiThreshold
}
//--------------------------------------------------------
/**
 *	Write attribute RoiThreshold related method
 *	Description: Scaling factor of the automatic roi threshold.
 *               Automatic roi threshol is the background level that is automatically found by autoroi routine.(Processing Setup)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_RoiThreshold(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_RoiThreshold(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_RoiThreshold) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_RoiThreshold(): " << w_val << endl;

	cam_param[cam_idx].roi_threshold = w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_RoiThreshold
}
//--------------------------------------------------------
/**
 *	Read attribute AutoRoi related method
 *	Description: Enable autoroi.\nAutoroi routine find the background level and the roi region where the spot is.\nWhen swroi is enabled also, the roi is the roi defined by the user by roiparam\nattribute but the background level is automatically calculated and taken into accout\nin position and standard deviation calculations.(Processing Setup)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_AutoRoi(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_AutoRoi(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_AutoRoi) ENABLED START -----*/
	bool attr_bool;
	in_mutex->lock();
	if (img8.size() > 0) {
		attr_bool = img8[img_cur]->get_enable_auto_roi();
	}
	else if (img16.size() > 0) {
		attr_bool = img16[img_cur]->get_enable_auto_roi();
	}
	else
		attr_bool = false;
	in_mutex->unlock();

	if (attr_bool)
		*attr_AutoRoi_read = true;
	else
		*attr_AutoRoi_read = false;

	attr.set_value(attr_AutoRoi_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_AutoRoi
}
//--------------------------------------------------------
/**
 *	Write attribute AutoRoi related method
 *	Description: Enable autoroi.\nAutoroi routine find the background level and the roi region where the spot is.\nWhen swroi is enabled also, the roi is the roi defined by the user by roiparam\nattribute but the background level is automatically calculated and taken into accout\nin position and standard deviation calculations.(Processing Setup)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_AutoRoi(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_AutoRoi(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_AutoRoi) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_Autoroi(): " << w_val << endl;
	
	if (w_val) {
		cam_param[cam_idx].auto_roi = true;
	}
	else {
		cam_param[cam_idx].auto_roi = false;
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_AutoRoi
}
//--------------------------------------------------------
/**
 *	Read attribute SwRoi related method
 *	Description: Enable roi define by the user.(Display Options) 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_SwRoi(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_SwRoi(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_SwRoi) ENABLED START -----*/
	bool attr_bool;
	in_mutex->lock();
	if (img8.size() > 0) {
		attr_bool = img8[img_cur]->get_enable_manual_roi();
	}
	else if (img16.size() > 0) {
		attr_bool= img16[img_cur]->get_enable_manual_roi();
	}	
	else
		attr_bool = false;
	in_mutex->unlock();

	if (attr_bool)
		*attr_SwRoi_read = true;
	else
		*attr_SwRoi_read = false;

	attr.set_value(attr_SwRoi_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_SwRoi
}
//--------------------------------------------------------
/**
 *	Write attribute SwRoi related method
 *	Description: Enable roi define by the user.(Display Options) 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_SwRoi(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_SwRoi(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_SwRoi) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_PlotMode(): " << w_val << endl;
	
	if (w_val) {
		cam_param[cam_idx].sw_roi = true;
	}
	else {
		cam_param[cam_idx].sw_roi = false;
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_SwRoi
}
//--------------------------------------------------------
/**
 *	Read attribute HwRoi related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_HwRoi(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_HwRoi(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_HwRoi) ENABLED START -----*/

	attr.set_value(attr_HwRoi_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_HwRoi
}
//--------------------------------------------------------
/**
 *	Write attribute HwRoi related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_HwRoi(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_HwRoi(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_HwRoi) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_HwRoi(): " << w_val << endl;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_HwRoi
}
//--------------------------------------------------------
/**
 *	Read attribute PlotMode related method
 *	Description: Set plotting mode.(Display Options)\n0x0: plot nothing\n0x1:plot position axis and roi region\n0x2:plot spot region\n0x3: plot both
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_PlotMode(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_PlotMode(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_PlotMode) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_PlotMode_read = img8[img_cur]->get_plot_mode();
	}
	else if (img16.size() > 0) {
		*attr_PlotMode_read = img16[img_cur]->get_plot_mode();
	}	
	else
		*attr_PlotMode_read = cam_param[cam_idx].plot_mode;
	in_mutex->unlock();

	attr.set_value(attr_PlotMode_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_PlotMode
}
//--------------------------------------------------------
/**
 *	Write attribute PlotMode related method
 *	Description: Set plotting mode.(Display Options)\n0x0: plot nothing\n0x1:plot position axis and roi region\n0x2:plot spot region\n0x3: plot both
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_PlotMode(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_PlotMode(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevShort	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_PlotMode) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_PlotMode(): " << w_val << endl;
	
	cam_param[cam_idx].plot_mode = w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_PlotMode
}
//--------------------------------------------------------
/**
 *	Read attribute IconSize related method
 *	Description: Horizontal size of the decimated image used to find automatically the roi.\n(Shadow Parameters)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_IconSize(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_IconSize(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_IconSize) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_IconSize_read = img8[img_cur]->get_imageicon_width();
	}
	else if (img16.size() > 0) {
		*attr_IconSize_read = img16[img_cur]->get_imageicon_width();
	}
	else
		*attr_IconSize_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_IconSize_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_IconSize
}
//--------------------------------------------------------
/**
 *	Write attribute IconSize related method
 *	Description: Horizontal size of the decimated image used to find automatically the roi.\n(Shadow Parameters)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_IconSize(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_IconSize(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_IconSize) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_IconSize(): " << w_val << endl;
	
	cam_param[cam_idx].gridsize = (int32_t) w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_IconSize
}
//--------------------------------------------------------
/**
 *	Read attribute LowPassOrder related method
 *	Description: Low pass fiter order used to magnify the main spot on the camera.\n(Shadow Parameters)
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_LowPassOrder(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_LowPassOrder(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_LowPassOrder) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_LowPassOrder_read = img8[img_cur]->get_lp_order();
	}
	else if (img16.size() > 0) {
		*attr_LowPassOrder_read = img16[img_cur]->get_lp_order();
	}
	else
		*attr_LowPassOrder_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_LowPassOrder_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_LowPassOrder
}
//--------------------------------------------------------
/**
 *	Write attribute LowPassOrder related method
 *	Description: Low pass fiter order used to magnify the main spot on the camera.\n(Shadow Parameters)
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_LowPassOrder(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_LowPassOrder(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevShort	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_LowPassOrder) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_LowPassOrder(): " << w_val << endl;
	
	cam_param[cam_idx].lp_order = (int32_t)  w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_LowPassOrder
}
//--------------------------------------------------------
/**
 *	Read attribute Rotation related method
 *	Description: Image rotation in degrees. (Display Options)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Rotation(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Rotation(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Rotation) ENABLED START -----*/
	*attr_Rotation_read = cam_param[cam_idx].rotation;
	attr.set_value(attr_Rotation_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Rotation
}
//--------------------------------------------------------
/**
 *	Write attribute Rotation related method
 *	Description: Image rotation in degrees. (Display Options)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_Rotation(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_Rotation(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_Rotation) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_Rotation(): " << w_val << endl;
	
	in_mutex->lock();
	if (w_val != cam_param[cam_idx].rotation) {
		cam_param[cam_idx].rotation = w_val;
		if (*attr_Permutation_read) {
			perform_permutation(*attr_Permutation_read);
		}
		else {
			if (img8.size() > 0) {
				/* reset permutation vector */
				for (int32_t i = 0; i < img8[0]->get_height()*img8[0]->get_width(); i++) 
					cam_param[cam_idx].img_permutation[i] = i;
				if (cam_param[cam_idx].mirror)
					img8[0]->permutation_mirrorh(cam_param[cam_idx].img_permutation);
				if (cam_param[cam_idx].rotation != 0)
					img8[0]->permutation_rotation(cam_param[cam_idx].rotation,cam_param[cam_idx].img_permutation);
				for (uint32_t i = 0; i < img8.size(); i++) {
					img8[i]->reset();
				}
			}
			else if (img16.size() > 0) {
				/* reset permutation vector */
				for (int32_t i = 0; i < img16[0]->get_height()*img16[0]->get_width(); i++) 
					cam_param[cam_idx].img_permutation[i] = i;
				if (cam_param[cam_idx].mirror)
					img16[0]->permutation_mirrorh(cam_param[cam_idx].img_permutation);
				if (cam_param[cam_idx].rotation != 0)
					img16[0]->permutation_rotation(cam_param[cam_idx].rotation,cam_param[cam_idx].img_permutation);
				for (uint32_t i = 0; i < img16.size(); i++) {
					img16[i]->reset();
				}
			}
		}	
	}

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_Rotation
}
//--------------------------------------------------------
/**
 *	Read attribute Mirror related method
 *	Description: Mirror the image (true) on the horizontal plane. (Display Options)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Mirror(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Mirror(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Mirror) ENABLED START -----*/
	bool attr_bool;

	in_mutex->lock();
	if (img8.size() > 0) {
		attr_bool = img8[img_cur]->get_mirror();
	}
	else if (img16.size() > 0) {
		attr_bool = img16[img_cur]->get_mirror();
	}
	else
		attr_bool = false;
	in_mutex->unlock();

	if (attr_bool)
		*attr_Mirror_read = true;
	else
		*attr_Mirror_read = false;

	attr.set_value(attr_Mirror_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Mirror
}
//--------------------------------------------------------
/**
 *	Write attribute Mirror related method
 *	Description: Mirror the image (true) on the horizontal plane. (Display Options)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_Mirror(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_Mirror(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_Mirror) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_Mirror(): " << w_val << endl;
	
	in_mutex->lock();
	if (w_val != cam_param[cam_idx].mirror) {
		if (w_val) 
			cam_param[cam_idx].mirror = true;
		else 
			cam_param[cam_idx].mirror = false;

		if (*attr_Permutation_read) {
			perform_permutation(*attr_Permutation_read);
		}
		else {
			if (img8.size() > 0) {
				/* reset permutation vector */
				for (int32_t i = 0; i < img8[0]->get_height()*img8[0]->get_width(); i++) 
					cam_param[cam_idx].img_permutation[i] = i;

				if (cam_param[cam_idx].mirror)
					img8[0]->permutation_mirrorh(cam_param[cam_idx].img_permutation);
				if (cam_param[cam_idx].rotation != 0)
					img8[0]->permutation_rotation(cam_param[cam_idx].rotation,cam_param[cam_idx].img_permutation);

				for (uint32_t i = 0; i < img8.size(); i++) {
					img8[i]->reset();
				}

			}
			else if (img16.size() > 0) {
				/* reset permutation vector */
				for (int32_t i = 0; i < img16[0]->get_height()*img16[0]->get_width(); i++) 
					cam_param[cam_idx].img_permutation[i] = i;

				if (cam_param[cam_idx].mirror)
					img16[0]->permutation_mirrorh(cam_param[cam_idx].img_permutation);
				if (cam_param[cam_idx].rotation != 0)
					img16[0]->permutation_rotation(cam_param[cam_idx].rotation,cam_param[cam_idx].img_permutation);

				for (uint32_t i = 0; i < img16.size(); i++) {
					img16[i]->reset();
				}

			}
		}	
	}

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_Mirror
}
//--------------------------------------------------------
/**
 *	Read attribute ImageDepth related method
 *	Description: Image depth. Valide values are 8 o 16.(Camera Setup)
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ImageDepth(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ImageDepth(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ImageDepth) ENABLED START -----*/
	if ((img8.size() == 0) && (img16.size() == 0))  {
		*attr_ImageDepth_read = 0;
	}
	attr.set_value(attr_ImageDepth_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ImageDepth
}
//--------------------------------------------------------
/**
 *	Write attribute ImageDepth related method
 *	Description: Image depth. Valide values are 8 o 16.(Camera Setup)
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ImageDepth(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ImageDepth(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevShort	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ImageDepth) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_ImageDepth(): " << w_val << endl;


	if (((img8.size() > 0) || (img16.size() > 0)) && (acqloop != NULL))  {
		if ((w_val == 8) || (w_val == 16)) {
			stop();
			/* wait the loop stopping */
			int32_t cnt = 0;
			while (link_flag) {
				usleep(500000);
				if (cnt++ > 20)
					break;
			}
			/*
			if ((img8.size() == 0) && (img16.size() == 0)) {
				cam_param[cam_idx].img_depth = w_val;
				acqloop = new acqthread::acqthread(this);
				acqloop->start();
			}
			*/
			cam_param[cam_idx].img_depth = w_val;
			start((char *)cam_param[cam_idx].name.c_str());
		}
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ImageDepth
}
//--------------------------------------------------------
/**
 *	Read attribute BackgroundSubtraction related method
 *	Description: Enable (true) background subtraction. (Processing Setup)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_BackgroundSubtraction(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_BackgroundSubtraction(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_BackgroundSubtraction) ENABLED START -----*/
	bool attr_bool;

	in_mutex->lock();
	if (img8.size() > 0) {
		attr_bool = cam_param[cam_idx].background_subtraction;
	}
	else if (img16.size() > 0) {
		attr_bool = cam_param[cam_idx].background_subtraction;
	}
	else
		attr_bool = false;
	in_mutex->unlock();

	if (attr_bool)
		*attr_BackgroundSubtraction_read = true;
	else
		*attr_BackgroundSubtraction_read = false;

	attr.set_value(attr_BackgroundSubtraction_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_BackgroundSubtraction
}
//--------------------------------------------------------
/**
 *	Write attribute BackgroundSubtraction related method
 *	Description: Enable (true) background subtraction. (Processing Setup)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_BackgroundSubtraction(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_BackgroundSubtraction(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_BackgroundSubtraction) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_BackgroundSubtraction(): " << w_val << endl;

	if (w_val) {
		cam_param[cam_idx].background_subtraction = true;
	}
	else {
		cam_param[cam_idx].background_subtraction = false;
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_BackgroundSubtraction
}
//--------------------------------------------------------
/**
 *	Read attribute EnableProcessing related method
 *	Description: Enable (true) image processing.(Processing Setup)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_EnableProcessing(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_EnableProcessing(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_EnableProcessing) ENABLED START -----*/
	bool attr_bool;

	in_mutex->lock();
	if (img8.size() > 0) {
		attr_bool = img8[img_cur]->get_enable_process();
	}
	else if (img16.size() > 0) {
		attr_bool = img16[img_cur]->get_enable_process();
	}
	else
		attr_bool = 0;
	in_mutex->unlock();

	if (attr_bool)
		*attr_EnableProcessing_read = true;
	else
		*attr_EnableProcessing_read = false;

	attr.set_value(attr_EnableProcessing_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_EnableProcessing
}
//--------------------------------------------------------
/**
 *	Write attribute EnableProcessing related method
 *	Description: Enable (true) image processing.(Processing Setup)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_EnableProcessing(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_EnableProcessing(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_EnableProcessing) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_EnableProcessing(): " << w_val << endl;
	
	if (w_val) {
		cam_param[cam_idx].process_enable = true;
	}
	else {
		cam_param[cam_idx].process_enable = false;
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_EnableProcessing
}
//--------------------------------------------------------
/**
 *	Read attribute FastProcessing related method
 *	Description: Enable fast processing.\nImage processing will be done automatically on a \ndecimate (binning) image to reduce computation time. (Processing Setup)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_FastProcessing(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_FastProcessing(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_FastProcessing) ENABLED START -----*/
	bool attr_bool;

	in_mutex->lock();
	if (img8.size() > 0) {
		attr_bool = img8[img_cur]->get_enable_fast_calc();
	}
	else if (img16.size() > 0) {
		attr_bool = img16[img_cur]->get_enable_fast_calc();
	}
	else
		attr_bool = false;
	in_mutex->unlock();

	if (attr_bool)
		*attr_FastProcessing_read = true;
	else
		*attr_FastProcessing_read = false;

	attr.set_value(attr_FastProcessing_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_FastProcessing
}
//--------------------------------------------------------
/**
 *	Write attribute FastProcessing related method
 *	Description: Enable fast processing.\nImage processing will be done automatically on a \ndecimate (binning) image to reduce computation time. (Processing Setup)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_FastProcessing(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_FastProcessing(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_FastProcessing) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_FastProcessing(): " << w_val << endl;
	
	if (w_val) {
		cam_param[cam_idx].fast_calc = true;
	}
	else {
		cam_param[cam_idx].fast_calc = false;
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_FastProcessing
}
//--------------------------------------------------------
/**
 *	Read attribute Saturation related method
 *	Description: Ratio between the number of the pixels that reached the\nmaximum value and the number of the pixels of the spot.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Saturation(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Saturation(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Saturation) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_Saturation_read = img8[img_cur]->saturation;
	}
	else if (img16.size() > 0) {
		*attr_Saturation_read = img16[img_cur]->saturation;
	}
	else
		*attr_Saturation_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_Saturation_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Saturation
}
//--------------------------------------------------------
/**
 *	Read attribute PlotLevel related method
 *	Description: Color of the plots (0-255 8bit) (0-4095 16bit).(Display Options)
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_PlotLevel(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_PlotLevel(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_PlotLevel) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_PlotLevel_read = img8[img_cur]->get_plot_level();
	}
	else if (img16.size() > 0) {
		*attr_PlotLevel_read = img16[img_cur]->get_plot_level();
	}
	else
		*attr_PlotLevel_read = cam_param[cam_idx].plot_level;
	in_mutex->unlock();

	attr.set_value(attr_PlotLevel_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_PlotLevel
}
//--------------------------------------------------------
/**
 *	Write attribute PlotLevel related method
 *	Description: Color of the plots (0-255 8bit) (0-4095 16bit).(Display Options)
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_PlotLevel(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_PlotLevel(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevShort	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_PlotLevel) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_PlotLevel(): " << w_val << endl;

	cam_param[cam_idx].plot_level = (short) w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_PlotLevel
}
//--------------------------------------------------------
/**
 *	Read attribute Pause related method
 *	Description: Pause the image acquisition but don't give up the connection with the camera. (Camera Control)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Pause(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Pause(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Pause) ENABLED START -----*/
	*attr_Pause_read = pause_flag;
	attr.set_value(attr_Pause_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Pause
}
//--------------------------------------------------------
/**
 *	Write attribute Pause related method
 *	Description: Pause the image acquisition but don't give up the connection with the camera. (Camera Control)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_Pause(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_Pause(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_Pause) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_Pause(): " << w_val << endl;
	
	if (link_flag) {
		if (w_val) {
			pause_flag = true;
		}
		else {
			pause_flag = false;
		}
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_Pause
}
//--------------------------------------------------------
/**
 *	Read attribute Label related method
 *	Description: Label of the camera currently acquired. It corresponds to the first element of the row\nof the table which contains the camera initialization paramenters.(Camera Control)\n
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Label(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Label(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Label) ENABLED START -----*/
	if (link_flag) {
		strncpy(*attr_Label_read, cam_param[cam_idx].name.c_str(), GIGEVCAM_DESC);
	}
	else {
		*attr_Label_read[0] = 0;;
	}
	attr.set_value(attr_Label_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Label
}
//--------------------------------------------------------
/**
 *	Read attribute HorPosRelative related method
 *	Description: Horizontal relative position of the spot int the image. (Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_HorPosRelative(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_HorPosRelative(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_HorPosRelative) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		if (img8[img_cur]->ref_hor_pos_offset != 0)
			*attr_HorPosRelative_read = -((double) img8[img_cur]->hor_pos-img8[img_cur]->ref_hor_pos_offset + 2 * img8[0]->hor_pos_offset)*img8[img_cur]->hor_sign;
		else
			*attr_HorPosRelative_read = 0;
	}
	else if (img16.size() > 0) {		
		if (img16[img_cur]->ref_hor_pos_offset != 0)
			*attr_HorPosRelative_read = -((double) img16[img_cur]->hor_pos-img16[img_cur]->ref_hor_pos_offset +  2 * img16[0]->hor_pos_offset)*img16[img_cur]->hor_sign;
		else
			*attr_HorPosRelative_read = 0;	
	}
	else
		*attr_HorPosRelative_read = 0;
	in_mutex->unlock();
	attr.set_value(attr_HorPosRelative_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_HorPosRelative
}
//--------------------------------------------------------
/**
 *	Read attribute VerPosRelative related method
 *	Description: Vertical relative position of the spot int the image. (Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_VerPosRelative(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_VerPosRelative(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_VerPosRelative) ENABLED START -----*/
	in_mutex->lock();


	if (img8.size() > 0) {
		if (img8[img_cur]->ref_ver_pos_offset != 0)
			*attr_VerPosRelative_read = -((double) img8[img_cur]->ver_pos-img8[img_cur]->ref_ver_pos_offset) * img8[img_cur]->ver_sign;
		else
			*attr_VerPosRelative_read = 0;
	}
	else if (img16.size() > 0) {
		if (img16[img_cur]->ref_ver_pos_offset != 0)
			*attr_VerPosRelative_read = -((double) img16[img_cur]->ver_pos-img16[img_cur]->ref_ver_pos_offset) * img16[img_cur]->ver_sign;
		else
			*attr_VerPosRelative_read = 0;
	}
	else
		*attr_VerPosRelative_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_VerPosRelative_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_VerPosRelative
}
//--------------------------------------------------------
/**
 *	Read attribute Phase related method
 *	Description: Inclination in degrees of the image.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Phase(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Phase(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Phase) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_Phase_read = (double) img8[img_cur]->phase;
	}
	else if (img16.size() > 0) {
		*attr_Phase_read = (double) img16[img_cur]->phase;
	}
	else
		*attr_Phase_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_Phase_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Phase
}
//--------------------------------------------------------
/**
 *	Read attribute ScaleHor related method
 *	Description: Set the number of pixels per mm in the horizontal plane
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ScaleHor(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ScaleHor(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ScaleHor) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_ScaleHor_read = (double) img8[img_cur]->hor_calibration  / (double) cam_param[cam_idx].binning;
	}
	else if (img16.size() > 0) {
		*attr_ScaleHor_read = (double) img16[img_cur]->hor_calibration / (double) cam_param[cam_idx].binning;
	}
	else {
		*attr_ScaleHor_read = cam_param[cam_idx].hor_calibration;
	}
	in_mutex->unlock();

	attr.set_value(attr_ScaleHor_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ScaleHor
}
//--------------------------------------------------------
/**
 *	Write attribute ScaleHor related method
 *	Description: Set the number of pixels per mm in the horizontal plane
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ScaleHor(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ScaleHor(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ScaleHor) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_ScaleHor(): " << w_val << endl;

	if (cam_param[cam_idx].binning != 1) {
		Tango::Except::throw_exception(
			(const char *) "Binning must be set to 1",
			(const char *) "Failed to set horizontal scale factor",
			(const char *) "GigeVCam::write_ScaleHor()",Tango::ERR);
	}

	in_mutex->lock();

	cam_param[cam_idx].hor_calibration = w_val;
	if (img8.size() > 0) {
		for (uint32_t j = 0; j < img8.size(); j++) {
			img8[j]->hor_calibration = w_val;
		}
	}
	else if (img16.size() > 0) {
		for (uint32_t j = 0; j < img16.size(); j++) {
			img16[j]->hor_calibration = w_val;
		}
	}

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ScaleHor
}
//--------------------------------------------------------
/**
 *	Read attribute ScaleVer related method
 *	Description: Set the number of pixels per mm in the vertical plane
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ScaleVer(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ScaleVer(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ScaleVer) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_ScaleVer_read = (double) img8[img_cur]->ver_calibration / (double) cam_param[cam_idx].binning;
	}
	else if (img16.size() > 0) {
		*attr_ScaleVer_read = (double) img16[img_cur]->ver_calibration  / (double) cam_param[cam_idx].binning;
	}
	else {
		*attr_ScaleVer_read = cam_param[cam_idx].ver_calibration;
	}
	in_mutex->unlock();

	attr.set_value(attr_ScaleVer_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ScaleVer
}
//--------------------------------------------------------
/**
 *	Write attribute ScaleVer related method
 *	Description: Set the number of pixels per mm in the vertical plane
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ScaleVer(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ScaleVer(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ScaleVer) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_ScaleVer(): " << w_val << endl;

	if (cam_param[cam_idx].binning != 1) {
		Tango::Except::throw_exception(
			(const char *) "Binning must be set to 1",
			(const char *) "Failed to set vertical scale factor",
			(const char *) "GigeVCam::write_ScaleVer()",Tango::ERR);
	}

	in_mutex->lock();
	cam_param[cam_idx].ver_calibration = w_val;
	for (uint32_t j = 0; j < img8.size(); j++) {
		img8[j]->ver_calibration = w_val;
	}
	for (uint32_t j = 0; j < img16.size(); j++) {
		img16[j]->ver_calibration = w_val;
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ScaleVer
}
//--------------------------------------------------------
/**
 *	Read attribute OffsetHor related method
 *	Description: Horizontal mechanical offset in mm
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_OffsetHor(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_OffsetHor(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_OffsetHor) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_OffsetHor_read = (double) img8[img_cur]->hor_pos_offset*cam_param[cam_idx].hor_sign;
	}
	else if (img16.size() > 0) {
		*attr_OffsetHor_read = (double) img16[img_cur]->hor_pos_offset*cam_param[cam_idx].hor_sign;
	}
	else {
		*attr_OffsetHor_read = cam_param[cam_idx].hor_pos_offset;
	}

	in_mutex->unlock();

	attr.set_value(attr_OffsetHor_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_OffsetHor
}
//--------------------------------------------------------
/**
 *	Write attribute OffsetHor related method
 *	Description: Horizontal mechanical offset in mm
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_OffsetHor(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_OffsetHor(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_OffsetHor) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_OffsetHor(): " << w_val << endl;

	w_val *= cam_param[cam_idx].hor_sign;

	in_mutex->lock();
	cam_param[cam_idx].delta_ref_hor_pos_offset -= (cam_param[cam_idx].hor_pos_offset*cam_param[cam_idx].hor_sign - w_val);
	cam_param[cam_idx].hor_pos_offset = w_val*cam_param[cam_idx].hor_sign;

	if (img8.size() > 0) {
		for (uint32_t j = 0; j < img8.size(); j++) {
			if (img8[j]->ref_hor_pos_offset != 0) {
				img8[j]->ref_hor_pos_offset -= (img8[j]->hor_pos_offset - w_val);
			}
			img8[j]->hor_pos_offset = w_val;
		}
	}
	else if (img16.size() > 0) {
		for (uint32_t j = 0; j < img16.size(); j++) {
			if (img16[j]->ref_hor_pos_offset != 0) {
				img16[j]->ref_hor_pos_offset -= (img16[j]->hor_pos_offset - w_val);
			}
			img16[j]->hor_pos_offset = w_val;
		}
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_OffsetHor
}
//--------------------------------------------------------
/**
 *	Read attribute OffsetVer related method
 *	Description: Vertical mechanical offset in mm
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_OffsetVer(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_OffsetVer(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_OffsetVer) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_OffsetVer_read = (double) img8[img_cur]->ver_pos_offset;
	}
	else if (img16.size() > 0) {
		*attr_OffsetVer_read = (double) img16[img_cur]->ver_pos_offset;
	}
	else {
		*attr_OffsetVer_read = cam_param[cam_idx].ver_pos_offset;
	}
	in_mutex->unlock();

	attr.set_value(attr_OffsetVer_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_OffsetVer
}
//--------------------------------------------------------
/**
 *	Write attribute OffsetVer related method
 *	Description: Vertical mechanical offset in mm
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_OffsetVer(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_OffsetVer(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_OffsetVer) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_OffsetVer(): " << w_val << endl;

	in_mutex->lock();
	cam_param[cam_idx].delta_ref_ver_pos_offset -= (cam_param[cam_idx].ver_pos_offset - w_val);
	cam_param[cam_idx].ver_pos_offset = w_val;

	if (img8.size() > 0) {
		for (uint32_t j = 0; j < img8.size(); j++) {
			if (img8[j]->ref_ver_pos_offset != 0) {
				img8[j]->ref_ver_pos_offset +=  (img8[j]->ver_pos_offset - w_val)*img8[j]->ver_sign;
			}
			img8[j]->ver_pos_offset = w_val;	
		}
	}
	else if (img16.size() > 0) {
		for (uint32_t j = 0; j < img16.size(); j++) {
			if (img16[j]->ref_ver_pos_offset != 0) {
				img16[j]->ref_ver_pos_offset += (img16[j]->ver_pos_offset - w_val)*img16[j]->ver_sign;
			}
			img16[j]->ver_pos_offset = w_val;	
		}
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_OffsetVer
}
//--------------------------------------------------------
/**
 *	Read attribute PlotOffsetAxis related method
 *	Description: Plot reference position axis. Reference position coordinates are (OffsetHor,OffsetVer)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_PlotOffsetAxis(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_PlotOffsetAxis(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_PlotOffsetAxis) ENABLED START -----*/
	bool attr_bool;

	in_mutex->lock();
	if (img8.size() > 0) {
		attr_bool = img8[img_cur]->get_plot_offset_axis();
	}
	else if (img16.size() > 0) {
		attr_bool = img16[img_cur]->get_plot_offset_axis();
	}
	else
		attr_bool = false;
	in_mutex->unlock();

	if (attr_bool)
		*attr_PlotOffsetAxis_read = true;
	else
		*attr_PlotOffsetAxis_read = false;

	attr.set_value(attr_PlotOffsetAxis_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_PlotOffsetAxis
}
//--------------------------------------------------------
/**
 *	Write attribute PlotOffsetAxis related method
 *	Description: Plot reference position axis. Reference position coordinates are (OffsetHor,OffsetVer)
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_PlotOffsetAxis(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_PlotOffsetAxis(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_PlotOffsetAxis) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_PlotOffsetAxis(): " << w_val << endl;
	
	if (w_val) {
		cam_param[cam_idx].plot_offset_axis = true;
	}
	else {
		cam_param[cam_idx].plot_offset_axis = false;
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_PlotOffsetAxis
}
//--------------------------------------------------------
/**
 *	Read attribute Intensity related method
 *	Description: Sum of pixel values inside the roi
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Intensity(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Intensity(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Intensity) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_Intensity_read = img8[img_cur]->intensity;
	}
	else if (img16.size() > 0) {
		*attr_Intensity_read = img16[img_cur]->intensity;
	}
	else
		*attr_Intensity_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_Intensity_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Intensity
}
//--------------------------------------------------------
/**
 *	Read attribute SignHor related method
 *	Description: Horizontal versus of the axis (1,-1)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_SignHor(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_SignHor(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_SignHor) ENABLED START -----*/
	
	*attr_SignHor_read = cam_param[cam_idx].hor_sign;
	attr.set_value(attr_SignHor_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_SignHor
}
//--------------------------------------------------------
/**
 *	Read attribute SignVer related method
 *	Description: Vertical versus of the axis (1,-1)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_SignVer(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_SignVer(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_SignVer) ENABLED START -----*/

	*attr_SignVer_read = cam_param[cam_idx].ver_sign;
	attr.set_value(attr_SignVer_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_SignVer
}
//--------------------------------------------------------
/**
 *	Read attribute ProcessMode related method
 *	Description: Processing mode:\n0=raw rms\n1=gaussian fit\n2=asymmetric gaussian fit\n3=confiteor
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ProcessMode(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ProcessMode(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ProcessMode) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_ProcessMode_read = img8[img_cur]->get_algorithm();
	}
	else if (img16.size() > 0) {
		*attr_ProcessMode_read = img16[img_cur]->get_algorithm();
	}
	else
		*attr_ProcessMode_read = 0;
	in_mutex->unlock();

	attr.set_value(attr_ProcessMode_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ProcessMode
}
//--------------------------------------------------------
/**
 *	Write attribute ProcessMode related method
 *	Description: Processing mode:\n0=raw rms\n1=gaussian fit\n2=asymmetric gaussian fit\n3=confiteor
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ProcessMode(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ProcessMode(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ProcessMode) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_ProcessMode(): " << w_val << endl;

	cam_param[cam_idx].process_mode = w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ProcessMode
}
//--------------------------------------------------------
/**
 *	Read attribute DebugThread related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_DebugThread(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_DebugThread(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_DebugThread) ENABLED START -----*/
	attr.set_value(attr_DebugThread_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_DebugThread
}
//--------------------------------------------------------
/**
 *	Read attribute FrameRate related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_FrameRate(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_FrameRate(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_FrameRate) ENABLED START -----*/

	//	Set the attribute value
	attr.set_value(attr_FrameRate_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_FrameRate
}
//--------------------------------------------------------
/**
 *	Write attribute FrameRate related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_FrameRate(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_FrameRate(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_FrameRate) ENABLED START -----*/
	*attr_FrameRate_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_FrameRate
}
//--------------------------------------------------------
/**
 *	Read attribute ImageBufferSize related method
 *	Description: Set image circular buffer size
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ImageBufferSize(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ImageBufferSize(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ImageBufferSize) ENABLED START -----*/

	//	Set the attribute value
	*attr_ImageBufferSize_read = imageBufferSize;
	attr.set_value(attr_ImageBufferSize_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ImageBufferSize
}
//--------------------------------------------------------
/**
 *	Write attribute ImageBufferSize related method
 *	Description: Set image circular buffer size
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ImageBufferSize(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ImageBufferSize(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ImageBufferSize) ENABLED START -----*/
	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ImageBufferSize
}
//--------------------------------------------------------
/**
 *	Read attribute ImageAcquisitionCounter related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ImageAcquisitionCounter(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ImageAcquisitionCounter(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ImageAcquisitionCounter) ENABLED START -----*/

	//	Set the attribute value
	*attr_ImageAcquisitionCounter_read = attr_ImageAcquisitionCounter_write;
	attr.set_value(attr_ImageAcquisitionCounter_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ImageAcquisitionCounter
}
//--------------------------------------------------------
/**
 *	Write attribute ImageAcquisitionCounter related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ImageAcquisitionCounter(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ImageAcquisitionCounter(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ImageAcquisitionCounter) ENABLED START -----*/
	attr_ImageAcquisitionCounter_write = w_val;
	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ImageAcquisitionCounter
}
//--------------------------------------------------------
/**
 *	Read attribute ExposureAutoEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ExposureAutoEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ExposureAutoEnable(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ExposureAutoEnable) ENABLED START -----*/

	//	Set the attribute value
	*attr_ExposureAutoEnable_read = cam_param[cam_idx].exposure_auto;
	attr.set_value(attr_ExposureAutoEnable_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ExposureAutoEnable
}
//--------------------------------------------------------
/**
 *	Write attribute ExposureAutoEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ExposureAutoEnable(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ExposureAutoEnable(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ExposureAutoEnable) ENABLED START -----*/
	if (w_val != cam_param[cam_idx].exposure_auto) {
		if (w_val == true)  {
			/* disable gain loop */
			cam_param[cam_idx].gain_auto = false;
		}
		cam_param[cam_idx].exposure_auto_cnt = 0;
		cam_param[cam_idx].exposure_auto = w_val;
	}
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ExposureAutoEnable
}
//--------------------------------------------------------
/**
 *	Read attribute ExposureAutoMin related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ExposureAutoMin(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ExposureAutoMin(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ExposureAutoMin) ENABLED START -----*/

	//	Set the attribute value
	*attr_ExposureAutoMin_read = cam_param[cam_idx].exposure_auto_min;
	attr.set_value(attr_ExposureAutoMin_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ExposureAutoMin
}
//--------------------------------------------------------
/**
 *	Write attribute ExposureAutoMin related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ExposureAutoMin(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ExposureAutoMin(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ExposureAutoMin) ENABLED START -----*/
	if (w_val < cam_param[cam_idx].gain_auto_min) {
		Tango::Except::throw_exception (
			(const char *)"Failed to set exposure auto min",
			(const char *)"Exposure higher than exposure auto max",
			(const char *)"GigeVCam::write_ExposureAutoMin()",Tango::ERR);
	}
	cam_param[cam_idx].exposure_auto_min = w_val;
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ExposureAutoMin
}
//--------------------------------------------------------
/**
 *	Read attribute ExposureAutoMax related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ExposureAutoMax(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ExposureAutoMax(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ExposureAutoMax) ENABLED START -----*/

	//	Set the attribute value
	*attr_ExposureAutoMax_read = cam_param[cam_idx].exposure_auto_max;
	attr.set_value(attr_ExposureAutoMax_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ExposureAutoMax
}
//--------------------------------------------------------
/**
 *	Write attribute ExposureAutoMax related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ExposureAutoMax(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ExposureAutoMax(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ExposureAutoMax) ENABLED START -----*/
	if (w_val < cam_param[cam_idx].exposure_auto_min) {
		Tango::Except::throw_exception (
			(const char *)"Failed to set exposure auto max",
			(const char *)"Exposure lower than exposure auto min",
			(const char *)"GigeVCam::write_ExposureAutoMax()",Tango::ERR);
	}
	cam_param[cam_idx].exposure_auto_max = w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ExposureAutoMax
}
//--------------------------------------------------------
/**
 *	Read attribute GainAutoEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_GainAutoEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_GainAutoEnable(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_GainAutoEnable) ENABLED START -----*/

	//	Set the attribute value
	*attr_GainAutoEnable_read = cam_param[cam_idx].gain_auto;
	attr.set_value(attr_GainAutoEnable_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_GainAutoEnable
}
//--------------------------------------------------------
/**
 *	Write attribute GainAutoEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_GainAutoEnable(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_GainAutoEnable(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_GainAutoEnable) ENABLED START -----*/
	if (w_val != cam_param[cam_idx].gain_auto) {
		if (w_val == true)  {
			/* disable exposure loop */
			cam_param[cam_idx].exposure_auto = false;
		}
		cam_param[cam_idx].gain_auto_cnt = 0;
		cam_param[cam_idx].gain_auto = w_val;
	}
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_GainAutoEnable
}
//--------------------------------------------------------
/**
 *	Read attribute GainAutoMin related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_GainAutoMin(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_GainAutoMin(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_GainAutoMin) ENABLED START -----*/

	//	Set the attribute value
	*attr_GainAutoMin_read = cam_param[cam_idx].gain_auto_min;
	attr.set_value(attr_GainAutoMin_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_GainAutoMin
}
//--------------------------------------------------------
/**
 *	Write attribute GainAutoMin related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_GainAutoMin(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_GainAutoMin(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_GainAutoMin) ENABLED START -----*/
	if (w_val > cam_param[cam_idx].gain_auto_max) {
		Tango::Except::throw_exception (
			(const char *)"Failed to set gain auto min",
			(const char *)"Gain higher than gain auto max",
			(const char *)"GigeVCam::write_GainAutoMin()",Tango::ERR);
	}
	cam_param[cam_idx].gain_auto_min = w_val;
	cam_param[cam_idx].gain_auto_raw_min = 190;
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_GainAutoMin
}
//--------------------------------------------------------
/**
 *	Read attribute GainAutoMax related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_GainAutoMax(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_GainAutoMax(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_GainAutoMax) ENABLED START -----*/

	//	Set the attribute value
	*attr_GainAutoMax_read = cam_param[cam_idx].gain_auto_max;
	attr.set_value(attr_GainAutoMax_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_GainAutoMax
}
//--------------------------------------------------------
/**
 *	Write attribute GainAutoMax related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_GainAutoMax(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_GainAutoMax(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_GainAutoMax) ENABLED START -----*/
	if (w_val < cam_param[cam_idx].gain_auto_min) {
		Tango::Except::throw_exception (
			(const char *)"Failed to set gain auto max",
			(const char *)"Gain lower than gain auto min",
			(const char *)"GigeVCam::write_GainAutoMax()",Tango::ERR);
	}
	cam_param[cam_idx].gain_auto_raw_max = 1023;
	cam_param[cam_idx].gain_auto_max = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_GainAutoMax
}
//--------------------------------------------------------
/**
 *	Read attribute AutoTargetValue related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_AutoTargetValue(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_AutoTargetValue(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_AutoTargetValue) ENABLED START -----*/

	//	Set the attribute value
	*attr_AutoTargetValue_read = cam_param[cam_idx].auto_target_value;
	attr.set_value(attr_AutoTargetValue_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_AutoTargetValue
}
//--------------------------------------------------------
/**
 *	Write attribute AutoTargetValue related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_AutoTargetValue(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_AutoTargetValue(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_AutoTargetValue) ENABLED START -----*/
	if (cam_param[cam_idx].auto_feedback_target_thres > w_val) {
		Tango::Except::throw_exception (
			(const char *)"Failed to set target value",
			(const char *)"Target value lower then threshold",
			(const char *)"write_AutoFeedbackTargetThres()",Tango::ERR);		
	}
	cam_param[cam_idx].auto_target_value = w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_AutoTargetValue
}
//--------------------------------------------------------
/**
 *	Read attribute AutoFeedbackGain related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_AutoFeedbackGain(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_AutoFeedbackGain(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_AutoFeedbackGain) ENABLED START -----*/

	//	Set the attribute value
	*attr_AutoFeedbackGain_read = cam_param[cam_idx].auto_feedback_gain;
	attr.set_value(attr_AutoFeedbackGain_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_AutoFeedbackGain
}
//--------------------------------------------------------
/**
 *	Write attribute AutoFeedbackGain related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_AutoFeedbackGain(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_AutoFeedbackGain(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_AutoFeedbackGain) ENABLED START -----*/
	cam_param[cam_idx].auto_feedback_gain = w_val;
	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_AutoFeedbackGain
}
//--------------------------------------------------------
/**
 *	Read attribute AutoFeedbackDws related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_AutoFeedbackDws(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_AutoFeedbackDws(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_AutoFeedbackDws) ENABLED START -----*/

	//	Set the attribute value
	*attr_AutoFeedbackDws_read = cam_param[cam_idx].auto_feedback_dws;
	attr.set_value(attr_AutoFeedbackDws_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_AutoFeedbackDws
}
//--------------------------------------------------------
/**
 *	Write attribute AutoFeedbackDws related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_AutoFeedbackDws(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_AutoFeedbackDws(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_AutoFeedbackDws) ENABLED START -----*/
	cam_param[cam_idx].auto_feedback_dws = w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_AutoFeedbackDws
}
//--------------------------------------------------------
/**
 *	Read attribute AutoFeedbackTargetThres related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_AutoFeedbackTargetThres(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_AutoFeedbackTargetThres(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_AutoFeedbackTargetThres) ENABLED START -----*/

	//	Set the attribute value
	*attr_AutoFeedbackTargetThres_read = cam_param[cam_idx].auto_feedback_target_thres;
	attr.set_value(attr_AutoFeedbackTargetThres_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_AutoFeedbackTargetThres
}
//--------------------------------------------------------
/**
 *	Write attribute AutoFeedbackTargetThres related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_AutoFeedbackTargetThres(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_AutoFeedbackTargetThres(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_AutoFeedbackTargetThres) ENABLED START -----*/
	if (w_val > cam_param[cam_idx].auto_target_value) {
		Tango::Except::throw_exception (
			(const char *)"Failed to set threshold",
			(const char *)"Minimum threshold higher than target value",
			(const char *)"write_AutoFeedbackTargetThres()",Tango::ERR);		
	}
	cam_param[cam_idx].auto_feedback_target_thres = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_AutoFeedbackTargetThres
}
//--------------------------------------------------------
/**
 *	Read attribute AutoFeedbackDeadband related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_AutoFeedbackDeadband(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_AutoFeedbackDeadband(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_AutoFeedbackDeadband) ENABLED START -----*/
	//	Set the attribute value
	*attr_AutoFeedbackDeadband_read = cam_param[cam_idx].auto_feedback_deadband;
	attr.set_value(attr_AutoFeedbackDeadband_read);
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_AutoFeedbackDeadband
}
//--------------------------------------------------------
/**
 *	Write attribute AutoFeedbackDeadband related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_AutoFeedbackDeadband(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_AutoFeedbackDeadband(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_AutoFeedbackDeadband) ENABLED START -----*/
	cam_param[cam_idx].auto_feedback_deadband = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_AutoFeedbackDeadband
}
//--------------------------------------------------------
/**
 *	Read attribute FitError related method
 *	Description: Fit error
 *               0=no error
 *               1=hor fit err
 *               2= ver fit err
 *               3=hor/ver fit err
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_FitError(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_FitError(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_FitError) ENABLED START -----*/

	//	Set the attribute value
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_FitError_read = img8[img_cur]->fit_error;
	}
	else if (img16.size() > 0) {
		*attr_FitError_read = img16[img_cur]->fit_error;
	}
	else
		*attr_FitError_read = 0;
	in_mutex->unlock();
	attr.set_value(attr_FitError_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_FitError
}
//--------------------------------------------------------
/**
 *	Read attribute BunchNumber related method
 *	Description: Absolute acquisition image counter.(Connection Diagnostics)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_BunchNumber(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_BunchNumber(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_BunchNumber) ENABLED START -----*/

	//	Set the attribute value
	in_mutex->lock();
	if (img8.size() > 0) {
		*attr_BunchNumber_read = img8[img_cur]->acquisition_counter;
	}
	else if (img16.size() > 0) {
		*attr_BunchNumber_read = img16[img_cur]->acquisition_counter;
	}
	else
		*attr_BunchNumber_read = 0;
	in_mutex->unlock();
	attr.set_value(attr_BunchNumber_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_BunchNumber
}
//--------------------------------------------------------
/**
 *	Read attribute MeanIntensityThreshold related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_MeanIntensityThreshold(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_MeanIntensityThreshold(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_MeanIntensityThreshold) ENABLED START -----*/

	//	Set the attribute value
	attr.set_value(attr_MeanIntensityThreshold_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_MeanIntensityThreshold
}
//--------------------------------------------------------
/**
 *	Write attribute MeanIntensityThreshold related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_MeanIntensityThreshold(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_MeanIntensityThreshold(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_MeanIntensityThreshold) ENABLED START -----*/
	*attr_MeanIntensityThreshold_read = w_val; 
	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_MeanIntensityThreshold
}
//--------------------------------------------------------
/**
 *	Read attribute Binning related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Binning(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Binning(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Binning) ENABLED START -----*/

	//	Set the attribute value
	*attr_Binning_read = cam_param[cam_idx].binning;
	attr.set_value(attr_Binning_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Binning
}
//--------------------------------------------------------
/**
 *	Write attribute Binning related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_Binning(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_Binning(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_Binning) ENABLED START -----*/

	INFO_STREAM << "Gigecam::write_Binning(): " << w_val << endl;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_Binning
}
//--------------------------------------------------------
/**
 *	Read attribute DISEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_DISEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_DISEnable(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_DISEnable) ENABLED START -----*/

	//	Set the attribute value
	*attr_DISEnable_read = cam_param[cam_idx].dis_enable;
	attr.set_value(attr_DISEnable_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_DISEnable
}
//--------------------------------------------------------
/**
 *	Write attribute DISEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_DISEnable(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_DISEnable(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_DISEnable) ENABLED START -----*/

	in_mutex->lock();
	bool	attr_bool;
	if (img8.size() > 0) {
		attr_bool = img8[img_cur]->get_enable_manual_roi();
	}
	else if (img16.size() > 0) {
		attr_bool= img16[img_cur]->get_enable_manual_roi();
	}	
	else
		attr_bool = false;
	in_mutex->unlock();
	
	if (attr_bool == false) {
		Tango::Except::throw_exception (
			(const char *)"Manual roi ha to be defined",
			(const char *)"Failed to enable DIS",
			(const char *)"Gigecam::write_DISEnable()",Tango::ERR);
	}


	in_mutex->lock();
	if (w_val == true) {
		if (img8.size() > 0) {
			img8[img_cur]->get_raw_pos(&cam_param[cam_idx].ref_raw_hor_pos, &cam_param[cam_idx].ref_raw_ver_pos);
		}
		else if (img16.size() > 0) {
			img16[img_cur]->get_raw_pos(&cam_param[cam_idx].ref_raw_hor_pos, &cam_param[cam_idx].ref_raw_ver_pos);
		}
	}
	in_mutex->unlock();

	cam_param[cam_idx].dis_enable = w_val;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_DISEnable
}
//--------------------------------------------------------
/**
 *	Read attribute DisableTriggerTimeout related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_DisableTriggerTimeout(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_DisableTriggerTimeout(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_DisableTriggerTimeout) ENABLED START -----*/

	//	Set the attribute value
	attr.set_value(attr_DisableTriggerTimeout_read);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_DisableTriggerTimeout
}
//--------------------------------------------------------
/**
 *	Write attribute DisableTriggerTimeout related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_DisableTriggerTimeout(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_DisableTriggerTimeout(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_DisableTriggerTimeout) ENABLED START -----*/
	*attr_DisableTriggerTimeout_read = w_val;
	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_DisableTriggerTimeout
}
//--------------------------------------------------------
/**
 *	Read attribute ImageSumSamples related method
 *	Description: number of images summed in a ``moving`` sum image
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_ImageSumSamples(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ImageSumSamples(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ImageSumSamples) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_ImageSumSamples_read);
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ImageSumSamples
}
//--------------------------------------------------------
/**
 *	Write attribute ImageSumSamples related method
 *	Description: number of images summed in a ``moving`` sum image
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_ImageSumSamples(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_ImageSumSamples(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_ImageSumSamples) ENABLED START -----*/
	if (w_val > imageBufferSize) {
		Tango::Except::throw_exception (
			(const char *)"Value greater then ImageBufferSize property",
			(const char *)"Failed to set image sum sample value",
			(const char *)"Gigecam::write_ImageSumSamples()",Tango::ERR);
	}	
	
	in_mutex->lock();
	if (img8.size() > 0) {
		memset(attr_ImageSum_read, 0, img8[img_cur]->get_width()*img8[img_cur]->get_height()*4);
	}
	else if (img16.size() > 0) {
		memset(attr_ImageSum_read, 0, img16[img_cur]->get_width()*img16[img_cur]->get_height()*4);
	}
	in_mutex->unlock();

	imagesum_counter = 0;
	*attr_ImageSumSamples_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_ImageSumSamples
}
//--------------------------------------------------------
/**
 *	Read attribute Model related method
 *	Description: 
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Model(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Model(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Model) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Model_read);
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Model
}
//--------------------------------------------------------
/**
 *	Read attribute FeedbackMode related method
 *	Description: true: when MeanIntensity is below MeanIntensityThreshold, the camera state is turned to FAULT
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_FeedbackMode(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_FeedbackMode(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_FeedbackMode) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_FeedbackMode_read);
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_FeedbackMode
}
//--------------------------------------------------------
/**
 *	Write attribute FeedbackMode related method
 *	Description: true: when MeanIntensity is below MeanIntensityThreshold, the camera state is turned to FAULT
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_FeedbackMode(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_FeedbackMode(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_FeedbackMode) ENABLED START -----*/
	*attr_FeedbackMode_read = w_val;
	
	if (w_val == false)
		*attr_FeedbackError_read = false;

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_FeedbackMode
}
//--------------------------------------------------------
/**
 *	Read attribute MeanIntensity related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_MeanIntensity(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_MeanIntensity(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_MeanIntensity) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_MeanIntensity_read);
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_MeanIntensity
}
//--------------------------------------------------------
/**
 *	Read attribute FeedbackError related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_FeedbackError(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_FeedbackError(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_FeedbackError) ENABLED START -----*/
	//	Set the attribute value

	if (*attr_FeedbackMode_read == false)
		*attr_FeedbackError_read = false;

	attr.set_value(attr_FeedbackError_read);
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_FeedbackError
}
//--------------------------------------------------------
/**
 *	Read attribute EnableMedianFilter related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_EnableMedianFilter(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_EnableMedianFilter(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_EnableMedianFilter) ENABLED START -----*/
	//	Set the attribute value

	bool attr_bool;

	in_mutex->lock();
	if (img8.size() > 0) {
		attr_bool = img8[img_cur]->get_enable_median_filter();
	}
	else if (img16.size() > 0) {
		attr_bool = img16[img_cur]->get_enable_median_filter();
	}
	else
		attr_bool = 0;
	in_mutex->unlock();

	if (attr_bool)
		*attr_EnableMedianFilter_read = true;
	else
		*attr_EnableMedianFilter_read = false;

	attr.set_value(attr_EnableMedianFilter_read);
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_EnableMedianFilter
}
//--------------------------------------------------------
/**
 *	Write attribute EnableMedianFilter related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_EnableMedianFilter(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_EnableMedianFilter(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_EnableMedianFilter) ENABLED START -----*/

	if (w_val) {
		cam_param[cam_idx].enable_median_filter = true;
	}
	else {
		cam_param[cam_idx].enable_median_filter = false;
	}
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_EnableMedianFilter
}
//--------------------------------------------------------
/**
 *	Read attribute Permutation related method
 *	Description: Enable image permutation by looking to PermutationTable.
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_Permutation(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Permutation(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Permutation) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Permutation_read);
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Permutation
}
//--------------------------------------------------------
/**
 *	Write attribute Permutation related method
 *	Description: Enable image permutation by looking to PermutationTable.
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_Permutation(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_Permutation(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_Permutation) ENABLED START -----*/
	
	// permutation function is executed inside acqthread::update_configuration() loop
	*attr_Permutation_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_Permutation
}
//--------------------------------------------------------
/**
 *	Read attribute MasterCamera related method
 *	Description: Label of the camera currently acquired. It corresponds to the first element of the row
 *               of the table which contains the camera initialization paramenters.(Camera Control)
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_MasterCamera(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_MasterCamera(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_MasterCamera) ENABLED START -----*/
	//	Set the attribute value

	if (link_flag) {
		strncpy(*attr_MasterCamera_read, cam_param[cam_idx].ipaddress.c_str(), GIGEVCAM_DESC);
	}
	else {
		*attr_MasterCamera_read[0] = 0;
	}
	attr.set_value(attr_MasterCamera_read);
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_MasterCamera
}
//--------------------------------------------------------
/**
 *	Read attribute NumAverageFrames related method
 *	Description: Maximum number of averaged images
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::read_NumAverageFrames(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_NumAverageFrames(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_NumAverageFrames) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_NumAverageFrames_read);
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_NumAverageFrames
}
//--------------------------------------------------------
/**
 *	Write attribute NumAverageFrames related method
 *	Description: Maximum number of averaged images
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void GigeVCam::write_NumAverageFrames(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_NumAverageFrames(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_NumAverageFrames) ENABLED START -----*/
	*attr_NumAverageFrames_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_NumAverageFrames
}
//--------------------------------------------------------
/**
 *	Read attribute SearchBackgroundParam related method
 *	Description: param[0]: number of background levels
 *               param[1]: derivative percentage threshold
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
//--------------------------------------------------------
void GigeVCam::read_SearchBackgroundParam(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_SearchBackgroundParam(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_SearchBackgroundParam) ENABLED START -----*/
	
	attr_SearchBackgroundParam_read[0] = cam_param[cam_idx].search_background_param[0];
	attr_SearchBackgroundParam_read[1] = cam_param[cam_idx].search_background_param[1];	
	attr.set_value(attr_SearchBackgroundParam_read, 2, 0, false);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_SearchBackgroundParam
}
//--------------------------------------------------------
/**
 *	Write attribute SearchBackgroundParam related method
 *	Description: param[0]: number of background levels
 *               param[1]: derivative percentage threshold
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
//--------------------------------------------------------
void GigeVCam::write_SearchBackgroundParam(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_SearchBackgroundParam(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_SearchBackgroundParam) ENABLED START -----*/
	const double *back_param;
	attr.get_write_value(back_param);

	int size = attr.get_write_value_length();

	if (size != 2) {
		Tango::Except::throw_exception (
			(const char *)"Parameters must be 2",
			(const char *)"Invalid number of parameters",
			(const char *)"Gigecam::write_SearchBackgroundParam()",Tango::ERR);
	}

	if ((back_param[0] == 0) && (back_param[1] == 0)) {
		/* do nothing */
	}
	else if ((back_param[0] < 1) || (back_param[0] > IMAGEPROC_MAX_LEVELS)) {
		Tango::Except::throw_exception (
			(const char *)"Invalid number of background levels",
			(const char *)"Failed to set search background parameters",
			(const char *)"Gigecam::write_SearchBackgroundParam()",Tango::ERR);
	}

	if ((back_param[0] == 0) && (back_param[1] == 0)) {
		/* do nothing */
	}
	else if ((back_param[1] < 0.01) || (back_param[1] > 0.9)) {
		Tango::Except::throw_exception (
			(const char *)"Invalid derivative parameter",
			(const char *)"Failed to set search background parameters",
			(const char *)"Gigecam::write_SearchBackgroundParam()",Tango::ERR);
	}

	for (int32_t i = 0; i < 2; i++)
		cam_param[cam_idx].search_background_param[i] = back_param[i];

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_SearchBackgroundParam
}
//--------------------------------------------------------
/**
 *	Read attribute HorProfile related method
 *	Description: Horizontal profile of the image.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4200
 */
//--------------------------------------------------------
void GigeVCam::read_HorProfile(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_HorProfile(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_HorProfile) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		attr_HorProfile_read = img8[img_cur]->hor_profile;
		attr.set_value(attr_HorProfile_read, img8[img_cur]->get_width(), 0, false);
	}
	else if (img16.size() > 0) {
		attr_HorProfile_read = img16[img_cur]->hor_profile;
		attr.set_value(attr_HorProfile_read, img16[img_cur]->get_width(), 0, false);
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_HorProfile
}
//--------------------------------------------------------
/**
 *	Read attribute VerProfile related method
 *	Description: Vertical profile of the image.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4200
 */
//--------------------------------------------------------
void GigeVCam::read_VerProfile(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_VerProfile(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_VerProfile) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		attr_VerProfile_read = img8[img_cur]->ver_profile;
		attr.set_value(attr_VerProfile_read, img8[img_cur]->get_height(), 0, false);
	}
	else if (img16.size() > 0) {
		attr_VerProfile_read = img16[img_cur]->ver_profile;
		attr.set_value(attr_VerProfile_read, img16[img_cur]->get_height(), 0, false);
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_VerProfile
}
//--------------------------------------------------------
/**
 *	Read attribute ErrorCounters related method
 *	Description: Error counters:\ncounter[0]: acquisition error\ncounter[1]: packets lost\ncounter[2]: packets canceled\n(Connection Diagnostics)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 10
 */
//--------------------------------------------------------
void GigeVCam::read_ErrorCounters(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ErrorCounters(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ErrorCounters) ENABLED START -----*/

	for (int32_t i = 0; i < GIGEVCAM_ERR_CNT_SIZE; i++) {
		attr_ErrorCounters_read[i] = error_counters[i];
	}

	attr.set_value(attr_ErrorCounters_read, GIGEVCAM_ERR_CNT_SIZE, 0, false);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ErrorCounters
}
//--------------------------------------------------------
/**
 *	Read attribute TimeStamp related method
 *	Description: Current timestamp (unix time) of the current image
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 2
 */
//--------------------------------------------------------
void GigeVCam::read_TimeStamp(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_TimeStamp(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_TimeStamp) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		attr_TimeStamp_read[0] = img8[img_cur]->acquisition_time.tv_sec;
		attr_TimeStamp_read[1] = img8[img_cur]->acquisition_time.tv_usec;
	}
	else if (img16.size() > 0) {
		attr_TimeStamp_read[0] = img16[img_cur]->acquisition_time.tv_sec;
		attr_TimeStamp_read[1] = img16[img_cur]->acquisition_time.tv_usec;
	}
	else {
		attr_TimeStamp_read[0] = 0;
		attr_TimeStamp_read[1] = 0;
	}
	in_mutex->unlock();

	attr.set_value(attr_TimeStamp_read, 2, 0, false);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_TimeStamp
}
//--------------------------------------------------------
/**
 *	Read attribute AutoRoiParam related method
 *	Description: Coordinates of the automatic roi (x1,y1,x2,y2)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void GigeVCam::read_AutoRoiParam(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_AutoRoiParam(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_AutoRoiParam) ENABLED START -----*/
	in_mutex->lock();
	attr_AutoRoiParam_read = 0;
	if (img8.size() > 0) {
		attr_AutoRoiParam_read = (Tango::DevLong *)img8[img_cur]->get_auto_roi();
		attr.set_value(attr_AutoRoiParam_read, 4, 0, false);
	}
	else if (img16.size() > 0) {
		attr_AutoRoiParam_read = (Tango::DevLong *)img16[img_cur]->get_auto_roi();
		attr.set_value(attr_AutoRoiParam_read, 4, 0, false);
	} 
	in_mutex->unlock();	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_AutoRoiParam
}
//--------------------------------------------------------
/**
 *	Read attribute RoiParam related method
 *	Description: Coordinates of the manual roi(x1,y1,x2,y2). (Display Options)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void GigeVCam::read_RoiParam(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_RoiParam(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_RoiParam) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		attr_RoiParam_read = (Tango::DevLong *)img8[img_cur]->get_manual_roi();
		attr.set_value(attr_RoiParam_read, 4, 0, false);
	}
	else if (img16.size() > 0) {
		attr_RoiParam_read = (Tango::DevLong *)img16[img_cur]->get_manual_roi();
		attr.set_value(attr_RoiParam_read, 4, 0, false);
	}
	in_mutex->unlock();	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_RoiParam
}
//--------------------------------------------------------
/**
 *	Write attribute RoiParam related method
 *	Description: Coordinates of the manual roi(x1,y1,x2,y2). (Display Options)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void GigeVCam::write_RoiParam(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_RoiParam(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevLong	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_RoiParam) ENABLED START -----*/
	INFO_STREAM << "Gigecam::write_RoiParam(): " << endl;

#ifdef CONFIG_64
	const int32_t *roi_array;
	attr.get_write_value(roi_array);
#else
	const long *roi_array;
	attr.get_write_value(roi_array);	
#endif
	
	int size = attr.get_write_value_length();

	if (size != 4) {
		Tango::Except::throw_exception (
			(const char *)"Parameters must be 4",
			(const char *)"Invalid number of parameters",
			(const char *)"Gigecam::write_ROI()",Tango::ERR);
	}

	int32_t roi_param_p[4];
	for (int i = 0; i < 4; i++)
		roi_param_p[i] = roi_array[i];

	set_roi_param(roi_param_p);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_RoiParam
}
//--------------------------------------------------------
/**
 *	Read attribute HorFitProfile related method
 *	Description: Horizontal profile of the image.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4200
 */
//--------------------------------------------------------
void GigeVCam::read_HorFitProfile(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_HorFitProfile(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_HorFitProfile) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		attr_HorFitProfile_read = img8[img_cur]->hor_fit_profile;
		attr.set_value(attr_HorFitProfile_read, img8[img_cur]->get_width(), 0, false);
	}
	else if (img16.size() > 0) {
		attr_HorFitProfile_read = img16[img_cur]->hor_fit_profile;
		attr.set_value(attr_HorFitProfile_read, img16[img_cur]->get_width(), 0, false);
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_HorFitProfile
}
//--------------------------------------------------------
/**
 *	Read attribute VerFitProfile related method
 *	Description: Horizontal profile of the image.(Measure)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4200
 */
//--------------------------------------------------------
void GigeVCam::read_VerFitProfile(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_VerFitProfile(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_VerFitProfile) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		attr_VerFitProfile_read = img8[img_cur]->ver_fit_profile;
		attr.set_value(attr_VerFitProfile_read, img8[img_cur]->get_height(), 0, false);
	}
	else if (img16.size() > 0) {
		attr_VerFitProfile_read = img16[img_cur]->ver_fit_profile;
		attr.set_value(attr_VerFitProfile_read, img16[img_cur]->get_height(), 0, false);
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_VerFitProfile
}
//--------------------------------------------------------
/**
 *	Read attribute HwRoiParam related method
 *	Description: Coordinates of the hardware roi(x1,y1,x2,y2). (Display Options)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void GigeVCam::read_HwRoiParam(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_HwRoiParam(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_HwRoiParam) ENABLED START -----*/

	//	Set the attribute value

	if (cam_param[cam_idx].hw_roi) {
		for (int i = 0; i < 4; i++) {
			attr_HwRoiParam_read[i] = cam_param[cam_idx].hw_roi_param[i];
		}
		attr.set_value(attr_HwRoiParam_read, 4);
	}
	else {
		Tango::Except::throw_exception (
			(const char *)"Hw roi disabled",
			(const char *)"Failed to retreive hw roi coordinates",
			(const char *)"GigeVCam::read_HwRoiParam()");				
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_HwRoiParam
}
//--------------------------------------------------------
/**
 *	Write attribute HwRoiParam related method
 *	Description: Coordinates of the hardware roi(x1,y1,x2,y2). (Display Options)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void GigeVCam::write_HwRoiParam(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "GigeVCam::write_HwRoiParam(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevLong	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(GigeVCam::write_HwRoiParam) ENABLED START -----*/

	

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::write_HwRoiParam
}
//--------------------------------------------------------
/**
 *	Read attribute Image8 related method
 *	Description: Image (8bit depth) of the camera.
 *
 *	Data type:	Tango::DevUChar
 *	Attr type:	Image max = 4200 x 4200
 */
//--------------------------------------------------------
void GigeVCam::read_Image8(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Image8(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Image8) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
		attr_Image8_read = img8[img_cur]->image;
		attr.set_value(attr_Image8_read, img8[img_cur]->get_width(), img8[img_cur]->get_height(), false);
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Image8
}
//--------------------------------------------------------
/**
 *	Read attribute Image8Icon related method
 *	Description: Decimated version of the image used to find the roi region
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Image max = 800 x 800
 */
//--------------------------------------------------------
void GigeVCam::read_Image8Icon(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Image8Icon(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Image8Icon) ENABLED START -----*/
	in_mutex->lock();
	if (img8.size() > 0) {
#ifdef CONFIG_64
		attr_Image8Icon_read = img8[img_cur]->imageicon;
#else
		attr_Image8Icon_read = (long *)img8[img_cur]->imageicon;
#endif
		attr.set_value(attr_Image8Icon_read, img8[img_cur]->get_imageicon_width(),
		 img8[img_cur]->get_imageicon_height(), false);
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Image8Icon
}
//--------------------------------------------------------
/**
 *	Read attribute Image16 related method
 *	Description: Image (18bit depth) of the camera
 *
 *	Data type:	Tango::DevUShort
 *	Attr type:	Image max = 4200 x 4200
 */
//--------------------------------------------------------
void GigeVCam::read_Image16(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Image16(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Image16) ENABLED START -----*/
	in_mutex->lock();
	if (img16.size() > 0) {
		attr_Image16_read = img16[img_cur]->image;
		attr.set_value(attr_Image16_read, img16[img_cur]->get_width(), img16[img_cur]->get_height(), false);
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Image16
}
//--------------------------------------------------------
/**
 *	Read attribute Image16Icon related method
 *	Description: Decimated version of the image used to find the roi region
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Image max = 800 x 800
 */
//--------------------------------------------------------
void GigeVCam::read_Image16Icon(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Image16Icon(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Image16Icon) ENABLED START -----*/
	in_mutex->lock();
	if (img16.size() > 0) {
#ifdef CONFIG_64
		attr_Image16Icon_read = img16[img_cur]->imageicon;
#else
		attr_Image16Icon_read = (long *)img16[img_cur]->imageicon;
#endif
		attr.set_value(attr_Image16Icon_read, img16[img_cur]->get_imageicon_width(),
		 img16[img_cur]->get_imageicon_height(), false);
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Image16Icon
}
//--------------------------------------------------------
/**
 *	Read attribute Image8Counter related method
 *	Description: Image (8bit depth) of the camera.
 *
 *	Data type:	Tango::DevUChar
 *	Attr type:	Image max = 4200 x 4200
 */
//--------------------------------------------------------
void GigeVCam::read_Image8Counter(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Image8Counter(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Image8Counter) ENABLED START -----*/

	//	Set the attribute value
	in_mutex->lock();
	int32_t param_in[2], mode = BUF_GET_BUNCH_MODE, param_out[4];
	param_in[0] = attr_ImageAcquisitionCounter_write;
	param_in[1] = attr_ImageAcquisitionCounter_write;
	/* keep the same code of the get_image command inout */
	/* param_out executes only one "for" loop just for one cycle */ 
	if (get_indexes(mode, 3, param_in, param_out) == true) {
		if (img8.size() > 0) {
			for (int32_t i = param_out[0]; i < param_out[1]; i++) {
				attr_Image8Counter_read = img8[i]->image;
				attr.set_value(attr_Image8Counter_read, img8[i]->get_width(),
					img8[i]->get_height(), false);
			}
			for (int32_t i = param_out[2]; i < param_out[3]; i++) {
				attr_Image8Counter_read = img8[i]->image;
				attr.set_value(attr_Image8Counter_read, img8[i]->get_width(),
					img8[i]->get_height(), false);
				break;
			}
		}
	}
	else {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalida data indexes",
			(const char *) "GigeVCam::read_Image8Counter()",Tango::ERR);

	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Image8Counter
}
//--------------------------------------------------------
/**
 *	Read attribute Image16Counter related method
 *	Description: Image (18bit depth) of the camera
 *
 *	Data type:	Tango::DevUShort
 *	Attr type:	Image max = 4200 x 4200
 */
//--------------------------------------------------------
void GigeVCam::read_Image16Counter(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_Image16Counter(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_Image16Counter) ENABLED START -----*/

	//	Set the attribute value
	in_mutex->lock();
	int32_t param_in[2], mode = BUF_GET_BUNCH_MODE, param_out[4];
	param_in[0] = attr_ImageAcquisitionCounter_write;
	param_in[1] = attr_ImageAcquisitionCounter_write;
	/* keep the same code of the get_image command inout */
	/* param_out executes only one "for" loop just for one cycle */ 
	if (get_indexes(mode, 3, param_in, param_out) == true) {
		if (img16.size() > 0) {
			for (int32_t i = param_out[0]; i < param_out[1]; i++) {
				attr_Image16Counter_read = img16[i]->image;
				attr.set_value(attr_Image16Counter_read, img16[i]->get_width(),
					img8[i]->get_height(), false);
			}
			for (int32_t i = param_out[2]; i < param_out[3]; i++) {
				attr_Image16Counter_read = img16[i]->image;
				attr.set_value(attr_Image16Counter_read, img16[i]->get_width(),
					img16[i]->get_height(), false);
				break;
			}
		}
	}
	else {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalida data indexes",
			(const char *) "GigeVCam::read_Image8Counter()",Tango::ERR);

	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_Image16Counter
}
//--------------------------------------------------------
/**
 *	Read attribute ImageSum related method
 *	Description: Moving sum image
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Image max = 4200 x 4200
 */
//--------------------------------------------------------
void GigeVCam::read_ImageSum(Tango::Attribute &attr)
{
	DEBUG_STREAM << "GigeVCam::read_ImageSum(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(GigeVCam::read_ImageSum) ENABLED START -----*/
	//	Set the attribute value

	if (*attr_ImageSumSamples_read <= 0) {
		Tango::Except::throw_exception(
			(const char *) "Number of image samples set to 0",
			(const char *) "Failed to get sum image",
			(const char *) "GigeVCam::read_ImageSum()",Tango::ERR);		
	}

	in_mutex->lock();
	if (img8.size() > 0) {
		attr.set_value(attr_ImageSum_read, img8[img_cur]->get_width(), img8[img_cur]->get_height(), false);
	}
	else if (img16.size() > 0) {
		attr.set_value(attr_ImageSum_read, img16[img_cur]->get_width(), img16[img_cur]->get_height(), false);
	}
	in_mutex->unlock();
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::read_ImageSum
}

//--------------------------------------------------------
/**
 *	Method      : GigeVCam::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void GigeVCam::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(GigeVCam::add_dynamic_attributes) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic attributes if any
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command Start related method
 *	Description: Connect to a camera and start image acquisition. (Camera Controls)
 *
 *	@param argin camera label or ip number
 */
//--------------------------------------------------------
void GigeVCam::start(Tango::DevString argin)
{
	DEBUG_STREAM << "GigeVCam::Start()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::start) ENABLED START -----*/

	//	Add your own code

	int32_t ip_idx, name_idx, tmp_idx = 0;

	{ string msg_log("command start executing");info_log(msg_log); }

	if (num_cameras == 1) {
		tmp_idx = 1;
	}
	else {
		ip_idx = camera_ip_index[argin];
		name_idx = camera_name_index[argin];
	
		if ((ip_idx == 0) && (name_idx == 0))	{
			Tango::Except::throw_exception(
				(const char *) "Can't connect to camera",
				(const char *) "Not find camera label",
				(const char *) "GigeVCam::start()",Tango::ERR);
		}
		else if (name_idx >  0) {
			tmp_idx = name_idx;
		}
		else if (ip_idx > 0) {
			tmp_idx = ip_idx;
		}
	}

	/* wait the loop stopping */
	if (link_flag) {
		Tango::Except::throw_exception(
			(const char *) "Still acquiring data",
			(const char *) "Stop acquisition before starting it",
			(const char *) "GigeVCam::start()",Tango::ERR);
	}

	/* if no camera actually running */
	in_mutex->lock();
	if ((img8.size() == 0) && (img16.size() == 0) && (link_flag == false)) {
		cam_idx = tmp_idx-1;
		run_flag = true;
		acqloop = new acqthread(this);
		acqloop->start();
		int count = 0;
		/* wait link flag becoming true */
		while (link_flag == false) {
			usleep(50000);
			count++;
			if (count >= 20)
				break;
		}
		if (link_flag == false) {
			in_mutex->unlock();
			Tango::Except::throw_exception(
				(const char *) "Fault in starting acquisition loop",
				(const char *) "Acquisition loop starting timeout",
				(const char *) "GigeVCam::start()",Tango::ERR);
		}
	}
	in_mutex->unlock();

	{ string msg_log("command start done"); info_log(msg_log); }

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::start
}
//--------------------------------------------------------
/**
 *	Command Stop related method
 *	Description: Stop the acquisition on the camera and disconnect from it. (Camera Controls)
 *
 */
//--------------------------------------------------------
void GigeVCam::stop()
{
	DEBUG_STREAM << "GigeVCam::Stop()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::stop) ENABLED START -----*/

	//	Add your own code
	{
	string msg_log("command stop executing");
	info_log(msg_log);
	}

	if (((img8.size() > 0) || (img16.size() > 0)) && (acqloop != NULL)) {
		acqloop->CameraStop();
	}
	run_flag = false;
	if (link_flag == false) {
		set_state(Tango::OFF);
		set_status("acquisition stopped");
	}

	{
	string msg_log("command stop done");
	info_log(msg_log);
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::stop
}
//--------------------------------------------------------
/**
 *	Command Reset related method
 *	Description: Restart connection on the current selected camera. (Camera Controls)
 *
 */
//--------------------------------------------------------
void GigeVCam::reset()
{
	DEBUG_STREAM << "GigeVCam::Reset()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::reset) ENABLED START -----*/

	//	Add your own code
	//	Add your own code to control device here

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::reset
}
//--------------------------------------------------------
/**
 *	Command ListCamera related method
 *	Description: List cameras that could be connected to this device. (Camera Controls)
 *
 *	@returns camera labels
 */
//--------------------------------------------------------
Tango::DevVarStringArray *GigeVCam::list_camera()
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "GigeVCam::ListCamera()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::list_camera) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	argout = new Tango::DevVarStringArray();
	DEBUG_STREAM << "GigeVCam::list_camera(): entering... !" << endl;

	//	Add your own code to control device here
	argout->length(num_cameras);
	
	for (uint32_t i = 0; i < argout->length(); i++)
		(*argout)[i] = CORBA::string_dup(cam_param[i].name.c_str());

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::list_camera
	return argout;
}
//--------------------------------------------------------
/**
 *	Command AcquireBackground related method
 *	Description: Acquire background image. (Processing Setup)
 *
 */
//--------------------------------------------------------
void GigeVCam::acquire_background()
{
	DEBUG_STREAM << "GigeVCam::AcquireBackground()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::acquire_background) ENABLED START -----*/

	//	Add your own code
	in_mutex->lock();
	if ((img8.size() > 0) || (img16.size() > 0)) {
		trigger_background = true;
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::acquire_background
}
//--------------------------------------------------------
/**
 *	Command GetImage8 related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarCharArray *GigeVCam::get_image8(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarCharArray *argout;
	DEBUG_STREAM << "GigeVCam::GetImage8()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_image8) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_image8(): entering... !" << endl;

	//	Add your own code to control device here
	in_mutex->lock();

	argout = 0;

	int32_t data_size;
	int32_t param_out[4], i;

	if ((img16.size() > 0) || (img8.size() <=0)) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalid data format",
			(const char *) "GigeVCam::get_image8()",Tango::ERR);
	}

	/* get buffer indexes */
	if (get_indexes((*argin)[0], argin->length(), (int32_t *) (&(*argin)[1]), param_out) == true) {
		int32_t img_size = img8[img_cur]->get_height()*img8[img_cur]->get_width();
		data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]))* img_size;

		argout  = new Tango::DevVarCharArray();
		argout->length(data_size);

		if (img8.size() > 0) {
			int32_t offset = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((unsigned char *) (&(*argout)[0]) + offset, img8[i]->image, img_size * sizeof(char));
				offset += img_size;
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((unsigned char *) (&(*argout)[0]) + offset, img8[i]->image, img_size * sizeof(char));
				offset += img_size;
			}
		}

	}
	else {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalida data indexes",
			(const char *) "GigeVCam::get_image8()",Tango::ERR);
	}

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_image8
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetImage16 related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarUShortArray *GigeVCam::get_image16(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarUShortArray *argout;
	DEBUG_STREAM << "GigeVCam::GetImage16()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_image16) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_image16(): entering... !" << endl;

	//	Add your own code to control device here
	in_mutex->lock();

	argout = 0;

	int32_t data_size;
	int32_t param_out[4], i;

	if ((img8.size() > 0) || (img16.size() <=0)) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalid data format",
			(const char *) "GigeVCam::get_image16()",Tango::ERR);
	}

	/* get buffer indexes */
	if (get_indexes((*argin)[0], argin->length(), (int32_t *) (&(*argin)[1]), param_out) == true) {
		int32_t img_size = img16[img_cur]->get_height()*img16[img_cur]->get_width();
		data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]))* img_size;
		
		argout  = new Tango::DevVarUShortArray();
		argout->length(data_size);

		if (img16.size() > 0) {
			int32_t offset = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((unsigned short *) (&(*argout)[0]) + offset, img16[i]->image, img_size * sizeof(short));
				offset += img_size;
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((unsigned short *) (&(*argout)[0]) + offset, img16[i]->image, img_size * sizeof(short));
				offset += img_size;
			}
		}
	}
	else {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalida data indexes",
			(const char *) "GigeVCam::get_image16()",Tango::ERR);
	}

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_image16
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetHorPos related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_hor_pos(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetHorPos()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_hor_pos) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_hor_pos(): entering... !" << endl;

	//	Add your own code to control device here

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;
	int32_t buf_limit = 0;

	if (buf_hor_pos == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_hor_pos()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_hor_pos, buf_idx, &data_size, &unique);
	}	
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);			
	}

	argout  = new Tango::DevVarDoubleArray();

	if ((BUF_GET_LAST_MODE == (*argin)[0]) && (data_size < (*argin)[1])) {
		buf_limit = (*argin)[1]-data_size;
		argout->length((*argin)[1]);
		for (int32_t i = 0; i < buf_limit; i++)
			(*argout)[i] = 0;
	}
	else
		argout->length(data_size);
	
	read_data_calib(unique, buf_idx, &(*argout)[buf_limit], buf_hor_pos, 0, cam_param[cam_idx].hor_sign, 0);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_hor_pos
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetVerPos related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_ver_pos(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetVerPos()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_ver_pos) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_ver_pos(): entering... !" << endl;

	//	Add your own code to control device here

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;
	int32_t buf_limit = 0;

	if (buf_ver_pos == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_ver_pos()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_ver_pos, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarDoubleArray();

	if ((BUF_GET_LAST_MODE == (*argin)[0]) && (data_size < (*argin)[1])) {
		buf_limit = (*argin)[1]-data_size;
		argout->length((*argin)[1]);
		for (int32_t i = 0; i < buf_limit; i++)
			(*argout)[i] = 0;
	}
	else
		argout->length(data_size);

	read_data_calib(unique, buf_idx, &(*argout)[buf_limit], buf_ver_pos, 0, cam_param[cam_idx].ver_sign, 0);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_ver_pos
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetHorSigma related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_hor_sigma(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetHorSigma()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_hor_sigma) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_hor_sigma(): entering... !" << endl;

	//	Add your own code to control device here

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;
	int32_t buf_limit = 0;

	if (buf_hor_sigma == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_hor_sigma()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_hor_sigma, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarDoubleArray();

	if ((BUF_GET_LAST_MODE == (*argin)[0]) && (data_size < (*argin)[1])) {
		buf_limit = (*argin)[1]-data_size;
		argout->length((*argin)[1]);
		for (int32_t i = 0; i < buf_limit; i++)
			(*argout)[i] = 0;
	}
	else
		argout->length(data_size);

	read_data(unique, buf_idx, &(*argout)[buf_limit], buf_hor_sigma);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_hor_sigma
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetVerSigma related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_ver_sigma(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetVerSigma()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_ver_sigma) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_ver_sigma(): entering... !" << endl;

	//	Add your own code to control device here

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;
	int32_t buf_limit = 0;

	if (buf_ver_sigma == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_ver_sigma()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_ver_sigma, buf_idx, &data_size,&unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarDoubleArray();

	if ((BUF_GET_LAST_MODE == (*argin)[0]) && (data_size < (*argin)[1])) {
		buf_limit = (*argin)[1]-data_size;
		argout->length((*argin)[1]);
		for (int32_t i = 0; i < buf_limit; i++)
			(*argout)[i] = 0;
	}
	else
		argout->length(data_size);

	read_data(unique,buf_idx, &(*argout)[buf_limit], buf_ver_sigma);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_ver_sigma
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetArea related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_area(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetArea()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_area) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_area(): entering... !" << endl;

	//	Add your own code to control device here

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;
	int32_t buf_limit = 0;

	if (buf_area == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_area()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_area, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
	in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarDoubleArray();

	if ((BUF_GET_LAST_MODE == (*argin)[0]) && (data_size < (*argin)[1])) {
		buf_limit = (*argin)[1]-data_size;
		argout->length((*argin)[1]);
		for (int32_t i = 0; i < buf_limit; i++)
			(*argout)[i] = 0;
	}
	else
		argout->length(data_size);

	read_data(unique, buf_idx, &(*argout)[buf_limit], buf_area);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_area
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetAcquisitionCounter related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarLongArray *GigeVCam::get_acquisition_counter(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarLongArray *argout;
	DEBUG_STREAM << "GigeVCam::GetAcquisitionCounter()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_acquisition_counter) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_acquisition_counter(): entering... !" << endl;

	//	Add your own code to control device here
	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;

	if (buf_acquisition_counter == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_acquisition_counter()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_acquisition_counter, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
	in_mutex->unlock();	
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarLongArray();
	argout->length(data_size);
	read_data(unique, buf_idx, &(*argout)[0], buf_acquisition_counter);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_acquisition_counter
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetSaturation related method
 *	Description: Get saturation values from the circular buffers
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_saturation(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetSaturation()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_saturation) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_saturation(): entering... !" << endl;

	//	Add your own code to control device here	

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;
	int32_t buf_limit = 0;

	if (buf_saturation == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_saturation()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_saturation, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarDoubleArray();

	if ((BUF_GET_LAST_MODE == (*argin)[0]) && (data_size < (*argin)[1])) {
		buf_limit = (*argin)[1]-data_size;
		argout->length((*argin)[1]);
		for (int32_t i = 0; i < buf_limit; i++)
			(*argout)[i] = 0;
	}
	else
		argout->length(data_size);

	read_data(unique, buf_idx, &(*argout)[buf_limit], buf_saturation);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_saturation
	return argout;
}
//--------------------------------------------------------
/**
 *	Command SetReference related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void GigeVCam::set_reference()
{
	DEBUG_STREAM << "GigeVCam::SetReference()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::set_reference) ENABLED START -----*/

	//	Add your own code
	take_reference = true;

	cam_param[cam_idx].ref_hor_pos_offset=0;
	cam_param[cam_idx].ref_ver_pos_offset=0;

	if (img8.size() > 0) {
		cam_param[cam_idx].ref_hor_pos_offset =img8[img_cur]->hor_pos + 2 * img8[img_cur]->hor_pos_offset;
		cam_param[cam_idx].ref_ver_pos_offset = img8[img_cur]->ver_pos;
		cam_param[cam_idx].delta_ref_hor_pos_offset = 0;
		cam_param[cam_idx].delta_ref_ver_pos_offset = 0;
	}
	else if (img16.size() > 0) {
		cam_param[cam_idx].ref_hor_pos_offset = img16[img_cur]->hor_pos + 2 * img16[img_cur]->hor_pos_offset;
		cam_param[cam_idx].ref_ver_pos_offset = img16[img_cur]->ver_pos;
		cam_param[cam_idx].delta_ref_hor_pos_offset = 0;
		cam_param[cam_idx].delta_ref_ver_pos_offset = 0;
	}

	if ((cam_param[cam_idx].ref_hor_pos_offset != 0) && (cam_param[cam_idx].ref_ver_pos_offset != 0)) {
		Tango::Database *db = new Tango::Database(host_rw,port_rw);
		try {
			Tango::DbDatum config("RefPosOffset");
			Tango::DbData db_data_put;
			vector<string>	val;
			char conf_str[300];
			memset(conf_str,0,300);
			sprintf(conf_str,"%.4f",cam_param[cam_idx].ref_hor_pos_offset);
			val.push_back(conf_str);
			memset(conf_str,0,300);
			sprintf(conf_str,"%.4f",cam_param[cam_idx].ref_ver_pos_offset);
			val.push_back(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to set reference into database",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}
	}

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::set_reference
}
//--------------------------------------------------------
/**
 *	Command GetPhase related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_phase(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetPhase()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_phase) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_phase(): entering... !" << endl;

	//	Add your own code to control device here

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;
	int32_t buf_limit = 0;

	if (buf_phase == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_phase()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_phase, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarDoubleArray();

	if ((BUF_GET_LAST_MODE == (*argin)[0]) && (data_size < (*argin)[1])) {
		buf_limit = (*argin)[1]-data_size;
		argout->length((*argin)[1]);
		for (int32_t i = 0; i < buf_limit; i++)
			(*argout)[i] = 0;
	}
	else
		argout->length(data_size);
	
	in_mutex->unlock();

	read_data(unique, buf_idx, &(*argout)[buf_limit], buf_phase);

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_phase
	return argout;
}
//--------------------------------------------------------
/**
 *	Command SaveScale related method
 *	Description: Save pixel size calibration
 *
 */
//--------------------------------------------------------
void GigeVCam::save_scale()
{
	DEBUG_STREAM << "GigeVCam::SaveScale()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::save_scale) ENABLED START -----*/

	//	Add your own code
	modify_configuration(false,true); /* modify only scale */

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::save_scale
}
//--------------------------------------------------------
/**
 *	Command RestoreScale related method
 *	Description: Restore pixel calibration
 *
 */
//--------------------------------------------------------
void GigeVCam::restore_scale()
{
	DEBUG_STREAM << "GigeVCam::RestoreScale()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::restore_scale) ENABLED START -----*/

	//	Add your own code
	INFO_STREAM << "Gigecam::restore_scale" << endl;

	in_mutex->lock();
	for (uint32_t j = 0; j < img8.size(); j++) {
		img8[j]->hor_calibration = cam_param[cam_idx].hor_calibration;
		img8[j]->ver_calibration = cam_param[cam_idx].ver_calibration;
		img8[j]->hor_pos_offset = cam_param[cam_idx].hor_pos_offset;
		img8[j]->ver_pos_offset = cam_param[cam_idx].ver_pos_offset;
	}
	for (uint32_t j = 0; j < img16.size(); j++) {
		img16[j]->hor_calibration = cam_param[cam_idx].hor_calibration;
		img16[j]->ver_calibration = cam_param[cam_idx].ver_calibration;
		img16[j]->hor_pos_offset = cam_param[cam_idx].hor_pos_offset;
		img16[j]->ver_pos_offset = cam_param[cam_idx].ver_pos_offset;
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::restore_scale
}
//--------------------------------------------------------
/**
 *	Command SetMeasurePoint related method
 *	Description: Set measure point in the image
 *
 *	@param argin horizontal pixel position, vertical pixel position
 */
//--------------------------------------------------------
void GigeVCam::set_measure_point(const Tango::DevVarLongArray *argin)
{
	DEBUG_STREAM << "GigeVCam::SetMeasurePoint()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::set_measure_point) ENABLED START -----*/

	//	Add your own code
	if ((argin->length() & 0x1) || (argin->length() > IMAGEPROC_MAX_REF*2)) {
		Tango::Except::throw_exception (
			(const char *)"Invalid number of points",
			(const char *)"Failed to set meauser points",
			(const char *)"GigeVCam::set_measure_point()");
	}

	INFO_STREAM << "GigeVCam::set_measure_points()" << endl;

	in_mutex->lock();
	if (img8.size() > 0) {
		for (uint32_t i = 0; i < (argin->length() >> 1); i++) {
			for (uint32_t j = 0; j < img8.size(); j++) {
				img8[j]->set_reference((*argin)[i*2],(*argin)[i*2+1]);
			}
		}
	}

	if (img16.size() > 0) {
		for (uint32_t i = 0; i < (argin->length() >> 1); i++) {
			for (uint32_t j = 0; j < img16.size(); j++) {
				img16[j]->set_reference((*argin)[i*2],(*argin)[i*2+1]);
			}
		}
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::set_measure_point
}
//--------------------------------------------------------
/**
 *	Command ClearMeasurePoints related method
 *	Description: Clear all reference points in the image
 *
 */
//--------------------------------------------------------
void GigeVCam::clear_measure_points()
{
	DEBUG_STREAM << "GigeVCam::ClearMeasurePoints()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::clear_measure_points) ENABLED START -----*/

	//	Add your own code
	INFO_STREAM << "GigeVCam::set_measure_points()" << endl;

	in_mutex->lock();
	if (img8.size() > 0) {
		for (int32_t i = 0; i < IMAGEPROC_MAX_REF; i++) {
			for (uint32_t j = 0; j < img8.size(); j++) {
				img8[j]->set_reference(0,0);
				img8[j]->ref_hor_pos_offset = 0;
				img8[j]->ref_ver_pos_offset = 0;
			}
		}
	}

	if (img16.size() > 0) {
		for (int32_t i = 0; i < IMAGEPROC_MAX_REF; i++) {
			for (uint32_t j = 0; j < img16.size(); j++) {
				img16[j]->set_reference(0,0);
				img16[j]->ref_hor_pos_offset = 0;
				img16[j]->ref_ver_pos_offset = 0;
			}
		}
	}
	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::clear_measure_points
}
//--------------------------------------------------------
/**
 *	Command GetIntensity related method
 *	Description: Get intensity values from the circular buffers
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_intensity(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetIntensity()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_intensity) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_intensity(): entering... !" << endl;

	//	Add your own code to control device here

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;
	int32_t buf_limit = 0;

	if (buf_intensity == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_intensity()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_intensity, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarDoubleArray();

	if ((BUF_GET_LAST_MODE == (*argin)[0]) && (data_size < (*argin)[1])) {
		buf_limit = (*argin)[1]-data_size;
		argout->length((*argin)[1]);
		for (int32_t i = 0; i < buf_limit; i++)
			(*argout)[i] = 0;
	}
	else
		argout->length(data_size);

	read_data(unique, buf_idx, &(*argout)[buf_limit], buf_intensity);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_intensity
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetHorProfile related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_hor_profile(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetHorProfile()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_hor_profile) ENABLED START -----*/

	//	Add your own code

	in_mutex->lock();

	argout = 0;

	int32_t data_size, data_size_offset = 0;
	int32_t param_out[4], i;

	/* get buffer indexes */
	if (get_indexes((*argin)[0], argin->length(), (int32_t *) (&(*argin)[1]), param_out) == true) {

		if (img8.size() > 0) {
			int32_t vec_size = img8[img_cur]->get_width();
			data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0])) * vec_size;
			if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
				data_size_offset = data_size;
				data_size += ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]));
			}
			argout  = new Tango::DevVarDoubleArray();
			argout->length(data_size);
			int32_t offset = 0,  cnt = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img8[i]->hor_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img8[i]->acquisition_counter;
				}
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img8[i]->hor_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img8[i]->acquisition_counter;
				}
			}
		}
		else if (img16.size() > 0) {
			int32_t vec_size = img16[img_cur]->get_width();
			data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0])) * vec_size ;
			if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
				data_size_offset = data_size;
				data_size += ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]));
			}
			argout  = new Tango::DevVarDoubleArray();
			argout->length(data_size);
			int32_t offset = 0, cnt = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img16[i]->hor_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img16[i]->acquisition_counter;
				}
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img16[i]->hor_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img16[i]->acquisition_counter;
				}
			}
		}
		else {
			in_mutex->unlock();
			Tango::Except::throw_exception(
				(const char *) "Failed to retreive data",
				(const char *) "Data not available",
				(const char *) "GigeVCam::GetHorProfile()",Tango::ERR);
		}

	}
	else {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalida data indexes",
			(const char *) "GigeVCam::GetHorProfile()",Tango::ERR);
	}

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_hor_profile
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetVerProfile related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_ver_profile(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetVerProfile()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_ver_profile) ENABLED START -----*/

	//	Add your own code
	in_mutex->lock();

	argout = 0;

	int32_t data_size, data_size_offset = 0;
	int32_t param_out[4], i;

	/* get buffer indexes */
	if (get_indexes((*argin)[0], argin->length(), (int32_t *) (&(*argin)[1]), param_out) == true) {

		if (img8.size() > 0) {
			int32_t vec_size = img8[img_cur]->get_height();
			data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0])) * vec_size ;
			if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
				data_size_offset = data_size;
				data_size += ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]));
			}
			argout  = new Tango::DevVarDoubleArray();
			argout->length(data_size);
			int32_t offset = 0, cnt = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img8[i]->ver_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img8[i]->acquisition_counter;
				}
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img8[i]->ver_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img8[i]->acquisition_counter;
				}
			}
		}
		else if (img16.size() > 0) {
			int32_t vec_size = img16[img_cur]->get_height();
			data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0])) * vec_size ;
			argout  = new Tango::DevVarDoubleArray();
			argout->length(data_size);
			if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
				data_size_offset = data_size;
				data_size += ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]));
			}
			int32_t offset = 0, cnt = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img16[i]->ver_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img16[i]->acquisition_counter;
				}
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img16[i]->ver_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img16[i]->acquisition_counter;
				}
			}
		}
		else {
			in_mutex->unlock();
			Tango::Except::throw_exception(
				(const char *) "Failed to retreive data",
				(const char *) "Data not available",
				(const char *) "GigeVCam::GetVerProfile()",Tango::ERR);
		}

	}
	else {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalida data indexes",
			(const char *) "GigeVCam::GetVerProfile()",Tango::ERR);
	}

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_ver_profile
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetHorFitProfile related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_hor_fit_profile(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetHorFitProfile()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_hor_fit_profile) ENABLED START -----*/

	//	Add your own code
	in_mutex->lock();

	argout = 0;

	int32_t data_size, data_size_offset = 0;
	int32_t param_out[4], i;

	/* get buffer indexes */
	if (get_indexes((*argin)[0], argin->length(), (int32_t *) (&(*argin)[1]), param_out) == true) {

		if (img8.size() > 0) {
			int32_t vec_size = img8[img_cur]->get_width();
			data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0])) * vec_size ;
			if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
				data_size_offset = data_size;
				data_size += ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]));
			}
			argout  = new Tango::DevVarDoubleArray();
			argout->length(data_size);
			int32_t offset = 0, cnt = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img8[i]->hor_fit_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img8[i]->acquisition_counter;
				}
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img8[i]->hor_fit_profile, vec_size * sizeof(double));
				offset += vec_size; 
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img8[i]->acquisition_counter;
				}
			}
		}
		else if (img16.size() > 0) {
			int32_t vec_size = img16[img_cur]->get_width();
			data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0])) * vec_size ;
			if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
				data_size_offset = data_size;
				data_size += ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]));
			}
			argout  = new Tango::DevVarDoubleArray();
			argout->length(data_size);
			int32_t offset = 0, cnt = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img16[i]->hor_fit_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img16[i]->acquisition_counter;
				}
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img16[i]->hor_fit_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img16[i]->acquisition_counter;
				}
			}
		}
		else {
			in_mutex->unlock();
			Tango::Except::throw_exception(
				(const char *) "Failed to retreive data",
				(const char *) "Data not available",
				(const char *) "GigeVCam::GetHorFitProfile()",Tango::ERR);
		}

	}
	else {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalida data indexes",
			(const char *) "GigeVCam::GetHorFitProfile()",Tango::ERR);
	}

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_hor_fit_profile
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetVerFitProfile related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_ver_fit_profile(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetVerFitProfile()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_ver_fit_profile) ENABLED START -----*/

	//	Add your own code
	in_mutex->lock();

	argout = 0;

	int32_t data_size, data_size_offset = 0;
	int32_t param_out[4], i;

	/* get buffer indexes */
	if (get_indexes((*argin)[0], argin->length(), (int32_t *) (&(*argin)[1]), param_out) == true) {

		if (img8.size() > 0) {
			int32_t vec_size = img8[img_cur]->get_height();
			data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0])) * vec_size ;
			if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
				data_size_offset = data_size;
				data_size += ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]));
			}
			argout  = new Tango::DevVarDoubleArray();
			argout->length(data_size);
			int32_t offset = 0, cnt = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img8[i]->ver_fit_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img8[i]->acquisition_counter;
				}
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img8[i]->ver_fit_profile, vec_size * sizeof(double));
				offset += vec_size; 
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img8[i]->acquisition_counter;
				}
			}
		}
		else if (img16.size() > 0) {
			int32_t vec_size = img16[img_cur]->get_height();
			data_size = ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0])) * vec_size ;
			if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
				data_size_offset = data_size;
				data_size += ((param_out[3]-param_out[2]) + (param_out[1]-param_out[0]));
			}
			argout  = new Tango::DevVarDoubleArray();
			argout->length(data_size);
			int32_t offset = 0, cnt = 0;
			for (i = param_out[0]; i < param_out[1]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img16[i]->ver_fit_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img16[i]->acquisition_counter;
				}
			}
			for (i = param_out[2]; i < param_out[3]; i++) {
				memcpy((double *) (&(*argout)[0]) + offset, img16[i]->ver_fit_profile, vec_size * sizeof(double));
				offset += vec_size;
				if (((*argin)[0] == BUF_GET_LAST_BUNCH_MODE) || ((*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE)) {
					(*argout)[data_size_offset+cnt++] = img16[i]->acquisition_counter;
				}
			}
		}
		else {
			in_mutex->unlock();
			Tango::Except::throw_exception(
				(const char *) "Failed to retreive data",
				(const char *) "Data not available",
				(const char *) "GigeVCam::GetVerFitProfile()",Tango::ERR);
		}

	}
	else {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Invalida data indexes",
			(const char *) "GigeVCam::GetVerFitProfile()",Tango::ERR);
	}

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_ver_fit_profile
	return argout;
}
//--------------------------------------------------------
/**
 *	Command SaveParam related method
 *	Description: Save all ccd / processing parameters except scale
 *
 */
//--------------------------------------------------------
void GigeVCam::save_param()
{
	DEBUG_STREAM << "GigeVCam::SaveParam()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::save_param) ENABLED START -----*/

	//	Add your own code
	modify_configuration(true,false);
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::save_param
}
//--------------------------------------------------------
/**
 *	Command ClearReference related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void GigeVCam::clear_reference()
{
	DEBUG_STREAM << "GigeVCam::ClearReference()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::clear_reference) ENABLED START -----*/

	//	Add your own code

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::clear_reference
}
//--------------------------------------------------------
/**
 *	Command ExposureAutoOnce related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void GigeVCam::exposure_auto_once()
{
	DEBUG_STREAM << "GigeVCam::ExposureAutoOnce()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::exposure_auto_once) ENABLED START -----*/

	//	Add your own code
	cam_param[cam_idx].gain_auto = false;
	cam_param[cam_idx].gain_auto_once = false;	
	cam_param[cam_idx].exposure_auto_once = true;
	cam_param[cam_idx].exposure_auto_cnt = 0;
	cam_param[cam_idx].exposure_once_cnt = 0;
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::exposure_auto_once
}
//--------------------------------------------------------
/**
 *	Command GainAutoOnce related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void GigeVCam::gain_auto_once()
{
	DEBUG_STREAM << "GigeVCam::GainAutoOnce()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::gain_auto_once) ENABLED START -----*/

	//	Add your own code
	cam_param[cam_idx].exposure_auto = false;
	cam_param[cam_idx].exposure_auto_once = false;	
	cam_param[cam_idx].gain_auto_once = true;	
	cam_param[cam_idx].gain_auto_cnt = 0;
	cam_param[cam_idx].gain_once_cnt = 0;
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::gain_auto_once
}
//--------------------------------------------------------
/**
 *	Command GetFitError related method
 *	Description: 0=no error
 *               1=hor fit err
 *               2=ver fit err
 *               3=hor/ver fit er
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarLongArray *GigeVCam::get_fit_error(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarLongArray *argout;
	DEBUG_STREAM << "GigeVCam::GetFitError()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_fit_error) ENABLED START -----*/

	//	Add your own code
	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	in_mutex->lock();
	
	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;
	int32_t buf_limit = 0;

	if (buf_fit_error == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_fit_error()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_fit_error, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarLongArray();

	if ((BUF_GET_LAST_MODE == (*argin)[0]) && (data_size < (*argin)[1])) {
		buf_limit = (*argin)[1]-data_size;
		argout->length((*argin)[1]);
		for (int32_t i = 0; i < buf_limit; i++)
			(*argout)[i] = 0;
	}
	else
		argout->length(data_size);
	
	read_data(unique, buf_idx, &(*argout)[buf_limit], buf_fit_error);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_fit_error
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetBunchNumber related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarLongArray *GigeVCam::get_bunch_number(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarLongArray *argout;
	DEBUG_STREAM << "GigeVCam::GetBunchNumber()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_bunch_number) ENABLED START -----*/

	//	Add your own code
	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_acquisition_counter(): entering... !" << endl;
	//	Add your own code to control device here

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;

	if (buf_acquisition_counter == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_bunch_number()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_acquisition_counter, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();	
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarLongArray();
	argout->length(data_size);
	read_data(unique, buf_idx, &(*argout)[0], buf_acquisition_counter);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_bunch_number
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetMaxVal related method
 *	Description: 
 *
 *	@param argin mode (0,1,2), more parameters
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarLongArray *GigeVCam::get_max_val(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarLongArray *argout;
	DEBUG_STREAM << "GigeVCam::GetMaxVal()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_max_val) ENABLED START -----*/

	//	Add your own code
	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	DEBUG_STREAM << "GigeVCam::get_acquisition_counter(): entering... !" << endl;
	//	Add your own code to control device here

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;

	if (buf_max_val == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_max_val()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_max_val, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);			
	}

	argout  = new Tango::DevVarLongArray();
	argout->length(data_size);
	read_data(unique, buf_idx, &(*argout)[0], buf_max_val);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_max_val
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetGain related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *GigeVCam::get_gain(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "GigeVCam::GetGain()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_gain) ENABLED START -----*/

	//	Add your own code
	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;

	if (buf_gain == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_gain()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_gain, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);			
	}

	argout  = new Tango::DevVarDoubleArray();
	argout->length(data_size);
	read_data(unique, buf_idx, &(*argout)[0], buf_gain);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_gain
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetExposure related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarLongArray *GigeVCam::get_exposure(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarLongArray *argout;
	DEBUG_STREAM << "GigeVCam::GetExposure()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(GigeVCam::get_exposure) ENABLED START -----*/

	//	Add your own code

	in_mutex->lock();

	argout = 0;

	long data_size, buf_idx[2];
	unsigned char unique;

	if (buf_exposure == 0) {
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Buffer not initialized",
			(const char *) "GigeVCam::get_exposure()",Tango::ERR);
	}

	try {
		check_get_data_param(argin, buf_exposure, buf_idx, &data_size, &unique);
	}
	catch(Tango::DevFailed &e) {
		in_mutex->unlock();
		Tango::Except::re_throw_exception(e,
			"Read buffer error", "Read buffer error",
			__FUNCTION__, Tango::ERR);		
	}

	argout  = new Tango::DevVarLongArray();
	argout->length(data_size);
	read_data(unique, buf_idx, &(*argout)[0], buf_exposure);

	in_mutex->unlock();

	/*----- PROTECTED REGION END -----*/	//	GigeVCam::get_exposure
	return argout;
}
//--------------------------------------------------------
/**
 *	Method      : GigeVCam::add_dynamic_commands()
 *	Description : Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void GigeVCam::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(GigeVCam::add_dynamic_commands) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic commands if any
	
	/*----- PROTECTED REGION END -----*/	//	GigeVCam::add_dynamic_commands
}

/*----- PROTECTED REGION ID(GigeVCam::namespace_ending) ENABLED START -----*/

	//	Additional Methods
/**
 *	method:	GigeVCam::get_configuration
 * Get the camera parameters
 *
*/
bool GigeVCam::get_configuration(void)
{
	
	long cam_cnt = 0, conf_size;

	/* local configuration size */
	num_cameras = configuration.size();
	global_flag_prop = false;

	/* if no camera locally defined, go to global property */
	if (num_cameras == 0) {
		Tango::DbData db_data;
		try {
			Tango::Database* db = new Tango::Database();
			db_data.push_back(Tango::DbDatum("Configuration"));
			db->get_property("Gigecam",db_data);
			if (db_data[0].is_empty()) { 
				cout << "ERROR gigevcam getting properties" << endl;
				return false;
			}
			else {
				db_data[0] >> configuration;
				global_flag_prop = true;
			}
		}
		catch (CORBA::Exception &e) {
			cout << "ERROR gigevcam getting properties" << endl;
			return false;
		}
	}

	cam_param = new cam_param_t[num_cameras];
	if (cam_param == 0)
		return false;

	for (vector<string>::const_iterator it = configuration.begin(); it < configuration.end(); it++) {
		std::string token;
		char separator;
		std::istringstream iss(*it);
		// check if the property separator is a ";" instead of the default ":" 
		std::size_t found = it->find(";");
		
		if (found != std::string::npos) 
			separator = ';';
		else 
			separator = ':';
		
		conf_size = 0;
		while ( getline(iss, token, separator) ) {
			switch (conf_size) {
				case 0: cam_param[cam_cnt].name = token; 
					camera_name_index[token.c_str()] = cam_cnt+1; 
					printf("configuration: camera name = %s\n",cam_param[cam_cnt].name.c_str());break;
				case 1: cam_param[cam_cnt].ipaddress = token;
					camera_ip_index[token.c_str()] = cam_cnt+1; 
					printf("configuration: ip address = %s\n",cam_param[cam_cnt].ipaddress.c_str());break; 
				case 2: cam_param[cam_cnt].img_depth = atol(token.c_str());
					printf("configuration: image depth = %d\n", (int) cam_param[cam_cnt].img_depth);break;
				case 3: cam_param[cam_cnt].exposure = atol(token.c_str());
					printf("configuration: exposure = %d us.\n", (int) cam_param[cam_cnt].exposure);break;
				case 4: cam_param[cam_cnt].gain = atof(token.c_str());
					printf("configuration: gain = %.1f dB\n",cam_param[cam_cnt].gain);break;
				case 5: cam_param[cam_cnt].mirror = (bool) atol(token.c_str());
					printf("configuration: mirror = ");
					if (cam_param[cam_cnt].mirror) 
						printf("true\n");
					else
						printf("false\n");
					break;
				case 6: cam_param[cam_cnt].rotation = atof(token.c_str());
					printf("configuration: rotation = %.1f deg.\n",cam_param[cam_cnt].rotation);break;
				case 7: cam_param[cam_cnt].trigger = (bool) atol(token.c_str());
					printf("configuration: trigger = ");
					if (cam_param[cam_cnt].trigger) 
						printf("external\n");
					else
						printf("internal\n");
					break;
				case 8: cam_param[cam_cnt].hor_calibration = atof(token.c_str());
					printf("configuration: hor calib. = %.3f mm\n",cam_param[cam_cnt].hor_calibration);break;
				case 9: cam_param[cam_cnt].ver_calibration = atof(token.c_str());
					printf("configuration: ver calib. = %.3f mm\n",cam_param[cam_cnt].ver_calibration);break;
				case 10: cam_param[cam_cnt].hor_pos_offset = atof(token.c_str());
					printf("configuration: hor offset = %.3f mm\n",cam_param[cam_cnt].hor_pos_offset);break;
				case 11: cam_param[cam_cnt].ver_pos_offset = atof(token.c_str());
					printf("configuration: ver offset = %.3f mm\n",cam_param[cam_cnt].ver_pos_offset);break;
				case 12: cam_param[cam_cnt].hor_sign = atof(token.c_str());
					printf("configuration: hor sign = %0f\n",cam_param[cam_cnt].hor_sign);break;
				case 13: cam_param[cam_cnt].ver_sign = atof(token.c_str());
					printf("configuration: ver sign = %0f\n",cam_param[cam_cnt].ver_sign);break;
			}
			if ((++conf_size) > 13)
				break;
		}

		cam_param[cam_cnt].img_permutation = 0;

		cam_param[cam_cnt].background_subtraction = false;
		cam_param[cam_cnt].enable_median_filter = false;
		cam_param[cam_cnt].process_enable = imageProcessing;

		cam_param[cam_cnt].auto_roi = autoRoi;
		cam_param[cam_cnt].sw_roi = swRoi;

		cam_param[cam_cnt].gridsize = (long) autoRoiParam[0];
		cam_param[cam_cnt].lp_order = (long) autoRoiParam[1];

		cam_param[cam_cnt].plot_mode = plotMode[0];
		cam_param[cam_cnt].plot_level = plotMode[1];

		cam_param[cam_cnt].manual_roi_param[0] = roiParam[0];
		cam_param[cam_cnt].manual_roi_param[1] = roiParam[1];
		cam_param[cam_cnt].manual_roi_param[2] = roiParam[2];
		cam_param[cam_cnt].manual_roi_param[3] = roiParam[3];

		cam_param[cam_cnt].process_mode = processMode;

		cam_param[cam_cnt].fast_calc = fastCalculation;

		cam_param[cam_cnt].search_background_param[0] = searchBackgroundParam[0]; /* background levels */
		cam_param[cam_cnt].search_background_param[1] = searchBackgroundParam[1]; /* derivative coeff */

		cam_param[cam_cnt].roi_threshold = roiThreshold; /* derivative coeff */

		cam_param[cam_cnt].configured = false;

		cam_param[cam_cnt].ref_hor_pos_offset = 0;
		cam_param[cam_cnt].ref_ver_pos_offset = 0;
		
		cam_param[cam_cnt].delta_ref_hor_pos_offset = 0;
		cam_param[cam_cnt].delta_ref_ver_pos_offset = 0;

		cam_param[cam_cnt].plot_offset_axis = false;
		
		cam_param[cam_cnt].hw_roi = hwRoi;

		cam_param[cam_cnt].ref_hor_pos_offset = refPosOffset[0];
		cam_param[cam_cnt].ref_ver_pos_offset = refPosOffset[1];

		cam_param[cam_cnt].gain_auto = gainAutoParam[0];
		cam_param[cam_cnt].gain_auto_min = gainAutoParam[1];
		cam_param[cam_cnt].gain_auto_max = gainAutoParam[2];

		cam_param[cam_cnt].exposure_auto = exposureAutoParam[0];
		cam_param[cam_cnt].exposure_auto_min = exposureAutoParam[1];
		cam_param[cam_cnt].exposure_auto_max = exposureAutoParam[2];

		cam_param[cam_cnt].hw_roi_param[0] = hwRoiParam[0];
		cam_param[cam_cnt].hw_roi_param[1] = hwRoiParam[1];
		cam_param[cam_cnt].hw_roi_param[2] = hwRoiParam[2];
		cam_param[cam_cnt].hw_roi_param[3] = hwRoiParam[3];

		cam_param[cam_cnt].auto_target_value = autoTargetValue;

		cam_param[cam_cnt].dis_enable = false;

		cam_param[cam_cnt].frame_rate = 10;

		cam_param[cam_cnt].auto_feedback_gain = autoFeedbackGain;
		cam_param[cam_cnt].auto_feedback_dws = autoFeedbackDws;
		cam_param[cam_cnt].auto_feedback_deadband = autoFeedbackDeadband;
		cam_param[cam_cnt].auto_feedback_target_thres = autoFeedbackTargetThres;
		cam_param[cam_cnt].binning = binning;

		/* 
		if *attr_Permutation_read and cam_param[cam_cnt].permutation do not match,
		enable_permutation is executed inside the acquisition loop (acqthread::update_configuration)
		*/
		*attr_Permutation_read = (permutationTable.size() > 0);	
		cam_param[cam_cnt].permutation = false;

		//cam_param[cam_cnt].hor_pos_offset *= cam_param[cam_cnt].hor_sign;
		//cam_param[cam_cnt].ver_pos_offset *= cam_param[cam_cnt].ver_sign;		
		

		cam_cnt++;

	}

	return true;

}
//+------------------------------------------------------------------
/**
 *	method:	GigeVCam::get_indexes
 * Get indexes from the image buffer
 *
*/
bool GigeVCam::get_indexes(int32_t mode, int32_t nargin, int32_t *param_in, int32_t *param_out)
{
	int32_t i,  image_now = img_cur;

	/* get the last samples */
	if ((nargin == 2) && ((mode == BUF_GET_LAST_MODE) || (mode == BUF_GET_LAST_BUNCH_MODE)) && (param_in[0] < imageBufferSize)) {
		if (image_now >= param_in[0]) {
			param_out[0] = image_now - param_in[0] + 1;
			param_out[1] = param_out[0] + param_in[0];
			param_out[2] = param_out[3] = 0;
		}
		else {
			param_out[0] = imageBufferSize - (param_in[0] - image_now) + 1;
			param_out[1] = imageBufferSize;
			param_out[2] = 0;
			param_out[3] = image_now + 1;
		}
	}
	/* get the time between acquisition counters */
	else if ((nargin == 3) && ((mode == BUF_GET_BUNCH_MODE) || (mode == BUF_GET_BUNCH_BUNCH_MODE))) {
		if (img8.size() > 0) {
			if (param_in[1] >= (int)img8[image_now]->acquisition_counter) {
				return false;
			}
		}
		else if (img16.size() > 0) {
			if (param_in[1] >= (int)img16[image_now]->acquisition_counter) {
				return false;
			}
		}
		else {
			return false;
		}

		if ((param_in[1]-param_in[0]) > (imageBufferSize)) {
			return false;
		}

		int32_t idx_start = -1, idx_end = -1;
		param_in[1] += 1;
		if (img8.size() > 0) {
			for (i = 0; i < imageBufferSize; i++) {
				if (((int)img8[i]->acquisition_counter == param_in[0]) && (idx_start == -1))
					idx_start = i;
				if (((int)img8[i]->acquisition_counter ==  param_in[1]) && (idx_end == -1))
					idx_end = i;
				if ((idx_start != -1) && (idx_end != -1))
					break;
			}
		}
		else if (img16.size() > 0) {
			for (i = 0; i < imageBufferSize; i++) {
				if (((int)img16[i]->acquisition_counter == param_in[0]) && (idx_start == -1))
					idx_start = i;
				if (((int)img16[i]->acquisition_counter ==  param_in[1]) && (idx_end == -1))
					idx_end = i;
				if ((idx_start != -1) && (idx_end != -1))
					break;
			}
		}
		else {
			return false;
		}

		if ((idx_start == -1) || (idx_end == -1)) {
			return false;
		}
		else if (idx_start <= idx_end) {
			param_out[0] = idx_start; param_out[1] = idx_end;
			param_out[2] = 0; param_out[3] = 0;
		}
		else {
			param_out[0] = idx_start; param_out[1] = imageBufferSize;
			param_out[2] = 0; param_out[3] = idx_end;
		}
		/*
		printf("bs=%d be=%d, in[0]=%d in[1]=%d out[0]=%d out[1]=%d out[2]=%d out[3]=%d\n",
			img8[idx_start]->acquisition_counter,img8[idx_end]->acquisition_counter,
			param_in[0],param_in[1],param_out[0],param_out[1],
			param_out[2],param_out[3]);
		*/
	}
	/* get the samples between timestamps */
	/*
	else if ((nargin == 5) && (mode == BUF_GET_TIME_MODE)) {
		unsigned long long time_start,time_end, time_buf_pre, time_buf_post;
		int32_t idx_start = -1, idx_end = -1;
		time_start = (((unsigned long long)param_in[0]) << 32) + (unsigned long long)param_in[1];
		time_end =   (((unsigned long long)param_in[2]) << 32) + (unsigned long long)param_in[3];
		for (i = 0; i < imageBufferSize-1; i++) {
			time_buf_pre =  ((unsigned long long)img8[i]->acquisition_time.tv_sec << 32) + 
				(unsigned long long)img8[i]->acquisition_time.tv_usec;
			time_buf_post = ((unsigned long long)img8[i+1]->acquisition_time.tv_sec << 32) + 
				(unsigned long long)img8[i+1]->acquisition_time.tv_usec;
			if ((time_buf_pre <= time_start) && (time_buf_post >= time_start) && (idx_start == -1))
				idx_start = i;
			if ((time_buf_pre <= time_end) && (time_buf_post >= time_end) && (idx_end == -1))
				idx_end = i;
			if ((idx_start != -1) && (idx_end != -1))
				break;
		}
		if (idx_start < idx_end) {
			param_out[0] = idx_start; param_out[1] = idx_end;
			param_out[2] = 0; param_out[3] = 0;
		}
		else {
			param_out[0] = idx_start; param_out[1] = imageBufferSize;
			param_out[2] = 0; param_out[3] = idx_end;
		}
	}
	*/
	else
		return false;

	return true;

}

//+------------------------------------------------------------------
/**
 *	method:	GigeVCam::init_buffers
 *
 *	description:	method to execute "init_buffers"
 *
 * @return	
 *
 */
//+------------------------------------------------------------------
bool GigeVCam::init_buffers(GigeVCam *gige)
{
	{string msg_log("init_buffers: initializing circular buffers");info_log(msg_log);}

	in_mutex->lock();

	gige->buf_phase = 0;
	init_rt_buffer(dataBufferSize, typeDouble, &gige->buf_phase);
	if (gige->buf_phase == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_phase);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_phase);

	gige->buf_area = 0;
	init_rt_buffer(dataBufferSize, typeDouble, &gige->buf_area);
	if (gige->buf_area == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_area);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_area);

	gige->buf_hor_pos = 0;
	init_rt_buffer(dataBufferSize, typeDouble, &gige->buf_hor_pos);
	if (gige->buf_hor_pos == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_hor_pos);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_hor_pos);

	gige->buf_hor_sigma = 0;
	init_rt_buffer(dataBufferSize, typeDouble, &gige->buf_hor_sigma);
	if (gige->buf_hor_sigma == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_hor_sigma);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_hor_sigma);

	gige->buf_saturation = 0;
	init_rt_buffer(dataBufferSize, typeDouble, &gige->buf_saturation);
	if (gige->buf_saturation == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_saturation);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_saturation);

	gige->buf_ver_pos = 0;
	init_rt_buffer(dataBufferSize, typeDouble, &gige->buf_ver_pos);
	if (gige->buf_ver_pos == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_ver_pos);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_ver_pos);

	gige->buf_ver_sigma = 0;
	init_rt_buffer(dataBufferSize, typeDouble, &gige->buf_ver_sigma);
	if (gige->buf_ver_sigma == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_ver_sigma);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_ver_sigma);

	gige->buf_intensity = 0;
	init_rt_buffer(dataBufferSize, typeDouble, &gige->buf_intensity);
	if (gige->buf_intensity == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_intensity);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_intensity);

	gige->buf_fit_error = 0;
	init_rt_buffer(dataBufferSize, typeLong, &gige->buf_fit_error);
	if (gige->buf_fit_error == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_fit_error);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_fit_error);

	gige->buf_acquisition_counter = 0;
	init_rt_buffer(dataBufferSize, typeLong, &gige->buf_acquisition_counter);
	if (gige->buf_acquisition_counter == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_acquisition_counter);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_acquisition_counter);

	gige->buf_max_val = 0;
	init_rt_buffer(dataBufferSize, typeLong, &gige->buf_max_val);
	if (gige->buf_max_val == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_max_val);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_max_val);

	gige->buf_gain = 0;
	init_rt_buffer(dataBufferSize, typeDouble, &gige->buf_gain);
	if (gige->buf_gain == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_gain);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_gain);

	gige->buf_exposure = 0;
	init_rt_buffer(dataBufferSize, typeLong, &gige->buf_exposure);
	if (gige->buf_exposure == 0) {
		in_mutex->unlock();return false;
	}
	set_mode(BUF_RT_CIRCULAR, gige->buf_exposure);
	set_fill_hole(BUF_RT_NO_FILL, gige->buf_exposure);

	in_mutex->unlock();

	{string msg_log("init_buffers: circular buffers initialized");info_log(msg_log);}

	return true;

}
//+------------------------------------------------------------------
/**
 *	method:	GigeVCam::delete_buffers
 *
 *	description:	method to execute "delete_buffers"
 *
 * @return	
 *
 */
//+------------------------------------------------------------------
void GigeVCam::delete_buffers(GigeVCam *gige)
{

	in_mutex->lock();

	if (gige->buf_phase) {
		delete_rt_buffer(gige->buf_phase); gige->buf_phase = 0;
	}
	if (gige->buf_area) {
		delete_rt_buffer(gige->buf_area); gige->buf_area = 0;
	}
	if (gige->buf_hor_pos) {
		delete_rt_buffer(gige->buf_hor_pos); gige->buf_hor_pos = 0;
	}
	if (gige->buf_hor_sigma) {
		delete_rt_buffer(gige->buf_hor_sigma);  gige->buf_hor_sigma = 0;
	}
	if (gige->buf_saturation) {
		delete_rt_buffer(gige->buf_saturation); gige->buf_saturation = 0;
	}
	if (gige->buf_ver_pos) {
		delete_rt_buffer(gige->buf_ver_pos); gige->buf_ver_pos = 0;
	}
	if (gige->buf_ver_sigma) {
		delete_rt_buffer(gige->buf_ver_sigma); gige->buf_ver_sigma = 0;
	}
	if (gige->buf_intensity) {
		delete_rt_buffer(gige->buf_intensity); gige->buf_intensity = 0;
	}
	if (gige->buf_fit_error) {
		delete_rt_buffer(gige->buf_fit_error); gige->buf_fit_error = 0;	
	}
	if (gige->buf_acquisition_counter) {
		delete_rt_buffer(gige->buf_acquisition_counter); gige->buf_acquisition_counter = 0;
	}
	if (gige->buf_max_val) {
		delete_rt_buffer(gige->buf_max_val); gige->buf_max_val = 0;
	}
	if (gige->buf_gain) {
		delete_rt_buffer(gige->buf_gain); gige->buf_gain = 0;
	}
	if (gige->buf_exposure) {
		delete_rt_buffer(gige->buf_exposure); gige->buf_exposure = 0;
	}

	in_mutex->unlock();

}
//+------------------------------------------------------------------
/**
 *	method:	GigeVCam::modify_configuration
 * Modify configuration parameters (horizontal, vertical calibration)
 *
*/
void GigeVCam::modify_configuration(bool ccd_param, bool calib_param)
{
	
	int32_t cur_cam_cnt = 0, conf_size, cur_cameras;
	vector<string>	cur_configuration, new_configuration;
	cam_param_t *cur_cam_param;
	Tango::DbData db_data;

	try {
		db_data.push_back(Tango::DbDatum("Configuration"));
		/* global configuration */
		if (global_flag_prop) {
			Tango::Database* db = new Tango::Database();
			db->get_property("Gigecam",db_data);
		}
		/* local configuration */
		else {
			get_db_device()->get_property(db_data); 
		}
	}
	catch (CORBA::Exception &e) {
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't read configuration",
			(const char *)"GigeVCam::modify_configuration()");
	}
	/* read data */
	if (db_data[0].is_empty()) { 
		Tango::Except::throw_exception (
			(const char *)"Can't read global configuration",
			(const char *)"Configuration is empty",
			(const char *)"GigeVCam::modify_configuration()");
	}
	else {
		db_data[0] >> cur_configuration;
	}

	/* local configuration size */
	cur_cameras = cur_configuration.size();

	cur_cam_param = new cam_param_t[cur_cameras];
	if (cur_cam_param == 0) {
		Tango::Except::throw_exception (
			(const char *)"Can't locate memory",
			(const char *)"Configuration failed",
			(const char *)"GigeVCam::modify_configuration()");
	}


	for (vector<string>::const_iterator it = cur_configuration.begin(); it < cur_configuration.end(); it++) {
		std::string token;
		std::istringstream iss(*it);
		conf_size = 0;
		while ( getline(iss, token, ':') ) {
			switch (conf_size) {
				case 0: cur_cam_param[cur_cam_cnt].name = token; 
					camera_name_index[token.c_str()] = cur_cam_cnt+1; break;
				case 1: cur_cam_param[cur_cam_cnt].ipaddress = token;
					camera_ip_index[token.c_str()] = cur_cam_cnt+1; break; 
				case 2: cur_cam_param[cur_cam_cnt].img_depth = atol(token.c_str());break;
				case 3: cur_cam_param[cur_cam_cnt].exposure = atol(token.c_str());break; 
				case 4: cur_cam_param[cur_cam_cnt].gain = (double) atof(token.c_str());break;
				case 5: cur_cam_param[cur_cam_cnt].mirror = (bool) atol(token.c_str());break;
				case 6: cur_cam_param[cur_cam_cnt].rotation = atof(token.c_str());break;
				case 7: cur_cam_param[cur_cam_cnt].trigger = (bool) atol(token.c_str());break;
				case 8: cur_cam_param[cur_cam_cnt].hor_calibration = (double) atof(token.c_str());break;
				case 9: cur_cam_param[cur_cam_cnt].ver_calibration = (double) atof(token.c_str());break;
				case 10: cur_cam_param[cur_cam_cnt].hor_pos_offset = (double) atof(token.c_str());break;
				case 11: cur_cam_param[cur_cam_cnt].ver_pos_offset = (double) atof(token.c_str());break;
				case 12: cur_cam_param[cur_cam_cnt].hor_sign = (double) atof(token.c_str());break;
				case 13: cur_cam_param[cur_cam_cnt].ver_sign = (double) atof(token.c_str());break;
			}
			if ((++conf_size) > 13)
				break;
  	}

		if (cur_cam_param[cur_cam_cnt].name == cam_param[cam_idx].name) {
	
			if (!global_flag_prop) {
				if ((img8.size() > 0) || (img16.size() > 0))  {
					if (ccd_param) {
						cur_cam_param[cur_cam_cnt].exposure = *attr_Exposure_read;
						cur_cam_param[cur_cam_cnt].gain = *attr_Gain_read;
						cur_cam_param[cur_cam_cnt].trigger = (bool) *attr_Trigger_read;
						cur_cam_param[cur_cam_cnt].img_depth = *attr_ImageDepth_read;
					}
				}
			}

			if (calib_param) {
				if (img8.size() > 0) {
					cur_cam_param[cur_cam_cnt].hor_calibration = img8[img_cur]->hor_calibration;
					cur_cam_param[cur_cam_cnt].ver_calibration = img8[img_cur]->ver_calibration;
					cur_cam_param[cur_cam_cnt].hor_pos_offset = img8[img_cur]->hor_pos_offset;
					cur_cam_param[cur_cam_cnt].ver_pos_offset = img8[img_cur]->ver_pos_offset;
					cur_cam_param[cur_cam_cnt].mirror = img8[img_cur]->get_mirror();
				}
				if (img16.size() > 0) {
					cur_cam_param[cur_cam_cnt].hor_calibration = img16[img_cur]->hor_calibration;
					cur_cam_param[cur_cam_cnt].ver_calibration = img16[img_cur]->ver_calibration;
					cur_cam_param[cur_cam_cnt].hor_pos_offset = img16[img_cur]->hor_pos_offset;
					cur_cam_param[cur_cam_cnt].ver_pos_offset = img16[img_cur]->ver_pos_offset;
					cur_cam_param[cur_cam_cnt].mirror = img16[img_cur]->get_mirror();
				}
				cam_param[cur_cam_cnt].hor_calibration = cur_cam_param[cur_cam_cnt].hor_calibration;
				cam_param[cur_cam_cnt].ver_calibration = cur_cam_param[cur_cam_cnt].ver_calibration;
				cam_param[cur_cam_cnt].hor_pos_offset = cur_cam_param[cur_cam_cnt].hor_pos_offset * cur_cam_param[cur_cam_cnt].hor_sign;
				cam_param[cur_cam_cnt].ver_pos_offset = cur_cam_param[cur_cam_cnt].ver_pos_offset;
				cur_cam_param[cam_idx].rotation = cam_param[cur_cam_cnt].rotation;
			}

		}
		cur_cam_cnt++;
	}

	char conf_str[300];
	for (int32_t i = 0; i < cur_cam_cnt; i++) {
		memset(conf_str,0,300);
		sprintf(conf_str,"%s:%s:%d:%d:%.1f:%d:%.2f:%d:%.2f:%.2f:%.3f:%.3f:%.0f:%.0f",
			cur_cam_param[i].name.c_str(),
			cur_cam_param[i].ipaddress.c_str(),
			cur_cam_param[i].img_depth, 
			cur_cam_param[i].exposure,
			cur_cam_param[i].gain,
			(int)cur_cam_param[i].mirror,
			cur_cam_param[i].rotation,
			(int)cur_cam_param[i].trigger, 
			cam_param[i].hor_calibration,
			cam_param[i].ver_calibration, 
			cam_param[i].hor_pos_offset,
			cam_param[i].ver_pos_offset,
			(double)cur_cam_param[i].hor_sign,
			(double)cur_cam_param[i].ver_sign);
		new_configuration.push_back(conf_str);
	}

	/* now write into database */

	Tango::Database *db = new Tango::Database(host_rw,port_rw);
	try {
		Tango::DbDatum config("Configuration");
		Tango::DbData db_data_put;
		config << new_configuration;
		db_data_put.push_back(config);
		if (global_flag_prop) {
			db->put_property("Gigecam",db_data_put);
		}
		else {
			db->put_device_property(get_name(), db_data_put);
		}
	}
	catch (Tango::DevFailed &e) {
		delete db;
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't write into database",
			(const char *)"GigeVCam::modify_configuration()");				
	}


	if ((!global_flag_prop) && ccd_param) {
		cur_cam_cnt = 0;

		/* fast calculation 0=disable, 1=enable */
		try {
			Tango::DbDatum config("FastCalculation");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			if (cam_param[cur_cam_cnt].fast_calc)
				sprintf(conf_str,"true");
			else
				sprintf(conf_str,"false");
			string val(conf_str); config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* 0=disable automatic roi, 1=enable automatic roi */
		try {
			Tango::DbDatum config("AutoRoi");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			if (cam_param[cur_cam_cnt].auto_roi)
				sprintf(conf_str,"true");
			else
				sprintf(conf_str,"false");
			string val(conf_str); config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* 0=disable manual roi, 1=enable manual roi */
		try {
			Tango::DbDatum config("SwRoi");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			if (cam_param[cur_cam_cnt].sw_roi)
				sprintf(conf_str,"true");
			else
				sprintf(conf_str,"false");
			string val(conf_str); config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* 0=disable hw roi, 1=enable hw roi */
		try {
			Tango::DbDatum config("HwRoi");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			if (cam_param[cur_cam_cnt].hw_roi)
				sprintf(conf_str,"true");
			else
				sprintf(conf_str,"false");
			string val(conf_str); config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* 0=rms 2D, 1=gaussian fit, 2=asym gaussian fit, 3=confiteor */
		try {
			Tango::DbDatum config("ProcessMode");
			Tango::DbData db_data_put;
			memset(conf_str,0,300); sprintf(conf_str,"%d",(int)cam_param[cam_idx].process_mode);
			string val(conf_str); config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* 0=rms 2D, 1=gaussian fit, 2=asym gaussian fit, 3=confiteor */
		try {
			Tango::DbDatum config("RoiThreshold");
			Tango::DbData db_data_put;
			memset(conf_str,0,300); sprintf(conf_str,"%.3f",cam_param[cam_idx].roi_threshold);
			string val(conf_str); config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* search background param */
		try {
			Tango::DbDatum config("SearchBackgroundParam");
			Tango::DbData db_data_put;
			vector<string>	val;
			for (int i = 0; i < 2; i++) {
				memset(conf_str,0,300);
				sprintf(conf_str,"%.2f",cam_param[cam_idx].search_background_param[i]);
				val.push_back(conf_str);
			}
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* manual ROI x1,y1,x2,y2 */
		try {
			Tango::DbDatum config("RoiParam");
			Tango::DbData db_data_put;
			vector<string>	val;
			for (int i = 0; i < 4; i++) {
				memset(conf_str,0,300);
				sprintf(conf_str,"%d",(int)cam_param[cam_idx].manual_roi_param[i]);
				val.push_back(conf_str);
			}
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		// reference target position
		try {
			Tango::DbDatum config("RefPosOffset");
			Tango::DbData db_data_put;
			vector<string>	val;
			memset(conf_str,0,300);
			sprintf(conf_str,"%.4f",cam_param[cam_idx].ref_hor_pos_offset);
			val.push_back(conf_str);
			memset(conf_str,0,300);
			sprintf(conf_str,"%.4f",cam_param[cam_idx].ref_ver_pos_offset);
			val.push_back(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		// automatic target value
		try {
			Tango::DbDatum config("AutoTargetValue");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",(int)cam_param[cam_idx].auto_target_value);
			string val(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* auto feedback gain */
		try {
			Tango::DbDatum config("AutoFeedbackGain");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			sprintf(conf_str,"%.3f",cam_param[cam_idx].auto_feedback_gain);
			string val(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		try {
			Tango::DbDatum config("AutoFeedbackDws");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",cam_param[cam_idx].auto_feedback_dws);
			string val(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		try {
			Tango::DbDatum config("AutoFeedbackDeadband");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",cam_param[cam_idx].auto_feedback_deadband);
			string val(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		try {
			Tango::DbDatum config("AutoFeedbackTargetThres");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",cam_param[cam_idx].auto_feedback_target_thres);
			string val(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		try {
			Tango::DbDatum config("Binning");
			Tango::DbData db_data_put;
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",cam_param[cam_idx].binning);
			string val(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* manual ROI x1,y1,x2,y2 */
		try {
			Tango::DbDatum config("HwRoiParam");
			Tango::DbData db_data_put;
			vector<string>	val;
			for (int i = 0; i < 4; i++) {
				memset(conf_str,0,300);
				sprintf(conf_str,"%d",(int)cam_param[cam_idx].hw_roi_param[i]);
				val.push_back(conf_str);
			}
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		/* auto roi param: grid size, lp order, roi threshold */
		try {
			Tango::DbDatum config("AutoRoiParam");
			Tango::DbData db_data_put;
			vector<string>	val;
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",(int)cam_param[cam_idx].gridsize);val.push_back(conf_str);
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",(int)cam_param[cam_idx].lp_order);val.push_back(conf_str);
			memset(conf_str,0,300);
			sprintf(conf_str,"%.1f",cam_param[cam_idx].roi_threshold);val.push_back(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		// automatic gain auto param
		try {
			Tango::DbDatum config("GainAutoParam");
			Tango::DbData db_data_put;
			vector<string>	val;
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",(int)cam_param[cam_idx].gain_auto);val.push_back(conf_str);
			memset(conf_str,0,300);
			sprintf(conf_str,"%.1f",cam_param[cam_idx].gain_auto_min);val.push_back(conf_str);
			memset(conf_str,0,300);
			sprintf(conf_str,"%.1f",cam_param[cam_idx].gain_auto_max);val.push_back(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

		// automatic exposure auto param
		try {
			Tango::DbDatum config("ExposureAutoParam");
			Tango::DbData db_data_put;
			vector<string>	val;
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",cam_param[cam_idx].exposure_auto);val.push_back(conf_str);
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",cam_param[cam_idx].exposure_auto_min);val.push_back(conf_str);
			memset(conf_str,0,300);
			sprintf(conf_str,"%d",cam_param[cam_idx].exposure_auto_max);val.push_back(conf_str);
			config << val;
			db_data_put.push_back(config);
			db->put_device_property(get_name(), db_data_put);
		}
		catch (Tango::DevFailed &e) {
			delete db;
			Tango::Except::throw_exception (
				(const char *)"Failed to update configuration",
				(const char *)"Can't write into database",
				(const char *)"GigeVCam::modify_configuration()");				
		}

	}
	
	delete db; 

}


/**
 *	method:	GigeVCam::set_roi_param
 * Modify configuration parameters (horizontal, vertical calibration)
 *
*/
void GigeVCam::set_roi_param(int32_t *roi_array_p)
{
	
	/* increments must be multiple of 4 */
	for (int32_t i = 0; i < 4; i++) {
		cam_param[cam_idx].manual_roi_param[i] = (roi_array_p[i] & 0xfffff0);
	}

	cam_param[cam_idx].dis_enable = false; /* disable digital image stabilization */

}



void GigeVCam::info_log(string str)
{
	/*INFO_STREAM << str << endl;*/
	struct timeval timenow;
	gettimeofday(&timenow,NULL);
	cout << ctime(&(timenow.tv_sec)) << __FUNCTION__ << ": " << str << endl;	
}

timespec GigeVCam::difftime(timespec start, timespec end)
{
	timespec temp;
	if ((end.tv_nsec-start.tv_nsec)<0) {
		temp.tv_sec = end.tv_sec-start.tv_sec-1;
		temp.tv_nsec = 1000000000+end.tv_nsec-start.tv_nsec;
	} else {
		temp.tv_sec = end.tv_sec-start.tv_sec;
		temp.tv_nsec = end.tv_nsec-start.tv_nsec;
	}
	return temp;
}

double GigeVCam::timespec2ms(timespec ts)
{
	return (double)ts.tv_sec+(double)ts.tv_nsec/1000000000;
}


//+------------------------------------------------------------------
/**
 *	method:	GigeVCam::string_explode
 * Explode a string 
*/
//+------------------------------------------------------------------
void GigeVCam::string_explode(string str, string separator, vector<string>* results)
{
	size_t found;
	
	found = str.find_first_of(separator);
	while(found != string::npos) {
		if(found > 0) {
			results->push_back(str.substr(0,found));
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		results->push_back(str);
	}

}


//+------------------------------------------------------------------
/**
 *	method:	GigeVCam::erase_sub_str
 * Erase First Occurrence of given  substring from main string. 
*/
//+------------------------------------------------------------------
void GigeVCam::erase_sub_str(std::string & mainStr, const std::string & toErase)
{
    // Search for the substring in string
    size_t pos = mainStr.find(toErase);
    if (pos != std::string::npos)
    {
        // If found then erase it from string
        mainStr.erase(pos, toErase.length());
    }
}

//+------------------------------------------------------------------
/**
 *	method:	GigeVCam::perform_permutation
 * Permutate image pixel by reading an attribute
*/
//+------------------------------------------------------------------
void GigeVCam::perform_permutation(bool enable)
{
	Tango::DeviceProxy *dev_permutation_table;

	if (permutationTable.size() != 2) {
		ERROR_STREAM << "Permutation table not configured" << endl;
		Tango::Except::throw_exception (
				(const char *)"Failed to enable image permutation",
				(const char *)"Permutation table not configured",
				(const char *)"GigeVCam::perform_permutation()");			
	}
	//permutation is true
	if (enable == true) {
		vector <Tango::DevDouble> permutation_table_vec[2],permutation_table_vec_tmp;

		// 0=horizontal,1=vertical
		for (int k = 0; k < 2; k++) {		
			vector<string> results;		
			string_explode(permutationTable[k], "/", &results);
		
			if (results.size() < 2) {
				ERROR_STREAM << "Invalid permutation table device names" << endl;				
				Tango::Except::throw_exception (
					(const char *)"Failed to enable image permutation",
					(const char *)"Invalid permutation table attribute name",
					(const char *)"GigeVCam::perform_permutation()");	
		
			}
			string device_name = permutationTable[k];
			// get the attribute name
			string attr_name = results[results.size()-1];	
			// remove attribute from device name
			erase_sub_str(device_name,attr_name);
			// remove "/" last char
			if (device_name.size () > 0)  device_name.resize (device_name.size () - 1);
			
			cout << __FUNCTION__ << " " << device_name << " " << attr_name << endl;
			
			// read permutation table from an attribute
			try {
				dev_permutation_table = 0;
				dev_permutation_table = new Tango::DeviceProxy(device_name);
				dev_permutation_table->read_attribute(attr_name) >> permutation_table_vec[k];
				delete dev_permutation_table;
			}	
			catch (Tango::DevFailed &e) {
				if (dev_permutation_table) 
					delete dev_permutation_table;
				ERROR_STREAM << "Failed to read permutation table" << endl;	
				Tango::Except::re_throw_exception(e,
					"Failed to read permutation table", "Failed to calculate permutation vector",
					__FUNCTION__, Tango::ERR);										
			}						

		}
		
		// get current image size
		long img_width,img_height;
		if (img8.size() > 0) { 
			img_height = img8[0]->get_height();
			img_width = img8[0]->get_width();
		}
		else if (img16.size() > 0) {
			img_height = img16[0]->get_height();
			img_width = img16[0]->get_width();		
		}
		else {
			ERROR_STREAM << "Software misconfiguration" << endl;			
			Tango::Except::throw_exception (
				(const char *)"Failed to enable image permutation",
				(const char *)"Software misconfiguration",
				__FUNCTION__);		
		}
		
		for (int32_t i = 0; i < img_height*img_width; i++) {
			cam_param[cam_idx].img_permutation[i] = 0;
			permutation_table_vec_tmp.push_back(0);			
		}	
		
		// hw_roi == false			
		if (cam_param[cam_idx].hw_roi == false) {
			int32_t cnt = 0;
			for (int32_t i = 0; i < img_height; i++) {
				for (int32_t j = 0; j < img_width; j++) {
					int32_t idx = j + i * img_width,
						offset_hor = (int32_t) permutation_table_vec[0][idx],
						offset_ver = (int32_t) permutation_table_vec[1][idx];										

					int32_t idx_perm = idx + offset_hor + offset_ver *img_width;
					if ((((j + offset_hor) >= 0) && ((j + offset_hor) < img_width) &&
						((i + offset_ver) >= 0) && ((i + offset_ver) < img_height)) && 
						(idx_perm < img_width*img_height)) {	

						cam_param[cam_idx].img_permutation[cnt] = idx_perm;
						
					}
					cnt++;
				}						
			}
		}
		
		
		
		// hw_roi == true
		// the permutation table takes into account the full image			
		else {	
			int32_t cnt = 0;
			for (int32_t i = cam_param[cam_idx].hw_roi_param[1]; i < cam_param[cam_idx].hw_roi_param[3]; i++) {
				for (int32_t j = cam_param[cam_idx].hw_roi_param[0]; j < cam_param[cam_idx].hw_roi_param[2]; j++) {
					long idx = (j + i * cam_param[cam_idx].max_width),
						offset_hor = permutation_table_vec[0][idx],
						offset_ver = permutation_table_vec[1][idx],
						i_red = i - cam_param[cam_idx].hw_roi_param[1],
						j_red = j - cam_param[cam_idx].hw_roi_param[0];

					int32_t idx_perm = cnt + offset_hor + offset_ver * img_width;									
					if ((((j_red + offset_hor) >= 0) && ((j_red + offset_hor) < img_width) &&
						((i_red + offset_ver) >= 0) && ((i_red + offset_ver) < img_height)) && 
						(idx_perm < img_width*img_height)) {	
						
						cam_param[cam_idx].img_permutation[cnt] = idx_perm;
									
					}
					cnt++;
				}		
			}
		}
			
	}
	// permutation is false
	else {
		if (img8.size() > 0) {
			for (int32_t i = 0; i < img8[0]->get_height()*img8[0]->get_width(); i++) {
				cam_param[cam_idx].img_permutation[i] = i;
			}		
		}
		else if (img16.size() > 0) {
			for (int32_t i = 0; i < img16[0]->get_height()*img16[0]->get_width(); i++) {
				cam_param[cam_idx].img_permutation[i] = i;
			}				
		}
			
	}


	// after applying matrix permutation then apply the other image transformations
	if (img8.size() > 0) {
		/* reset permutation vector */
		if (cam_param[cam_idx].mirror)
			img8[0]->permutation_mirrorh(cam_param[cam_idx].img_permutation);
		if (cam_param[cam_idx].rotation != 0)
			img8[0]->permutation_rotation(cam_param[cam_idx].rotation,cam_param[cam_idx].img_permutation);

		for (uint32_t i = 0; i < img8.size(); i++) {
			img8[i]->reset();
		}
	}
	else if (img16.size() > 0) {
		/* reset permutation vector */
		for (int32_t i = 0; i < img16[0]->get_height()*img16[0]->get_width(); i++) 
			cam_param[cam_idx].img_permutation[i] = i;

		if (cam_param[cam_idx].mirror)
			img16[0]->permutation_mirrorh(cam_param[cam_idx].img_permutation);
		if (cam_param[cam_idx].rotation != 0)
			img16[0]->permutation_rotation(cam_param[cam_idx].rotation,cam_param[cam_idx].img_permutation);

		for (uint32_t i = 0; i < img16.size(); i++) {
			img16[i]->reset();
		}
	}
	
}


	/*----- PROTECTED REGION END -----*/	//	GigeVCam::namespace_ending
} //	namespace
